/*
 * include/revirt/rlog.h
 *
 * This file contains definitions relating to the revirt logging system.
 */
#ifndef _RLOG_H
#define _RLOG_H

//#include <linux/sched.h>

#include <linux/types.h>

#include <revirt/rlog_min.h>
#include <revirt/replay.h>
#include <revirt/syscalls.h>
/*
 * Common definitions, used by both kernel and userland
 */
#define RLOG_SCHEDULE_ENTRY (0x1000)

#define ENTRY_SCHED  (0 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_FORK   (1 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXIT   (2 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_TOP    (3 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_SIGNAL	(4 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXECPARM (5 | RLOG_SCHEDULE_ENTRY)

//#define ENTRY_PUT_USER	15 // in rlog_min.h
#define ENTRY_COPY_USER	16
#define ENTRY_SYSRET	17
#define ENTRY_CLEAR_USER 18
#define ENTRY_COPY_USER_NODATA 19
#define ENTRY_SIGFRAME 20
#define ENTRY_RT_SIGFRAME 21
#define ENTRY_MARKER 22
#define ENTRY_SENDERIP 23

#define REPLAY_PBUF_MAX_SIZE 10

struct revirt_options_xreplay_st {
    int packetrate;
    int packetsize;
};

#ifdef __KERNEL__
typedef  __kernel_fd_set __revirt_syscall_set;
#else
typedef struct {
	unsigned long fds_bits[32];
} __revirt_syscall_set;
#endif

#ifdef __KERNEL__
typedef  __kernel_fd_set __revirt_fd_set;
#else
typedef struct {
	unsigned long fds_bits[8];
} __revirt_fd_set;
#endif
struct revirt_options_st { 
	/* FIXME -- there's really no need for two of these */
    unsigned int logflags;
    unsigned int replayflags;
    struct revirt_options_xreplay_st xreplay;
    int network_fd;
    __revirt_fd_set unlogged_fds;
    __revirt_syscall_set allowed_syscalls;
    struct {
	int size;
	unsigned int mask;
    } logbuf;
};

#define REPLAY_OPT_LOG_WHOLE_SIGFRAME	(1 << 0)
#define REPLAY_OPT_LOG_WHOLE_PTREGS		(1 << 1)
#define REPLAY_OPT_LOG_WHOLE_SIGINFO	(1 << 2)
#define REPLAY_OPT_LOG_SYSCALL_LOGS_BRANCHCOUNT	(1 << 3)
#define REPLAY_OPT_LOG_ALL_SYSCALLS		(1 << 4)
#define REPLAY_OPT_LOG_LOG_SENDERIP		(1 << 5)
#define REPLAY_OPT_REPLAY_DEBUG		(1 << 8)
#define	REPLAY_OPT_ZTRACE_DEBUG		(1 << 9)
#define REPLAY_OPT_THROTTLE_NETWORK	(1 << 10)
#define REPLAY_OPT_WARNING_HALTS	(1 << 11)
#define REPLAY_OPT_RESTRICT_SYSCALLS	(1 << 12)
#define REPLAY_OPT_LOGSYS_ON_REPLAY	(1 << 13)
#define REPLAY_OPT_VALUE		(1 << 16)
#define REPLAY_OPT_PACKETRATE		(REPLAY_OPT_VALUE | 1)
#define REPLAY_OPT_PACKETSIZE		(REPLAY_OPT_VALUE | 2)
#define REPLAY_OPT_NETWORKFD		(REPLAY_OPT_VALUE | 3)
#define REPLAY_OPT_UNLOGGEDFDS		(REPLAY_OPT_VALUE | 4)
#define REPLAY_OPT_LOGBUFSIZE		(REPLAY_OPT_VALUE | 5)

#define REPLAY_OPT_LOGMASK (0xff)
#define REPLAY_OPT_REPLAYMASK (0xff << 8)

//#define REPLAY_RECORD_ALL_EXCEPTIONS # DEFINED IN syscalls.h
#ifdef CONFIG_REVIRT_ALL_SYSCALLS
# warning "Logging all system calls"
# define REPLAY_ALL_SYSCALLS
#endif
//#define REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#ifdef CONFIG_REVIRT_SYSCALL_LOGS_BRANCHCOUNT
# define REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#endif
//#define REPLAY_WHOLE_PTREGS
#ifdef CONFIG_REVIRT_WHOLE_PTREGS
# define REPLAY_WHOLE_PTREGS
#endif
//#define REPLAY_WHOLE_SIGFRAME
#ifdef CONFIG_REVIRT_WHOLE_SIGFRAME
# define REPLAY_WHOLE_SIGFRAME
#endif
//#define REPLAY_WHOLE_SIGINFO
#ifdef CONFIG_REVIRT_WHOLE_SIGINFO
# define REPLAY_WHOLE_SIGINFO
#endif
//#define REPLAY_RECORD_ALL_PFCTRS
#ifdef CONFIG_REVIRT_RECORD_ALL_PFCTRS
# define REPLAY_RECORD_ALL_PFCTRS
#endif
//#define REPLAY_LOG_SENDERIP
#define REPLAY_THROTTLE_NETWORK
#define	REPLAY_UNIQID
#define	REPLAY_ONLY_USR_EXCEPTIONS

#define IS_SCHED(x) ( (x) & RLOG_SCHEDULE_ENTRY )

//#define UML_NETWORK_FD 0x44 // 2.2.20 guest
//#define UML_NETWORK_FD 0x56 // 2.4.18 guest
#define UML_NETWORK_FD 87 // 2.4.18 guest
#define UML_IS_NETWORK_FD(fd) ((fd)==(revirt_options.network_fd))
#define UML_IS_UNLOGGED_FD(fd) (FD_ISSET((fd), &revirt_options.unlogged_fds))

#define REPLAY_DEFAULT_LOGFLAGS REPLAY_OPT_RESTRICT_SYSCALLS
#define REPLAY_DEFAULT_REPLAYFLAGS \
   (REPLAY_OPT_WARNING_HALTS \
    |REPLAY_OPT_RESTRICT_SYSCALLS \
    |REPLAY_OPT_THROTTLE_NETWORK)


#define REVIRT_EFLAGS_MASK (0x40DD5)

#ifndef __KERNEL__
typedef long long s64;
#endif

struct revirt_ctrdata {
	s64 hw_branch;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	int exceptions[REVIRT_NR_EX];
#endif
#ifdef REPLAY_RECORD_ALL_PFCTRS
	unsigned int hw_intr;
	unsigned int ex_noerror;
	unsigned int ex_error;
	s64 counter[REVIRT_NR_PFCTRS];
#endif
};

/*
 * This is all calculated in copy_branchdata now.
 */
#define ADJ_BC(x) ((s64)(x).hw_branch )
#define ADJ_BC_R(x) ADJ_BC(x)

struct rlog_header {
	int type;
	unsigned long len;
#ifdef REPLAY_UNIQID
	unsigned long id;
#endif
};


struct revirt_senderip {
	unsigned long ip;
};

struct rlog_exit_entry {
	short type;
	int pid;
};

struct rlog_schedule_entry {
	short type;
	unsigned int flags;
	int pid;
	unsigned long eip;
	int data;
	struct revirt_ctrdata ctr;
	unsigned int edi;
	unsigned int nrxuser;
	unsigned long wchan;
};

#if 0
struct rlog_entry_fork {
	short type;
	int pid;
	int child;
};
#endif

#if 0 // in rlog_min.h
struct rlog_put_user_entry {
	short type;
	unsigned long xx, pptr;
	short ssize, rretval;
	int pid, syscall;
};
#endif

struct rlog_copy_user_entry {
	short type;
	unsigned long tto;
	unsigned long ssize;
	int pid;
	int nrsyscalls;
	int syscall;
};

struct rlog_clear_user_entry {
	short type;
	unsigned long to;
	unsigned long n;
	int pid, syscall;
};

# include <asm/ptrace.h>

struct __min_pt_regs { unsigned int eax; };

struct rlog_entry_sysret {
	short type;
	int pid;
	int nrxuser;
	int syscall;
#ifdef REPLAY_WHOLE_PTREGS
	struct pt_regs regs;
#else
	struct __min_pt_regs regs;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	struct revirt_ctrdata ctr;
#endif
};


/*
 * Stupid, stupid, stupid...
 */
#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/time.h>
#include <linux/times.h>
#include <asm/signal.h>

#else

#include <linux/posix_types.h>

#define _NSIG_K         64
#define _NSIG_BPW_K     32
#define _NSIG_WORDS_K   (_NSIG_K / _NSIG_BPW_K)

typedef struct {
        unsigned long sig[_NSIG_WORDS_K];
} __kernel_sigset_t;

/* from linux/time.h */
/*
 * Needed by struct rlog_marker_entry
 */
struct __kernel_timeval {
	__kernel_time_t		tv_sec;
	__kernel_suseconds_t	tv_usec;
};

#endif

struct rlog_marker_entry {
	short type;
	int mark;
#ifdef __KERNEL__
	struct timeval tv;
#else
	struct __kernel_timeval tv;
#endif
};

struct rlog_entry_signal {
	short type;
	int pid;
//	int n;		/* Number of signals in the queue */
#ifndef REPLAY_WHOLE_SIGINFO
	int si_signo, si_errno, si_code;
	int _pad[2];
#endif
#ifdef __KERNEL__	
	sigset_t blocked;
	sigset_t pending_signal;
	sigset_t dfl;
#else
	__kernel_sigset_t blocked;
	__kernel_sigset_t pending_signal;
	__kernel_sigset_t dfl;
#endif
};

#include <asm/sigcontext.h>

#ifdef __KERNEL__

# include <asm/signal.h>
# include <asm/siginfo.h>
# include <asm/ucontext.h>

#else  /* !__KERNEL__ */
# define _NSIG           64
# define _NSIG_BPW       32
# define _NSIG_WORDS     (_NSIG / _NSIG_BPW)

# define SI_MAX_SIZE     128
# define SI_PAD_SIZE     ((SI_MAX_SIZE/sizeof(int)) - 3)

typedef union __kernel_sigval {
	int sival_int;
	void * sival_ptr;
} __kernel_sigval_t;

struct __kernel_siginfo {
	int si_signo;
	int si_errno;
	int si_code;

	union {
		int _pad[SI_PAD_SIZE];
		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
		} _kill;

		struct {
			unsigned int _timer1;
			unsigned int _timer2;
		} _timer;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			__kernel_sigval_t _sigval;
		} _rt;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			int _status;
			__kernel_clock_t _utime;
			__kernel_clock_t _stime;
		} _sigchld;

		struct {
			void * _addr;
		} _sigfault;

		struct {
			int _band;
			int _fd;
		} _sigpoll;
	} _sifields;
};

typedef struct __kernel_sigaltstack {
	void * ss_sp;
	int ss_flags;
	__kernel_size_t ss_size;
} __kernel_stack_t;

struct __kernel_ucontext {
	unsigned long uc_flags;
	struct __kernel_ucontext * uc_link;
	__kernel_stack_t	uc_stack;
	struct sigcontext uc_mcontext;
	__kernel_sigset_t uc_sigmask; // defined above
};

#endif /* __KERNEL__ */


/* from arch/i386/kernel/signal.c */
struct __kernel_sigframe {
	char * pretcode;
	int sig;
	struct sigcontext sc;
	struct _fpstate fpstate;
	unsigned long extramask[_NSIG_WORDS-1];
	char retcode[8];
};

struct __kernel_rt_sigframe {
	char *pretcode;
        int sig;
#ifdef __KERNEL__
        struct siginfo *pinfo;
#else
        struct __kernel_siginfo *pinfo;
#endif
        void *puc;
#ifdef __KERNEL__
        struct siginfo info;
        struct ucontext uc;
#else
        struct __kernel_siginfo info;
        struct __kernel_ucontext uc;
#endif
        struct _fpstate fpstate;
        char retcode[8];
};

/*
 * struct sigrame is declared inside signal.c, so even the kernel needs
 * to re-declare it here...
 */
struct rlog_entry_sigframe {
	short type;
	int pid;
	struct __kernel_sigframe frame;
};

struct rlog_entry_rt_sigframe {
	short type;
	int pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	struct __kernel_rt_sigframe frame;
#else
	int sig;
	unsigned long trapno; // frame.uc.uc_mcontext.trapno
#endif
};

struct rlog_entry_scall_lexecve {
	int type;
	int pid;
	int fname_index;
	int argv_index;
	int env_index;
};

/*
 * Kernel-only definitions
 */

#ifdef __KERNEL__

#define LOGPAGESIZE 4096
#define NLOGSPERPAGE (LOGPAGESIZE/sizeof(struct log_schedule_entry))

#include <revirt/syscalls.h>

void save_perfctr(struct task_struct *prev);
void revirt_exit(struct task_struct *);

void revirt_log_schedule(struct task_struct* tsk);
void revirt_log_exit(void);
asmlinkage void revirt_log_sysret(struct pt_regs  * regs);
void revirt_log_sigframe(struct __kernel_sigframe * frame);
void revirt_log_signal(siginfo_t * sig);
void revirt_log_rt_sigframe(struct __kernel_rt_sigframe * frame);

void ztrace_tsk_gets_signal(void);
void ztrace_tsk_debug(void);
void ztrace_tsk_perf(void);

int revirt_replay_sigframe(struct __kernel_sigframe *, 
			struct __kernel_sigframe *);
int revirt_replay_rt_sigframe(struct __kernel_rt_sigframe *, 
			struct __kernel_rt_sigframe *);

int revirt_read_open(char *);
int revirt_read_log(struct rlog_header *);
void clear_branchdata(struct task_struct *);
void copy_branchdata(struct task_struct *, struct revirt_ctrdata *);
unsigned int revirt_getwchan(struct task_struct *);

int revirt_alloc_log_buf(int);

#ifdef REPLAY_UNIQID
int rlog_resetid(unsigned int i);
#endif

#include <linux/uio.h>
extern unsigned int revirt_interrupts;
extern unsigned int revirt_exceptions1;
extern unsigned int revirt_exceptions2;
extern unsigned int revirt_current_syscall;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
extern unsigned int revirt_ecount[REVIRT_NR_EX];
#endif
extern struct revirt_options_st revirt_options;

extern unsigned int revirt_edi;

extern s64 zt_perferr;

extern char revirt_scall_buf[];

# endif /* __KERNEL__ */

#endif /* _RLOG_H */
