/*
 * log.c - Functions associated with ReVirt logging system
 *
 * Copyright (C) 2002-2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

/*
 * FIXME -- inline, put in .h if necessary
 */
void copy_branchdata(struct task_struct *from, 
			    struct revirt_ctrdata *to) {
	struct revirt_tsk_s * rtsk = from->revirt_tsk;
	if(!rtsk)
		return;
	if(from == current) {
		save_perfctr_regs(from);
	}

#ifdef REPLAY_DEBUG
	if(rtsk->perfctr[0].ctr >= (1LL<<32)) {
		printk("copy_branchdata: warning: large count: 0x%llx\n",
			rtsk->perfctr[0].ctr);
	}
#endif

	to->hw_branch=rtsk->perfctr[revirt_pfctr.overflow.counter_index].ctr;
	if(revirt_pfctr.subtract_exceptions)
		to->hw_branch-=(rtsk->swctr[0]+rtsk->swctr[1]+rtsk->swctr[2]);
	if(from->revirt_flags & REPLAY_REPLAY)
		to->hw_branch+=zt_perferr;
#ifdef REPLAY_RECORD_ALL_PFCTRS
        to->hw_intr=rtsk->swctr[0];
	to->ex_noerror = rtsk->swctr[1];
	to->ex_error= rtsk->swctr[2];

	do {
	    int i;
	    for(i=0; i<REVIRT_NR_PFCTRS; i++)
		to->counter[i]=rtsk->perfctr[i].ctr;
	} while(0);
#endif

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do {
	    int i;
	    for(i=0; i<REVIRT_NR_EX; i++)
	    	to->exceptions[i]=rtsk->ecount[i];
	} while(0);
#endif
}

void clear_branchdata(struct task_struct *tsk) {
	int i;
	struct revirt_tsk_s * rtsk = tsk->revirt_tsk;
	if(rtsk==NULL)
		return;
	for(i=0; i<REVIRT_NR_PFCTRS; i++)
		rtsk->perfctr[i].ctr=0;
	for(i=0; i<REVIRT_NR_SWCTR; i++)
		rtsk->swctr[i]=0;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do { 
		int i; 
		for(i=0; i<REVIRT_NR_EX; i++)  {
    		    rtsk->ecount[i]=0; 
		}
	} while(0);
#endif
	if(tsk == current) {
		restore_perfctr_regs(tsk);
	}
}


/*
 * Copied from arch/i386/kernel/process.c:get_wchan()
 */
extern void scheduling_functions_start_here(void);
extern void scheduling_functions_end_here(void);
#define first_sched     ((unsigned long) scheduling_functions_start_here)
#define last_sched      ((unsigned long) scheduling_functions_end_here)

unsigned int revirt_getwchan(struct task_struct * prev) {
          unsigned long ebp, esp, eip;
          unsigned long stack_page;
          int count=0;

	  stack_page = (unsigned long) prev;
          esp=prev->thread.esp;
          if(!stack_page || esp < stack_page || esp > 8188+stack_page)
                count=16;
          ebp = * (unsigned long *) esp;
          do {
                if(ebp < stack_page || ebp > 8184+stack_page)
                        return 0;
                eip = *(unsigned long *) (ebp + 4);
                if(eip < first_sched || eip >= last_sched) {
                        return eip;
                }
                ebp = * (unsigned long *) ebp;
                        
          } while(count++<16);
	  return 0;
}

void revirt_log_schedule(struct task_struct* tsk) {
        struct rlog_schedule_entry le;


	if(!(tsk->revirt_flags & REPLAY_LOG)) {
	    printk("revirt_log_schedule called for unlogged process. \n pid: 0x%x\n" \
	    	   " revirt_flags: 0x%x\n", 
	    		tsk->pid, tsk->revirt_flags);
	    BUG();
	}
	if(!tsk->revirt_tsk) {
		printk("revirt_log_schedule: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	/*
	 * Don't bother logging schedules after a process has called
	 * do_exit.
	 */
	if(tsk->revirt_flags & REPLAY_ZOMBIE)
		return;

	le.type=ENTRY_SCHED;

	copy_branchdata(tsk,&le.ctr);
        le.pid= tsk->pid;

	le.eip=zreg(tsk,eip);
	le.data=zreg(tsk,ecx);

	le.nrxuser = tsk->revirt_tsk->nrxuser;
	le.flags = tsk->revirt_flags;
	le.flags |= REPLAY_TSK_SETSTATE(tsk->state);

	if(tsk->revirt_flags & REPLAY_IN_COPY_USER) {
	  le.edi = revirt_edi;
	  printk("revirt_log_schedule() called in __copy_user_atom. edi==0x%x\n", 
	  		revirt_edi);
	 }
	else
	  le.edi = 0;

	/*
	 * If we're blocked, find out where.  Code copied from
	 * arch/i386/kernel/process.c:get_wchan().  Note that the
	 * switch_to() macro has already stored the proper esp in the
	 * task struct, so the fact that tsk == current is OK.
	 *
	 * FIXME are we sure that tsk->state != TASK_RUNNING implies
	 * that revirt_log_schedule() was called from __switch_to?
	 *
	 * OK, for now we look anytime process is in umode.  Will this break
	 * anything?
	 */
	le.wchan=0;

	if(tsk->state != TASK_RUNNING) {
		le.wchan = revirt_getwchan(tsk);
	} 

	clear_branchdata(tsk);
	/* FIXME -- this should probably be included in clear_branchdata */
	tsk->revirt_tsk->nrsyscalls= 0;

	revirt_write_log((char *)&le, le.type, sizeof(le));
}

#undef last_sched
#undef first_sched

#if 0
void record_fork(struct task_struct *parent,struct task_struct * child) {
        struct rlog_entry_fork le;

        le.type= ENTRY_FORK;
        le.pid= parent->pid;
        le.child=child->pid;

        revirt_write_log((char *)&le, le.type, sizeof(le));
}
#endif

void revirt_log_exit(void) {
        struct rlog_exit_entry ee;

        ee.type= ENTRY_EXIT;
	ee.pid=current->pid;
	revirt_write_log((char *)&ee, ee.type, sizeof(ee));

	/*
	 * Let revirt_log_schedule know not to log anymore schedules.
	 */
	current->revirt_flags |= REPLAY_ZOMBIE;
}

asmlinkage void revirt_log_sysret(struct pt_regs  * regs) {
	struct rlog_entry_sysret se;

	if(!current->revirt_tsk) {
		printk("revirt_log_sysret: pid %d has no revirt_tsk!\n",
				current->pid);
	}

	se.type = ENTRY_SYSRET;
	se.pid = current->pid;
	// sigreturn clobbers orig_eax
	se.syscall = current->revirt_tsk->current_syscall;  
#ifdef REPLAY_WHOLE_PTREGS
	se.regs = *regs;
#else
	se.regs.eax = regs->eax;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	copy_branchdata(current, &se.ctr);
#endif
	se.nrxuser=current->revirt_tsk->nrxuser;

	current->revirt_tsk->nrxuser=0;

	revirt_write_log((char *)&se, se.type, sizeof(se));
}

void revirt_log_sigframe(struct __kernel_sigframe * frame) {
	struct rlog_entry_sigframe e;

	e.type = ENTRY_SIGFRAME;
	e.pid = current->pid;
	e.frame = *frame;

	revirt_write_log((char *)&e, e.type, sizeof(e));
}

void revirt_log_signal(siginfo_t * sig) {
#ifdef REPLAY_WHOLE_SIGINFO
	char d[sizeof(struct rlog_entry_signal) + sizeof(siginfo_t)];
	siginfo_t * si;
#else
	struct rlog_entry_signal le_r;
#endif
	int i, size;
	struct rlog_entry_signal * le;

#ifdef REPLAY_WHOLE_SIGINFO
	le = (struct rlog_entry_signal *)d;
	si = (siginfo_t *) (d + sizeof(struct rlog_entry_signal));

	size=sizeof(struct rlog_entry_signal) + sizeof(siginfo_t);
#else
	le = &le_r;
	size = sizeof(le_r);
#endif
	le->type = ENTRY_SIGNAL;
	le->pid = current->pid;

	/* XXX Should we lock current->sigmask_lock here? */

	le->pending_signal = current->pending.signal;
	le->blocked = current->blocked;

	/*
	 * We need to do this because force_siginfo may change SIG_IGN
	 * to SIG_DFL.
	 * 
	 * Note that there is no 0 signal; so all signals are shifted
	 * down one -- so sigaddset will change the i+1 back to i for
	 * action[0].
	 */

	if(current->sig) {
	    for(i=0; i<_NSIG_WORDS; i++) 
	    	le->dfl.sig[i] = 0;
	    for(i=0; i<_NSIG; i++) {
	    	if(current->sig->action[i].sa.sa_handler == SIG_DFL) {
			sigaddset(&le->dfl, i+1);
		}
	    }
	} else {
		printk("rlog: Strange, log_signals called for zombie...\n");
	}

#ifdef REPLAY_WHOLE_SIGINFO
	*si = *sig;
	revirt_write_log(d, le->type, size);
#else
	le->si_signo = sig->si_signo;
	le->si_errno = sig->si_errno;
	le->si_code = sig->si_code;
	le->_pad[0] = sig->_sifields._pad[0];
	le->_pad[1] = sig->_sifields._pad[1];
	revirt_write_log((char *)le, le->type, size);
#endif
}

void revirt_log_rt_sigframe(struct __kernel_rt_sigframe * frame) {
	struct rlog_entry_rt_sigframe e;

	e.type = ENTRY_RT_SIGFRAME;
	e.pid = current->pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	e.frame = *frame;
#else
	e.sig = frame->sig;
	e.trapno = frame->uc.uc_mcontext.trapno;
#endif
	revirt_write_log((char *)&e, e.type, sizeof(e));
}
