/*
 * logfile.c - Functions associated with reading a ReVirt logfile during replay
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/sched.h>
#include <asm/processor.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <revirt/rlog.h>

#define REPLAY_INPUT_BUF 4096*4

#define SEEK_CUR 1 /* see fs/read_write.c:39 */

static struct file * revirt_scallf = NULL;

/*
 * Yes, this is wasteful. We can optimize it later.
 *
 * Do we need some kind of locks around this?
 */
char revirt_scall_buf[REPLAY_INPUT_BUF];

int revirt_read_open(char * fname) {
	int error;

	error=PTR_ERR(fname);
	if(IS_ERR(fname))
		goto out;

	revirt_scallf = filp_open(fname, O_RDONLY, 0644);
	error=PTR_ERR(revirt_scallf);
	if(IS_ERR(revirt_scallf))
		goto out;
	error=0;
   out:
	return error;
}

int revirt_read_close(void) {
	int ret=0;
	if(revirt_scallf)
	    ret=filp_close(revirt_scallf, NULL);
	revirt_scallf=NULL;
	return ret;
}

int revirt_read_log(struct rlog_header * rh) {
	mm_segment_t oldfs;
	struct rlog_header r;
	int l;

	oldfs= get_fs();
	set_fs(KERNEL_DS);

	/*
	 * If we didn't split the schedule and syscall log entries
	 * into different files, skip scheduling entries.
	 */
	for(;;) {
	    l=(*revirt_scallf->f_op->read)(revirt_scallf, (char *)&r, sizeof(r),
			&revirt_scallf->f_pos);
	    if(l < 0) {
	    	printk("revirt_read_log: Error reading log: %d\n", -l);
		BUG();
	    }
	    if(IS_SCHED(r.type)) {
	    	(*revirt_scallf->f_op->llseek)(revirt_scallf, r.len, SEEK_CUR);
	    } else
	    	break;
	}
	*rh = r;
	l=(*revirt_scallf->f_op->read)(revirt_scallf, revirt_scall_buf, r.len,
			&revirt_scallf->f_pos);
	
	set_fs(oldfs);	

	return l;
}
