/*
 * perfctr-athlon.c - counters for Athlon
 *
 * Copyright (C) 2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <revirt/perfctr.h>
/************************ AMD ATHLON CODE *******************************
 * See "AMD Athlon Processor x86 Code Optimization Guide, appendix D
 ************************************************************************/
#define ATH_EV_INST_RETIRED		0xc0
#define ATH_EV_BR_TAKEN_RETIRED		0xc4
#define ATH_EV_HW_INT_RX		0xcf

#define ATH_USRMODE	 		(1<<16)
#define ATH_OSMODE			(1<<17)
#define ATH_INT				(1<<20)
#define ATH_EN				(1<<22)

#define ATH_EVTSEL			0xC0010000
#define ATH_EVSEL0			0xC0010000
#define ATH_EVSEL1			0xC0010001

#define ATH_CTR				0xC0010004
#define ATH_CTR0			0xC0010004
#define ATH_CTR1			0xC0010005

struct revirt_pfctr_s revirt_pfctr = {
	.count=2,
	.overflow= {
		.control_msr=ATH_EVSEL0,
		.enable_bit=ATH_INT,
		.clear_mask=0,
		.counter_index=0,
	},
	.subtract_exceptions=1,
	.list = {
		/* 0 
		 * Retired taken branches.  This explicitly counts branches,
		 * interrupts, and exceptions; subract interrupts and 
		 * exceptions to make repeatable.
		 */
		{ .counter_msr=ATH_CTR0,
		  .arch={.athlon={.evtsel_msr=ATH_EVSEL0,
			.evtsel_val=ATH_EV_BR_TAKEN_RETIRED | 
					ATH_EN | ATH_USRMODE } } },
		/* 1
		 * Number of taken hardware interrupts. 
		 *
		 * Note: Never seen a difference between sw and hw interrupt
		 * counts; either our platforms don't deliver SMIs, or none
		 * of the counters count them.
		 */
		{ .counter_msr=ATH_CTR1,
		  .arch={.athlon={.evtsel_msr=ATH_EVSEL1,
			.evtsel_val=ATH_EV_HW_INT_RX | 
					ATH_EN | ATH_USRMODE } } },
	}
};

