/*
 * replay.c - ReVirt replay functionality
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

/* Moved to rlog.h */
//#define	EFLAGS_MASK (0x40DD5)
//#define REPLAY_DEBUG

void copy_branchdata(struct task_struct *, struct revirt_ctrdata *);

void check_branchcount(struct revirt_ctrdata * lctr, struct pt_regs * regs,
		struct rlog_header rh) {
    int diff=0, pause=0;
    struct revirt_ctrdata rctr;

    if(!current->revirt_tsk) {
	    printk("check_branchcount: pid %d has no revirt_tsk!\n",
			    current->pid);
    }

    copy_branchdata(current, &rctr);

    if(ADJ_BC(*lctr) != ADJ_BC_R(rctr)) {
        printk("revirt_syscall: maincount differs! %llx %llx (%llx) \n",
	    	ADJ_BC_R(rctr), ADJ_BC((*lctr)), 
		ADJ_BC((*lctr)) - ADJ_BC_R(rctr));
	diff=1;
	pause=1;
    }
    if(diff) {
	int i;
	char buf[512], *c;
#ifdef RECORD_ALL_PFCTRS
	printk("revirt_syscall: #%lu sc %d @ 0x%x\n" \
		" [%llx %x %x %x] [%llx %x %x %x]\n" \
		" [%llx %llx %llx %llx] [%llx %llx %llx %llx]\n",
		rh.id,
		current->revirt_tsk->current_syscall,
		(unsigned int)regs->eip,
	    	rctr.hw_branch,
		rctr.hw_intr,
		rctr.ex_noerror,
		rctr.ex_error,
		lctr->hw_branch,
		lctr->hw_intr,
		lctr->ex_noerror,
		lctr->ex_error,
		rctr.counter[0],
		rctr.counter[1],
		rctr.counter[2],
		rctr.counter[3],
		lctr->counter[0],
		lctr->counter[1],
		lctr->counter[2],
		lctr->counter[3]);
#else
	printk("revirt_syscall: #%lu sc %d @ 0x%x [%llx] [%llx]\n",
		rh.id,
		current->revirt_tsk->current_syscall,
		(unsigned int)regs->eip,
	    	rctr.hw_branch,
		lctr->hw_branch);
#endif
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
 	 c=buf;
	 *c=0;
	 for(i=0; i<REVIRT_NR_EX; i++) {
		 if(rctr.exceptions[i]>0) {
			 c+=sprintf(c, "[%d] %x ", i, rctr.exceptions[i]);
		 }
	 }
	 if(c>buf) {
		 printk("%s\n", buf);
	 }
#endif
   	 if(diff && (revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS))
	    	ztrace_debug_pause();
    }
}

int revirt_replay_sigframe(struct __kernel_sigframe *user_frame, 
				struct __kernel_sigframe *frame){
	struct rlog_entry_sigframe * e;
	struct rlog_header rh;
	int pause, i;

	/*
	 * Read the log
	 */
	revirt_read_log(&rh);
	if(rh.type != ENTRY_SIGFRAME) {
		printk("revirt_replay_sigframe: expecting type %x, got %x (%lu)\n",
				ENTRY_SIGFRAME, rh.type, rh.id);
	}
	e=(struct rlog_entry_sigframe *)revirt_scall_buf;

	/*
	 * Check for any deviation
	 */
	pause=0;
#define LDIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_sigframe: %s differs: %lx %lx\n", \
	  	y, frame->x, e->frame.x);	  	 \
  	  pause = 1;					 \
	}
#define DIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->frame.x);	\
  	  pause = 1;					 \
	}
	DIFF(pretcode, "pretcode");
	DIFF(sig, "sig");
	LDIFF(sc.eax, "eax");
	LDIFF(sc.ebx, "ebx");
	LDIFF(sc.ecx, "ecx");
	LDIFF(sc.edx, "edx");

	LDIFF(sc.edi, "edi");
	LDIFF(sc.esi, "esi");
	LDIFF(sc.ebp, "ebp");
	LDIFF(sc.esp, "esp");
	LDIFF(sc.eip, "eip");
	DIFF(sc.cs, "cs");
	DIFF(sc.ss, "ss");
	DIFF(sc.gs, "gs");
	DIFF(sc.fs, "fs");
	DIFF(sc.es, "es");
	DIFF(sc.ds, "ds");
#undef DIFF
#undef LDIFF
	/*
	 * only the bits masked below are restored by restore_sigcontext()
	 * in arch/i386/kernel/signal.c.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.sc.eflags & REVIRT_EFLAGS_MASK) 
		!= (frame->sc.eflags & REVIRT_EFLAGS_MASK)) {
	    printk("revirt_replay_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->sc.eflags & REVIRT_EFLAGS_MASK,
	    		e->frame.sc.eflags & REVIRT_EFLAGS_MASK);
	    
	    pause=1;
	}

	for(i=0; i<8; i++) {
	 if(e->frame.retcode[i] != frame->retcode[i]) {
	   printk("revirt_replay_sigframe: retcode[%d] differs: %x %x\n",
	   		i, frame->retcode[i], e->frame.retcode[i]);
	   pause=1;
	 }
	}


     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}

	/*
	 * Finally, copy the frame.
	 */
	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
}

int revirt_replay_rt_sigframe(struct __kernel_rt_sigframe *user_frame, 
			struct __kernel_rt_sigframe *frame){
	struct rlog_entry_rt_sigframe * e;
	struct rlog_header rh;
	int pause;

	/*
	 * Read the log
	 */
	revirt_read_log(&rh);
	if(rh.type != ENTRY_RT_SIGFRAME) {
		printk("revirt_replay_rt_sigframe: expecting logtype %x, got %x(%lu)\n",
				ENTRY_RT_SIGFRAME, rh.type, rh.id);
	}
	e=(struct rlog_entry_rt_sigframe *)revirt_scall_buf;
	
	/*
	 * Check for any deviation
	 */
#ifdef REPLAY_DEBUG
	printk("revirt_replay_rt_sigframe: handling signal %x\n", frame->info.si_signo);
#endif

	pause=0;
#ifdef REPLAY_WHOLE_SIGFRAME
#define DIFF(x, y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_rt_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->frame.x);	 \
  	  pause = 1;					 \
	}
	DIFF(pretcode, "pretcode");
	DIFF(pinfo, "pinfo");
	DIFF(puc, "puc");

	DIFF(info.si_signo, "info.si_signo");
	DIFF(info.si_errno, "info.si_errno");
	DIFF(info.si_code, "info.si_code");
	DIFF(info.si_errno, "info.si_errno");

	// This is probably OK, as long as we clear the structure first
	// i.e., I suspect the differences are just random stack differences,
	// not true changes.  Nevertheless, we need it to be the same both
	// times.
	{
	 int i;
	 for(i=0;i < SI_PAD_SIZE; i++) {
	  if(e->frame.info._sifields._pad[i] != frame->info._sifields._pad[i]) {
	   printk("revirt_replay_rt_sigframe: frame.info._sifields._pad[%d] differs: %x %x\n",
	   		i, 
			frame->info._sifields._pad[i],
	 		e->frame.info._sifields._pad[i]);
	   pause=1;
	  }
	 }
	}

	DIFF(uc.uc_flags, "uc.uc_flags");
	DIFF(uc.uc_link, "uc.uc_link");
	DIFF(uc.uc_stack.ss_sp, "uc.uc_stack.ss_sp");
	DIFF(uc.uc_stack.ss_flags, "uc.uc_stack.ss_flags");
	DIFF(uc.uc_stack.ss_size, "uc.uc_stack.ss_size");
	
	DIFF(uc.uc_mcontext.gs, "gs");
	DIFF(uc.uc_mcontext.fs, "fs");
	DIFF(uc.uc_mcontext.es, "es");
	DIFF(uc.uc_mcontext.ds, "ds");
	DIFF(uc.uc_mcontext.__gsh, "__gsh");
	DIFF(uc.uc_mcontext.__fsh, "__fsh");
	DIFF(uc.uc_mcontext.__esh, "__esh");
	DIFF(uc.uc_mcontext.__dsh, "__dsh");
	DIFF(uc.uc_mcontext.edi, "edi");
	DIFF(uc.uc_mcontext.esi, "esi");
	DIFF(uc.uc_mcontext.ebp, "ebp");
	DIFF(uc.uc_mcontext.esp, "esp");
	DIFF(uc.uc_mcontext.ebx, "ebx");
	DIFF(uc.uc_mcontext.ecx, "ecx");
	DIFF(uc.uc_mcontext.edx, "edx");
	DIFF(uc.uc_mcontext.eax, "eax");
	// We know that trapno changes
	DIFF(uc.uc_mcontext.err, "err");
	DIFF(uc.uc_mcontext.eip, "eip");
	DIFF(uc.uc_mcontext.cs, "cs");
	DIFF(uc.uc_mcontext.__csh, "__csh");
	// eflags checked below
	DIFF(uc.uc_mcontext.esp_at_signal, "esp_at_signal");
	DIFF(uc.uc_mcontext.ss, "ss");
	DIFF(uc.uc_mcontext.__ssh, "__ssh");
	DIFF(uc.uc_mcontext.fpstate, "* fpstate");
	DIFF(uc.uc_mcontext.oldmask, "oldmask");
	DIFF(uc.uc_mcontext.cr2, "cr2");
	
	/*
	 * only the bits masked below are restored by restore_sigcontext.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK) 
		!= (frame->uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK)) {
	    printk("revirt_replay_rt_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK,
	    		e->frame.uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK);
	    pause=1;
	}
	{ int i;
	 for(i=0; i<8; i++) {
	  if(e->frame.retcode[i] != frame->retcode[i]) {
	    printk("revirt_replay_rt_sigframe: retcode[%d] differs: %x %x\n",
	   		i, 
			frame->retcode[i], 
			e->frame.retcode[i]);
	    pause=1;
	  }
	 }
	 for(i=0; i<_NSIG_WORDS; i++) {
	  if(e->frame.uc.uc_sigmask.sig[i] != frame->uc.uc_sigmask.sig[i]) {
	   printk("revirt_replay_rt_sigframe: uc.uc_sigmask.sig[%d] differs: %lx %lx\n",
	   		i, 
			frame->uc.uc_sigmask.sig[i], 
			e->frame.uc.uc_sigmask.sig[i]);
	    pause=1;
	  }
	 }
	}

	DIFF(fpstate.cw, "fpstate.cw");
	DIFF(fpstate.sw, "fpstate.sw");
	DIFF(fpstate.tag, "fpstate.tag");
	DIFF(fpstate.ipoff, "fpstate.ipoff");
	DIFF(fpstate.cssel, "fpstate.cssel");
	DIFF(fpstate.dataoff, "fpstate.dataoff");
	DIFF(fpstate.datasel, "fpstate.datasel");
	DIFF(fpstate.status, "fpstate.status");
	DIFF(fpstate.magic, "fpstate.magic");
	DIFF(fpstate.mxcsr, "fpstate.mxcsr");
	DIFF(fpstate.reserved, "fpstate.reserved");

	{ int i, j;
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++) {
			DIFF(fpstate._st[i].significand[j], 
					"fpstate._sti.significand");
		}
	 	DIFF(fpstate._st[i].exponent, "fpstate.sti[].exponent");
	 }

	 for(i=0; i<6; i++) {
		DIFF(fpstate._fxsr_env[i], "fpstate._fxsr_env[]");
	 }
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++)
			DIFF(fpstate._fxsr_st[i].significand[j],
					"fpstate.fxsr_st[].significand[]");
		DIFF(fpstate._fxsr_st[i].exponent, 
				"fpstate._fxsr_st[].exponent");
		for(j=0; j<3; j++) 
			DIFF(fpstate._fxsr_st[i].padding[j],
				"fpstate._fxsr_st[].padding[]");
	 }
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++)
			DIFF(fpstate._xmm[i].element[j],
				"fpstate._xmm[].element[]");
	 }
	 for(i=0; i<56; i++) {
		DIFF(fpstate.padding[i], "fpstate._padding[]");
	 }
	}
#else
#define DIFF(x, y) \
	if(e->x != frame->x) { 				 \
	  printk("revirt_replay_rt_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->x);	 \
  	  pause = 1;					 \
	}
#endif
	DIFF(sig, "sig");
#undef DIFF

     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}
	/*
	 * Finally, copy the frame
	 */
#ifdef REPLAY_WHOLE_SIGFRAME
	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
#else
	frame->uc.uc_mcontext.trapno = e->trapno;
	return __copy_to_user(user_frame, frame, sizeof(*frame));
#endif
}

/*
 * Move this to another file
 */

//#define REPLAY_DEBUG

extern struct ztrace_rununtil_struct zt_rununtil;

#define REPLAY_EXCLUDE_SOME

#ifndef REPLAY_ALL_SYSCALLS
# define REPLAY_UNLOGGED \
	case __NR_sigprocmask: \
	case __NR_mmap: \
	case __NR_rt_sigreturn: \
	case __NR_getpid: \
	case __NR_munmap: \
	case __NR_lseek: \
	case __NR_sigaltstack: \
	case __NR_write: 	\
	case __NR_getguestpgd: \
	case __NR_switchguest: \
	case __NR_rt_sigsuspend: \
	case __NR__llseek:
#else
# define REPLAY_UNLOGGED
#endif

#define REPLAY_LOGGED \
	case __NR_read: \
	case __NR_socketcall: \
	case __NR_gettimeofday:

/*
 * Called before a syscall either during logging or replay; turns on
 * XUSER and/or DATASQUASH if appropriate
 */
asmlinkage void revirt_presys(struct pt_regs regs) {
	struct revirt_ctrdata cnt;
	struct task_struct * tsk = current;
	if(!tsk->revirt_tsk) {
		printk("revirt_presys: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	/*
	 * Note that revirt_current_syscall should be kept 'up-to-date',
	 * but there's a time between entry.S and here that it's out of sync.
	 */
	tsk->revirt_tsk->current_syscall = revirt_current_syscall;

	/* FIXME -- make this a normal swctr, like interrupts &c */
	tsk->revirt_tsk->nrsyscalls++;
      	copy_branchdata(tsk, &cnt);
#ifdef REPLAY_DEBUG
	if(tsk->revirt_flags & REPLAY_REPLAY) {
	    printk("revirt_presys: pid %d @ 0x%lx calls syscall %d(0x%lx 0x%lx 0x%lx 0x%lx 0x%lx) [%llx %x %x %x]\n",
			tsk->pid, regs.eip, tsk->revirt_tsk->current_syscall, 
			regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi,
			cnt.hw_branch, cnt.hw_intr, cnt.ex_noerror, 
			cnt.ex_error);
	}
#endif
	/*
	 * Do some self-checks to make sure replay is working right
	 */
	if((zt_flags & ZTRACE_BKPT) && (ADJ_BC(cnt) > zt_rununtil.count)) {
		printk("revirt_presys: count %llx, zr wanted %llx\n",
				ADJ_BC(cnt), zt_rununtil.count);
		if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	}
	if((zt_flags & ZTRACE_PERF) && (cnt.hw_branch > 0x100 + zt_perferr)) {
		printk("revirt_presys: ZT_PERF, hw_count %llx\n",
				cnt.hw_branch - zt_perferr );
		if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	}
	/*
	 * Make sure this is a legal syscall for umlinux
	 */
	if((revirt_options.replayflags & REPLAY_OPT_RESTRICT_SYSCALLS)
		&& tsk->umlinux.umlTraced) {
		if(!FD_ISSET(tsk->revirt_tsk->current_syscall, &revirt_options.allowed_syscalls)) {
			printk("revirt_presys: pid %d @%lx makes forbidden syscall %d(0x%lx 0x%lx 0x%lx 0x%lx 0x%lx)\n",
				tsk->pid, regs.eip, tsk->revirt_tsk->current_syscall,
				regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi);
			if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
				ztrace_debug_pause();
			}
		}
	}

	/*
	 * Log or replay appropriate system calls
	 */
	switch(tsk->revirt_tsk->current_syscall) {
#ifdef REPLAY_EXCLUDE_SOME
	    REPLAY_UNLOGGED
	    case __NR_lexecve:
	    	break;
	    default:
#else
	    REPLAY_LOGGED
#endif
	    	if(tsk->revirt_flags & REPLAY_LOG) 
			tsk->revirt_flags |= REPLAY_XUSER;
		else if(tsk->revirt_flags & REPLAY_REPLAY) {
			tsk->revirt_flags |= REPLAY_DATASQUASH;	
			tsk->revirt_flags |= REPLAY_SYSCALL;
		}
		else {
			printk("revirt_presys: Huh? [pause]\n");
			ztrace_debug_pause();
		}
	}
	if(tsk->revirt_flags & REPLAY_REPLAY)
		if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (tsk->revirt_tsk->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
}

asmlinkage void revirt_replay_syscall(struct pt_regs * regs);
void revirt_log_sysret(struct pt_regs *);

asmlinkage void revirt_postsys(struct pt_regs regs) {
	struct pt_regs * pregs = &regs;
	struct task_struct * tsk=current;

	tsk->revirt_flags &= ~(REPLAY_DATASQUASH|REPLAY_XUSER|REPLAY_SYSCALL);
	if(!tsk->revirt_tsk) {
		printk("revirt_postsys: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	/*
	 * This was done in revirt_presys, but in certain cases ([lr]execve)
	 * we have a postsys w/o a presys.  Sync it up.
	 */
	tsk->revirt_tsk->current_syscall = revirt_current_syscall;

	switch(tsk->revirt_tsk->current_syscall) {
#ifdef REPLAY_EXCLUDE_SOME
	    REPLAY_UNLOGGED
	    case __NR_lexecve:
	    case __NR_rexecve:
	    	break;
	    default:
#else
	    REPLAY_LOGGED
#endif
		if(tsk->revirt_flags & REPLAY_LOG)
		    revirt_log_sysret(pregs);
		else if(tsk->revirt_flags & REPLAY_REPLAY) {
		    revirt_replay_syscall(pregs);
		    if(revirt_options.replayflags & REPLAY_OPT_LOGSYS_ON_REPLAY)
			    revirt_log_sysret(pregs);
		}
		else {
			printk("revirt_postsys: Huh? [pause]\n");
			ztrace_debug_pause();
		}
	}
}

void ztrace_tsk_rexec_done(void);

asmlinkage void revirt_replay_syscall(struct pt_regs * regs) {
	int ret = -ENOSYS, r=0;
	unsigned int t;
	struct rlog_header rh;
	struct rlog_put_user_entry * pe;
	struct rlog_copy_user_entry * ce;
	struct rlog_entry_sysret * se;
	struct rlog_clear_user_entry *cle;
	struct task_struct * tsk=current;

#ifdef REPLAY_DEBUG
	printk("revirt_replay_syscall: rexec returned %d.\n", regs->eax);
#endif
	ztrace_tsk_rexec_done();
	if(!tsk->revirt_tsk) {
		printk("revirt_replay_syscall: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}
	tsk->revirt_tsk->nrxuser=0;

#ifdef REPLAY_DEBUG
	printk("revirt_replay_syscall: replaying\n");
#endif
	while(1) {
	    revirt_read_log(&rh);
	    switch(rh.type) {
	        case ENTRY_SENDERIP:
			/*
			 * We need to take the senderip's out of the log,
			 * so we just skip them here.
			 */
			break;
	    	case ENTRY_PUT_USER:
		    pe=(struct rlog_put_user_entry *)revirt_scall_buf;

		    if(pe->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d (%lu):pid %d, expecting %d\n",
				tsk->revirt_tsk->current_syscall, 
				rh.id, tsk->pid, pe->pid);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    /*
		     * Ok, this is wierd.  put_user() is called with 
		     * any size integer, without being told its size, and
		     * is expected to figure it out with compiler techniques.
		     * It does so using the sizeof(*(ptr)) compiler operator.
		     * As a result, __put_user_size is always "called" with
		     * a constant: 1, 2, or 4 (for this architecture).  This
		     * is replaced with switch(2) { } (or 1 or 4), and the 
		     * optimizer optimizes away all other cases, including
		     * the default.
		     *
		     * If ever anyone calls put_user with a pointer not of
		     * size 1,2, or 4, the default: case is instantiated, 
		     * which calls __put_user_bad.  Since there is no
		     * __put_user_bad, this causes the link to fail.
		     *
		     * But it will do the same thing if we call __put_user_size
		     * with a variable, such as pe->ssize, since the compiler
		     * doesn't know what values pe->ssize may take.  So
		     * we call __put_user_size with constants.
		     */

		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (tsk->revirt_tsk->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
#ifdef REPLAY_DEBUG
		    printk("revirt_replay_user: pid %d(%u)__put_user_size(%lx, %lx, %d)",
		    		tsk->pid, rh.id, pe->xx, pe->pptr, pe->ssize);
#endif
		    switch(pe->ssize) {
		      case 1:
		    	__put_user_size(pe->xx, pe->pptr, 1, r);
			break;
		      case 2:
		    	__put_user_size(pe->xx, pe->pptr, 2, r);
			break;
		      case 4:
		    	__put_user_size(pe->xx, pe->pptr, 4, r);
			break;
		      case 8:
		    	__put_user_size(pe->xx, pe->pptr, 8, r);
			break;
		      default:
		      	printk("revirt_replay_user: bad put_user size: %d [pause]\n", 
					pe->ssize);
			ztrace_debug_pause();
		    }
		    tsk->revirt_tsk->nrxuser++;
		    if(r!=pe->rretval) {
		    	printk("revirt_replay_syscall %d:putuser ret %d,expecting %d\n", 
				tsk->revirt_tsk->current_syscall, r, pe->rretval);
		    }
		    break;


		case ENTRY_COPY_USER:
		  do {
		    char * d;
		    int i, l;

		    ce=(struct rlog_copy_user_entry *)revirt_scall_buf;
		    if(ce->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d: pid %d, expecting %d(%lu)\n",
				tsk->revirt_tsk->current_syscall, 
				tsk->pid, 
				ce->pid, rh.id);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    if(ce->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d(%lu) [pause]\n", 
				tsk->revirt_tsk->current_syscall, ce->syscall, rh.id);
			ztrace_debug_pause();
		    }

		    l=ce->ssize;
		    t=ce->tto;
		    i=0;
		    if((d=kmalloc(l, GFP_KERNEL))==NULL) {
		    	printk("revirt_replay copy_user: cannot allocate buffer.\n");
			BUG();
		    }
		    memcpy(d, revirt_scall_buf+sizeof(*ce), l);
		    /*
		     *  tsk->revirt_tsk->nrxuser is incremented at the beginning
		     * of a copy_user, but at the end of this, as it needs
		     * to be to interrupt between xusers.  But when 
		     * interrupting an xuser, the logged nrxuser is one
		     * more, so we compensate.
		     */
		    while(zt_flags & ZTRACE_XUSER_STOP && 
		    	   (zt_xstop.nrxuser-(zt_xstop.edi!=0))
			   			==tsk->revirt_tsk->nrxuser) {
			int delta;
			printk("copy_user: edi=0x%x\n", zt_xstop.edi);
			if(!zt_xstop.edi) {
				ztrace_tsk_xuser_schedule();
				continue;
			}
			delta=zt_xstop.edi - t;
			if(delta > l) {
			    printk("rp_copyuser: delta %d, l %d\n", 
			    		delta, l);
			    printk("  pausing...\n");
			    ztrace_debug_pause();
			    BUG();
			}
			r=delta;
			printk("revirt_replay_syscall: pid %d __copy_user_atom(0x%x,0x%lx,%d)\n",
					tsk->pid, t, (long unsigned int)d+i, r);
			__copy_user_atom(t,d+i,r);
			if(r) {
		    	 printk("rp_syscall:__copy_user_atom failed at %d\n",
					r);
			 printk("   syscall:%d nrsyscalls:%d\n", ce->syscall,
					ce->nrsyscalls);
			 //BUG();
		    	}
			t+=delta;
			i+=delta;
			l-=delta;
			ztrace_tsk_xuser_schedule();
		    }
		    r=l;
#ifdef REPLAY_DEBUG
		    printk("revirt_replay_syscall: pid %d(%lu) __copy_user_atom(0x%lx,0x%lx,%d)\n",
					tsk->pid, rh.id, 
					t, (long unsigned int)d+i, r);
#endif
		    __copy_user_atom(t, d+i, r);
		    tsk->revirt_tsk->nrxuser++;
		    kfree(d);
		    if(r) {
		    	printk("revirt_replay_syscall:__copy_user_atom failed at %d\n",
					r);
			printk("   syscall:%d(%lu) nrsyscalls:%d\n", ce->syscall,
					rh.id, ce->nrsyscalls);
			//BUG();
		    }
		  } while(0);
		    break;
		case ENTRY_CLEAR_USER:
		    cle=(struct rlog_clear_user_entry *)revirt_scall_buf;
		    if(cle->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d: pid %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, tsk->pid, 
				cle->pid);
			ztrace_debug_pause();
		    }
		    if(cle->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, cle->syscall);
			ztrace_debug_pause();
		    }
		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (tsk->revirt_tsk->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
		    r=__clear_user((void *)cle->to, cle->n);
		    tsk->revirt_tsk->nrxuser++;
		    if(r) {
		    	printk("revirt_replay_syscall:__clear_user failed at %d\n",
					r);
			printk("   syscall:%d \n", cle->syscall);
			//BUG();
		    }
		    break;
		case ENTRY_SYSRET:
		    se=(struct rlog_entry_sysret *)revirt_scall_buf;
		    if(se->pid != tsk->pid) {
		    	printk("revirt_replay_syscall sys %d id %lu: pid %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, 
				rh.id,
				tsk->pid, 
				se->pid);
			ztrace_debug_pause();
		    }
		    if(se->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d(%lu) [pause]\n", 
				tsk->revirt_tsk->current_syscall, se->syscall,
				rh.id);
			ztrace_debug_pause();
		    }
		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (tsk->revirt_tsk->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
		    /*
		     *  Check for syscalls with changed values....
		     */
		   if((se->syscall != __NR_read) 
		   	&& (se->syscall != __NR_socketcall)) 
		{
		    int pause=0;
#ifdef REPLAY_WHOLE_PTREGS
#define DIFF(x,y) \
	if(se->regs.x != regs->x) { 			 \
	  printk("revirt_replay_syscall: rex'd %s differs: %x %x\n", \
	  	y, (unsigned int)regs->x, (unsigned int)se->regs.x);	  \
  	  pause = 1;					 \
	}
		    DIFF(ebx, "ebx");
		    DIFF(ecx, "ebx");
		    DIFF(edx, "ebx");
		    DIFF(esi, "esi");
		    DIFF(edi, "edi");
		    DIFF(eip, "eip");
		    DIFF(esp, "esp");
		    DIFF(ebp, "ebp");
		    /*
		     * Only the bits from REVIRT_EFLAGS_MASK are actually
		     * accessible and affect flow in ring 3,
		     * so we only check those.
		     */
		    if((se->regs.eflags & REVIRT_EFLAGS_MASK)
			!= (regs->eflags & REVIRT_EFLAGS_MASK)) {
			printk("revirt_replay_syscall: rex'd eflags differs: %lx %lx\n",
				regs->eflags & REVIRT_EFLAGS_MASK,
				se->regs.eflags & REVIRT_EFLAGS_MASK);
			pause=1;
		   }
#undef DIFF
#endif
		    if(se->regs.eax != regs->eax) {
		    	printk("revirt_replay_syscall: rex'd eax differs: %lx %lx\n",
					regs->eax, se->regs.eax);
			pause=1;
		    }

#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
		    check_branchcount(&se->ctr, regs, rh);
#endif
		    if(pause){
		    	printk("revirt_replay_syscall: syscall %d(%lu) should probably be logged [pause]\n",
				se->syscall, rh.id);
			ztrace_debug_pause();
		    }
		   }
		    regs->eax = se->regs.eax;

		    ret = regs->eax;
#ifdef REPLAY_DEBUG
		    printk("revirt_replay_sycall: pid %d ret %d @ 0x%lx\n", 
		    		tsk->pid, ret, regs->eip);
# ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#  ifdef REPLAY_RECORD_ALL_PFCTRS
		    printk("revirt_replay_syscall: logged count: [%llx %x %x %x]\n",
		    		se->ctr.hw_branch,
		    		se->ctr.hw_intr,
		    		se->ctr.ex_noerror,
		    		se->ctr.ex_error);
#  else
		    printk("revirt_replay_syscall: logged count: [%llx]\n",
		    		se->ctr.hw_branch);
#  endif
# endif
#endif
		    tsk->revirt_tsk->nrxuser++;
		    goto out;
		default:
		    printk("Bad syscall entry: %d (%lu) [pause]\n", 
				    rh.type, rh.id);
		    if(rh.type == ENTRY_RT_SIGFRAME) {
		    	struct rlog_entry_rt_sigframe * sys;
		    	sys=(struct rlog_entry_rt_sigframe *)revirt_scall_buf;
#ifdef REPLAY_WHOLE_SIGFRAME
			printk(" sig: %d eip:%lx\n", sys->frame.sig,
					sys->frame.uc.uc_mcontext.eip);
#else
			printk(" sig: %d\n", sys->sig);
#endif
		    }
		    ztrace_debug_pause();
	    }
	}
    out:
	return;
}

asmlinkage int revirt_writev_log(int, const struct iovec *, int);

unsigned long revirt_copyuser(unsigned long to, unsigned long from, 
	unsigned long size) {
    struct rlog_copy_user_entry le;
    struct task_struct * tsk = current;

    if(!tsk->revirt_tsk) {
	    printk("revirt_copyuser: pid %d has no revirt_tsk!\n",
			    tsk->pid);
	    return -EFAULT;
    }

    /*
     * There are times when the same function may write either to userland
     * or to kernel; e.g., low-level file reads.  We need to make sure to
     * log & replay only those which go into userland.
     *
     * FIXME: we need to use a #define for this to make it portable
     */
    if(to > 0xc0000000) {
    	int flag;
	flag = tsk->revirt_flags & REPLAY_XUSER;
	tsk->revirt_flags &= ~REPLAY_XUSER;
	__copy_user_atom(to, from, size);
	tsk->revirt_flags |= flag;
	return size;
    }

    if(tsk->revirt_flags & REPLAY_LOG) {
#if 0
	if( 0 /*tsk->revirt_flags & REPLAY_DATAPASS */) {
	    /* log size, no data */
	    le.type=ENTRY_COPY_USER_NODATA;
	    le.tto=to;
	    le.ssize=size;
	    le.pid=tsk->pid;
	    le.nrsyscalls=tsk->revirt_tsk->nrsyscalls;
	    le.syscall=revirt_current_syscall; //tsk->revirt_tsk->current_syscall;
	    revirt_write_log((char *)&le, le.type, sizeof(le));
	} else 
#endif
	if(tsk->revirt_flags & REPLAY_XUSER) {
	    /* log both metadata and data */
            struct iovec v[2]; 
            le.type = ENTRY_COPY_USER;
            le.tto = to;
            le.ssize = size;
            le.pid = tsk->pid;
            le.nrsyscalls = tsk->revirt_tsk->nrsyscalls;
            le.syscall = revirt_current_syscall; //tsk->revirt_tsk->current_syscall;
            v[0].iov_base = &le;
            v[0].iov_len = sizeof(le);
            v[1].iov_base = (void *)from;
            v[1].iov_len = size;
            revirt_writev_log(le.type, v, 2);
	}
	__copy_user_atom(to, from, size);
	return size;
    }
    if(tsk->revirt_flags & REPLAY_REPLAY) {
	if(tsk->revirt_flags & REPLAY_DATASQUASH) {
	    if(to < 0xc0000000) {
#ifdef REPLAY_DEBUG
		printk("squashing copy_user(0x%lx,0x%lx,%ld)\n",
			to, from, size);
#endif
		return 0;
	    } else {
	    	printk("WARNING: calling copy_user with kernel destination\n");
		printk("   copy_user(0x%lx,0x%lx,%ld)\n", to, from, size);
		ztrace_debug_pause();
		BUG();
	    }
	}
	else {
		__copy_user_atom(to, from, size);
		return size;
	}
    }
    printk("revirt_replay_copy_user: huh?\n");
    ztrace_debug_pause();
    return -1;
}

void revirt_log_exit(void);
void revirt_read_close(void);

void revirt_exit(struct task_struct * tsk) {
        if (tsk->revirt_flags&REPLAY_LOG)  {
                revirt_log_exit();
        } else if(tsk->revirt_flags & REPLAY_REPLAY) {
	    if(tsk->revirt_tsk && tsk->revirt_tsk->sched) {
                if(zt_flags & ZTRACE_WAITEXIT) {
                        printk("revirt_exit: pid %d waking up scheduler\n",
                                tsk->pid);
                        /* FIXME: do we need to worry about locks here? */
                        tsk->revirt_tsk->sched->revirt_flags &= ~REPLAY_WAIT;
                        zt_flags &= ~ZTRACE_WAITEXIT;
                } else {
                        printk("revirt_exit: pid %d exiting unexpectedly?\n",
                                tsk->pid);
			tsk->revirt_flags=0;
			kfree(tsk->revirt_tsk);
			tsk->revirt_tsk=NULL;
                        BUG();
                }
	    } else {
		    printk("revirt_exit: tsk %d exiting, no scheduler\n",
				    tsk->pid);
		    tsk->revirt_flags=0;
		    kfree(tsk->revirt_tsk);
		    tsk->revirt_tsk=NULL;
		    BUG();
	    }
	    /*
	     * FIXME -- assumes only one proc for now, just for UML.  If we
	     * go back to host-based replay, we need to keep the file open.
	     */
	    revirt_read_close();
        }
	if(tsk->revirt_tsk) {
		kfree(tsk->revirt_tsk);
		tsk->revirt_tsk=NULL;
	} 
}
