/*
 * ztrace.c - Interface for zR user-level schedule replayer
 *
 * Copyright (C) 2002-2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/user.h>

#include <asm/perfctr.h>
#include <linux/slab.h>

#include <asm/hardirq.h> // for in_interrupt()

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>

/* for ZTRACE_SMI */
#include <asm/io.h>

struct ztrace_xuser_stop zt_xstop;
/*
 * FIXME -- take out syscalls 5 and 6 (open and close) 
 * Probably also some more
 */
struct revirt_options_st revirt_options = {
	.logflags= REPLAY_DEFAULT_LOGFLAGS,
	.replayflags=REPLAY_DEFAULT_REPLAYFLAGS,
	.xreplay = {
		.packetrate=5, 
		.packetsize=1000 },
	.network_fd=UML_NETWORK_FD, 	/* network_fd */
	.unlogged_fds={}, 		/* unlogged_fds */
	.allowed_syscalls = {
	 .fds_bits={
		/* FIXME -- we need to take out open, close, ptrace,
			setitimer, rt_sigaction, sigaltstack,  */
	  /* 1 exit 
	     3 read
	     4 write
	     5 open
	     6 close
	     19 lseek
	     20 getpid
	     26 ptrace */
	  0x418007a, 
	  /* 55  fcntl */
	  0x800000, 
	  /* 78 gettimeofday
	     90 mmap
	     91 munmap */
	  0xc004000, 
	  /* 104 setitimer
	     126 sigprocmask */
	  0x40000100, 
	  /* 140 _llseek*/
	  0x1000, 
	  /* 173 rt_sigreturn
	     174 rt_sigaction
	     179 rt_sigsuspend
	     186 sigaltstack */
	  0x4086000, 
	  0x0,
	  /* 233 ztrace  FIXME -- need to limit this functionality */ 
	  /* 235 getguestpgd 
	     236 switchguest */
	  0x1a00 } } ,
	 .logbuf= {
		 .size=(1 << 21),	/* This must be a power of 2...*/
		 .mask=((1 << 21)-1) }, /* for this to work */
 };



int zt_flags=0;
struct ztrace_rununtil_struct zt_rununtil;
int zt_yield=0;
int zt_wchan=0;
struct task_struct * zt_tsk=NULL;
s64 zt_perferr=0;

spinlock_t ztrace_yieldto_lock = SPIN_LOCK_UNLOCKED;

//#define ZTRACE_DEBUG

/****************************************************************
 *
 * Call-backs from important kernel places...
 *
 ****************************************************************/
/* 
 * Called from arch/i386/kernel/traps.c:do_debug(), for single-stepping and 
 * hardware breakpoints.
 */
void ztrace_tsk_debug(void) {
	struct task_struct *tsk=current;

	if(!tsk->revirt_tsk)
		return;

	if((tsk->revirt_flags & REPLAY_SSTEP) 
		|| (zt_flags & (ZTRACE_BKPT|ZTRACE_PERF))) {
    	    zt_yield=ZTRACE_SSTEP;
    	    tsk->revirt_flags &= ~REPLAY_SSTEP;
	    zreg(tsk,eflags) &= ~TRAP_FLAG;
    	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} else {
	    printk("ztrace_task_sstep: REPLAY_SSTEP and ZTRACE_BKPT unset!\n");
	    printk("   pausing...\n");
	    ztrace_debug_pause();
	}
}

/*
 * Called from arch/i386/kernel/apic.c:smp_perfcntr_overflow()
 * when we get a performance counter overflow interrupt
 */
void ztrace_tsk_perf(void) {
    struct task_struct *tsk=current;
    unsigned long flags, redir=0;

    save_flags(flags);
    cli();

    /*
     * Clear LVTPC mask, and overflow flag in counter control MSR (P4)
     */
    perfctr_clear_ovf();

    /*
     * Sanity checks
     */
    if(!(zt_flags & ZTRACE_PERF)) {
	restore_flags(flags);
    	printk("ztrace_task_perf: ZTRACE_PERF unset! [pause]\n");
	ztrace_debug_pause();
	goto out;
    }

    if(!zt_tsk) {
	restore_flags(flags);
    	printk("ztrace_task_perf: got ovf, but zt_tsk == NULL!\n");
	ztrace_debug_pause();
	goto out;
    }

    if(!zt_tsk->revirt_tsk) {
	    restore_flags(flags);
	    printk("zt_tsk_perf: zt_tsk (pid %d) has no revirt_tsk\n",
			    tsk->pid);
	    goto out;
	    return;
    }

    /*
     * This needs to be atomic, for if we get swapped out, we may never
     * get run again.  need_resched should be safe before this section,
     * because the scheduler can't run, but we'll protect it here just to
     * be safe.
     */
    zt_yield=ZTRACE_PERF;

    if(zt_tsk->revirt_tsk->sched)
    	zt_tsk->revirt_tsk->sched->revirt_flags &= ~REPLAY_WAIT;
    zt_tsk->revirt_flags |= REPLAY_WAIT;

    if(tsk == zt_tsk)
    	tsk->need_resched=1;
    else
    	redir=zt_tsk->pid;
    zt_tsk=NULL;

    restore_flags(flags);
  out:
    return;
}

/*
 * ztrace_tsk_enters_syscall -- called from entry.S just after entering
 * the kernel.
 * 
 * Stop if:
 *  - we've been asked to stop at the next system call entry (SYSENTER_STOP)
 *  or
 *  - We're doing breakpoint replay and the return eip equals the stop eip.
 *
 *  NOTE that nrsyscalls and current_syscall are now set in revirt_presys,
 *   which is AFTER ztrace_tsk_enters_syscall() is called; they can't be
 *   trusted at this point.
 */
asmlinkage void ztrace_tsk_enters_syscall(void) {
	struct task_struct *tsk= current;

	if(!tsk->revirt_tsk)
		return;
	if(zt_flags & ZTRACE_SYSENTER_STOP) {
#ifdef ZTRACE_DEBUG
		printk("ztrace: pid %d yielding due to SYSENTER_STOP\n",
				tsk->pid);
#endif
		//zt_flags &= ~ZTRACE_SYSENTER_STOP;
		zt_yield=ZTRACE_SYSCALL_ENTER;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if ((zt_flags & ZTRACE_BKPT_IN_KERNEL) 
			&& (zreg(tsk,eip)==zt_rununtil.eip)) {
#ifdef ZTRACE_DEBUG
		printk("ztrace: pid %d yielding in syscall" \
				" due to BKPT_IN_KERNEL\n",
				tsk->pid);
#endif
		zt_yield=ZTRACE_SYSCALL_ENTER;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	}
}

/*
 * Called from entry.S, after returning from the syscall, but before
 * signal delivery / rescheduling stuff.
 *
 * I think ZTRACE_BLOCK wants to stop here because it *expected* to be blocked
 * in the middle of the last system call; for some reason it wasn't.  Same
 * thing with ztrace_tsk_rexec_done(), which happens before this on the same
 * codepath.  Not sure why I have it both places... =(
 */
asmlinkage void ztrace_tsk_exits_syscall(void) {
	struct task_struct *tsk= current;

	if(!tsk->revirt_tsk)
		return;

	if(zt_flags & ZTRACE_XUSER_STOP) {
		printk("ztrace: exiting syscall, ZTRACE_XUSER_STOP set!\n");
		printk("ztrace: zt_xstop.nrexuser=%lx, tsk->revirt_tsk->nrxuser=%lx\n",
				zt_xstop.nrxuser, tsk->revirt_tsk->nrxuser);
		BUG();
	}
	
	if(zt_flags & ZTRACE_SYSEXIT_STOP) {
#ifdef ZTRACE_DEBUG
	    printk("ztrace: pid %d yielding at sysexit due to SYSEXIT_STOP\n",
	    		tsk->pid);
#endif
	    //tsk->revirt_flags |= REPLAY_AT_EXIT;
	    //zt_flags &= ~ZTRACE_SYSEXIT_STOP;
	    zt_yield=ZTRACE_SYSCALL_EXIT;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} 
	else if(tsk->revirt_flags & REPLAY_SSTEP ) {
#ifdef ZTRACE_DEBUG
	    printk("ztrace: pid %d yielding at sysexit due to REPLAY_SSTEP\n",
	    		tsk->pid);
#endif
	    tsk->revirt_flags &= ~REPLAY_SSTEP;
	    zt_yield=ZTRACE_SYSCALL_EXIT;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(zt_flags & ZTRACE_BKPT && zreg(tsk,eip)==zt_rununtil.eip) {
#ifdef ZTRACE_DEBUG
		printk("ztrace: pid %d yielding after syscall due to BKPT\n",
			tsk->pid);
#endif
	    zt_yield=ZTRACE_SYSCALL_EXIT;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if (zt_flags & ZTRACE_BLOCK_STOP 
			&& zreg(tsk,eip)==zt_rununtil.eip) {
#ifdef ZTRACE_DEBUG
	    printk("ztrace: pid %d yielding after syscall due to BLOCK\n",
			tsk->pid);
#endif
	    zt_yield=ZTRACE_SYSCALL_EXIT;
	    zt_wchan=0;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	}
}

/*
 * Called from revirt/replay.c:revirt_replay_sycall().
 * This is called during replay, after the system call is re-executed,
 * before the xusers &c are replayed.
 *
 * See note for ztrace_tsk_syscall_exit for ZTRACE_BLOCK.
 */
void ztrace_tsk_rexec_done(void) {
	struct task_struct * tsk = current;
	if(!tsk->revirt_tsk)
		return;
	if(zt_flags  & ZTRACE_BLOCK_STOP  
                        && zreg(tsk,eip)==zt_rununtil.eip) {
            printk("ztrace: pid %d yielding after rexec due to BLOCK eip:0x%lx\n",
                        tsk->pid, zreg(tsk,eip));
            zt_yield=ZTRACE_XUSER_STOP;
            zt_wchan=0;
            ztrace_yield_to(tsk->revirt_tsk->sched);
        }
}

/*
 * Called from revirt/replay.c:revirt_replay_sycall()
 * From varioius places: after / during put_user and copyuser replays.
 */
void ztrace_tsk_xuser_schedule(void) {
	struct task_struct * tsk = current;
	if(!tsk->revirt_tsk)
		return;
	/*
	 * Hmm... to be consistent, we should check for zt_flags here,
	 * not in replay.c....
	 */
	//zt_flags &= ~ZTRACE_XUSER_STOP;
	printk("ztrace: yielding due to xuser stop\n");
	zt_yield = ZTRACE_XUSER_STOP;
	ztrace_yield_to(tsk->revirt_tsk->sched);
}

/*
 * Called from arch/i386/kernel/signal.c:do_signal, after handle_signal.
 */
void ztrace_tsk_gets_signal(void) {
	struct task_struct * tsk = current;
	struct revirt_ctrdata ctr;

	/*
	 * After delivering a signal, we need to let it be handled so that
	 * zr has accurate information about what the process is going to do
	 * next.
	 *
	 * If we're single-stepping and receive a segv, the trap doesn't 
	 * happen, but the flag gets cleared; and if we're
	 * at a breakpoint and we've just received a SEGV, instruction
	 * breakpoints are temporarily disabled; so we need to check for
	 * those conditions and yield anyway.
	 */
	if(!tsk->revirt_tsk)
		return;
#ifdef ZTRACE_DEBUG
	copy_branchdata(tsk, &ctr);
	printk("ztrace_tsk_gets_signal: @ 0x%x %llx [%llx]\n", 
			zreg(tsk,eip), ADJ_BC_R(ctr), ctr.hw_branch);
#endif
	if(zt_flags & ZTRACE_SIGNAL_STEP) {
            zt_yield=ZTRACE_SIGNALQUEUE;
	    //zt_flags= 0;
	    if(tsk->revirt_tsk->sched)
	     	ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(tsk->revirt_flags & REPLAY_SSTEP) {
            zt_yield=ZTRACE_SIGNALQUEUE;
            tsk->revirt_flags &= ~REPLAY_SSTEP;
            ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(zt_flags & ZTRACE_BKPT
                        && zreg(tsk,eip)==zt_rununtil.eip) {
		/* 
		 * We need to be able to deliver a signal before the process
		 * continues executing; but at this point, do_signal has
		 * already been called.  So we clear the RF flag so that
		 * the breakpoint succeeds.
		 */
		if(zreg(tsk,eflags) & X86_EFLAGS_RF) {
			zreg(tsk,eflags) &= ~X86_EFLAGS_RF;
#ifdef ZTRACE_DEBUG
			printk("ztrace_tsk_gets_signal: clearing RF\n");
#endif
		}
        }
}

/********************************************************************
 *
 * Functions called from sys_ztrace
 *
 ********************************************************************/
int ztrace_set_opt(struct task_struct * tsk, unsigned long addr, 
			unsigned long data) {
	u32 tmask;
	int err=0;
	fd_set fd_tmp;

	printk("ztrace_set_opt: %lx %lx\n", addr, data);
	if(addr & REPLAY_OPT_VALUE) {
	    printk("ztrace_set_opt: using value\n");
	    switch(addr) {
	    	case REPLAY_OPT_PACKETRATE:
		    if(data <= REPLAY_PBUF_MAX_SIZE)	
			revirt_options.xreplay.packetrate=data;
		    else
		    	err=-EINVAL;
		    break;
		case REPLAY_OPT_PACKETSIZE:
		    revirt_options.xreplay.packetsize=data;
		    break;
		case REPLAY_OPT_NETWORKFD:
		    /* FIXME -- check to make sure we're not ptraced 
		     * (i.e., only lilo can switch these, not guest kernel) */
		    revirt_options.network_fd=data;
		    break;
		case REPLAY_OPT_UNLOGGEDFDS:
		    if(copy_from_user(&fd_tmp, (fd_set *)data, sizeof(fd_set))) 
			err=-EFAULT;
		    else
		    	revirt_options.unlogged_fds=fd_tmp;
		    break;
		case REPLAY_OPT_LOGBUFSIZE:
		    /*
		     * revirt_alloc_logbuf will resize this if necessary
		     */
		    revirt_options.logbuf.size=data;
		    revirt_options.logbuf.mask=data-1;
		    printk("logbufsize: size %d mask %x\n", 
				    revirt_options.logbuf.size,
				    revirt_options.logbuf.mask);
		    revirt_alloc_log_buf(1);
		    err=0;
		    break;
		default:
		    err=-EINVAL;
	    }
	} else {
	    printk("ztrace_set_opt: using optmask\n");
	    /*
	     * FIXME -- we need to make this more clean than looking in 
	     * the lower bits, because OPT_VALUE uses those...
	     */
	    if(addr & REPLAY_OPT_LOGMASK) {
		tmask = revirt_options.logflags;
		tmask &= ~(REPLAY_OPT_LOGMASK & addr);
		tmask |= (REPLAY_OPT_LOGMASK & data);
		printk("ztrace: setting logflags to %x\n", tmask);
		revirt_options.logflags = tmask;
	    }
	    if(addr & REPLAY_OPT_REPLAYMASK) {
		tmask = revirt_options.replayflags;
		tmask &= ~(REPLAY_OPT_REPLAYMASK & addr);
		tmask |= (REPLAY_OPT_REPLAYMASK & data);
		printk("ztrace: setting replayflags to %x\n", tmask);
		revirt_options.replayflags = tmask;
	    }
	}
	return err;
}

/************************
 * Data control functions
 ************************/
int ztrace_get_reg(struct task_struct *tsk,unsigned long addr,
		   unsigned long* val)
{
	long * p;
	unsigned long lval;

	if(addr > EAX)
		return -EINVAL;

	p=&(zreg(tsk,ebx));
	lval = *(p+addr);

	put_user(lval, val);
	return 0;
}

int ztrace_put_reg(struct task_struct *tsk,unsigned long addr,
			unsigned long val) 
{
	long * p;

	if(addr < FRAME_SIZE )  {
		/*
		 * Well, Linux changed the setup of the ptregs struct,
		 * but couldn't change asm/ptrace.h for backwards
		 * compatability; so they map the old junk onto the new junk.
		 * This is basically a copy of 
		 * arch/i386/kernel/ptrace.c:getreg.
		 */
		switch(addr) {
			case FS:
				tsk->thread.fs = val;
				break;
			case GS:
				tsk->thread.gs = val;
				break;
			case DS:	
			case ES:	
			case CS:	
			case SS:	
				val &= 0xffff;
				/* FALLTHRU */
			default:
				/*
				 * Don't let zr change trap flag, etc.
				 */
				if(addr==EFL) {
				    val = (REVIRT_EFLAGS_MASK & val)
					| ((~REVIRT_EFLAGS_MASK) 
					       & zreg(tsk,eflags));
				}
				if(addr > GS)
					addr -= 2;
				p=&(zreg(tsk,ebx));
				*(p+addr) = val;
		}
		return 0;
	} else 
		return -EINVAL;	
}

static int ztrace_getstate(struct task_struct *tsk,
			   struct ztrace_state_struct *state)
{
	/*
	 * sys_ztrace calls this with a tmp variable, and does copy_user
	 */
	int i;
	state->regs = *(((struct pt_regs *) tsk->thread.esp0)-1);
	state->revirt_flags = tsk->revirt_flags;
	/*
	 * Copy next 16 bytes of code, and previous 16 bytes of code
	 *
	 * Put 'canaries' in so we can see if the reads failed (probably
	 * due to invalid addresses).
	 */
	for (i=0;i<4;i++) {
		if(ztrace_read_long(tsk,state->regs.eip+i*4,
				 (unsigned long*) (state->code+i*4))) {
			*(unsigned long *)(state->code+i*4) = 0xdeadbeef;
		}
	}
	for (i=0;i<4;i++) {
		if(ztrace_read_long(tsk,state->regs.eip-16+i*4,
				 (unsigned long*) (state->pcode+i*4))) {
			*(unsigned long *)(state->pcode+i*4) = 0xdeadbeef;
		}
	}

	for(i=0; i<8; i++)
		state->debugreg[i]=tsk->thread.debugreg[i];
	state->cr2 = tsk->thread.cr2;
	state->trap_no = tsk->thread.trap_no;
	state->error_code = tsk->thread.error_code;
	copy_branchdata(tsk, &state->ctr);
	state->wchan = revirt_getwchan(tsk);
	state->zt_flags = zt_flags;
	state->zt_rununtil = zt_rununtil;
	state->zt_perferr = zt_perferr;
	return 0;
}

static int ztrace_peek(struct task_struct * tsk, unsigned long addr, 
				unsigned long * p) {
	unsigned long tmp;
	int copied;

	copied = access_process_vm(tsk, addr, &tmp, sizeof(tmp), 0);
	if(copied != sizeof(tmp))
		return -EIO;
	return put_user(tmp, p);
}

/*********************************************
 * Functions which block & let tracing pid run
 *********************************************/

int send_signal(int, struct siginfo *, struct sigpending *);

static int ztrace_signalqueue(struct task_struct * tsk, char * data) {
	siginfo_t si;
	struct rlog_entry_signal le;
	char * p;
	int i, ret;

	p=data;
	if(copy_from_user(&le, p, sizeof(le))) {
		return -EFAULT;
	}
	p += sizeof(le);

	if(tsk->sig == NULL) {
		return -EFAULT;
	}

	spin_lock_irq(&tsk->sigmask_lock);

	for(i=0; i<_NSIG; i++) {
	    if(sigismember(&le.dfl, i+1) && 
	    		(tsk->sig->action[i].sa.sa_handler == SIG_IGN)) {
		printk("ztrace: pid %d sig %d SIG_IGN to SIG_DFL\n", 
				current->pid, i+1);
		tsk->sig->action[i].sa.sa_handler = SIG_DFL;
	    }
	}
#ifdef ZTRACE_DEBUG
	printk("ztrace: changing pid %d sigmask 0x%x %x to 0x%x %x\n",
		tsk->pid, tsk->blocked.sig[0], tsk->blocked.sig[1],
		le.blocked.sig[0], le.blocked.sig[1]);
#endif
	tsk->blocked = le.blocked;

	flush_signals(tsk);

#ifdef REPLAY_WHOLE_SIGINFO
	if(copy_from_user(&si, p, sizeof(si))) {
			return -EFAULT;
	}
#else
	memset(&si, 0, sizeof(si));
	si.si_signo = le.si_signo;
	si.si_errno = le.si_errno;
	si.si_code = le.si_code;
	si._sifields._pad[0] = le._pad[0];
	si._sifields._pad[1] = le._pad[1];
#endif
#ifdef ZTRACE_DEBUG
	printk("ztrace: placing signal %d in queue\n", si.si_signo);
#endif
	if((ret=send_signal(si.si_signo, &si, &tsk->pending)))
		return ret;
	if(!sigismember(&tsk->blocked, si.si_signo) 
				&& tsk->state & TASK_INTERRUPTIBLE) 
		wake_up_process(tsk);

	recalc_sigpending(tsk);

	spin_unlock_irq(&tsk->sigmask_lock);

       /*
        * We need to step through the signal-delivery process so that 
	* zr gets the right values when it does a GETSTATE.
	*/
       zt_flags = ZTRACE_SIGNAL_STEP;
       ztrace_yield_to(tsk);
       zt_flags=0;
       return 0;
}

static int ztrace_sstep(struct task_struct *tsk,int waitfl)
{
	s64 count;
	struct revirt_ctrdata ctr;
	if (tsk==current) {
		printk("ztrace: single stepping self!\n");
		return -1;
	}

	/* Take count for a sanity check... */
	copy_branchdata(tsk, &ctr);
	count = ADJ_BC(ctr);

	/* Set SSTEP flag & trap flag, and yield */
	tsk->revirt_flags |= REPLAY_SSTEP;
	zreg(tsk,eflags) |= TRAP_FLAG;

	ztrace_yield_to(tsk);

	zreg(tsk,eflags) &= ~TRAP_FLAG;
	tsk->revirt_flags &= ~REPLAY_SSTEP;

	/* Sanity check */
	copy_branchdata(tsk, &ctr);
	if(ADJ_BC(ctr) - count > 2) {
		printk("ztrace: sstep more than one branch? c: %llx [pause]\n",
			count);
		ztrace_debug_pause();
	}
	return zt_yield;
}

static int ztrace_perf(struct task_struct * tsk, s64 * branches_p,
			struct ztrace_breturn * bret) {
    struct ztrace_breturn br;
    s64 perfctr_save, branches;
    struct revirt_tsk_s *rtsk=tsk->revirt_tsk;
    int counter=revirt_pfctr.overflow.counter_index;

    if(!tsk->revirt_tsk) {
	    printk("ztrace_perf: pid %d has no revirt_tsk!\n",
			    tsk->pid);
	    return -EINVAL;
    }

    /*
     * zR stores the "total branches taken since last record" in the
     * hardware counters; we need to maintain this illusion.
     */
    perfctr_save=rtsk->perfctr[counter].ctr;

    /*
     * They have a 64-bit put_user, but no 64-bit get_user! 
     */
    if(copy_from_user(&branches, branches_p, sizeof(branches))) {
    	return -EFAULT;
    }

    /*
     * zt_perferr should only be read by the replayed process, which should
     * be blocked right now; no need to put a lock around these two.
     */
    rtsk->perfctr[counter].ctr = -branches;
 
    zt_perferr = branches + perfctr_save;

    perfcntr_enable_overflow();

    // FIXME -- is this necessary?
    tsk->revirt_flags &= ~REPLAY_SSTEP;
    zreg(tsk,eflags) &= ~TRAP_FLAG;

    /*
     * Set the flags, and hand-off to the traced process
     */
    zt_flags = ZTRACE_PERF;
    zt_tsk = tsk;
    ztrace_yield_to(tsk);
    zt_tsk=NULL;
    zt_flags=0;

    perfcntr_disable_overflow();

    rtsk->perfctr[counter].ctr += branches + perfctr_save;
    zt_perferr=0;

    copy_branchdata(tsk, &br.ctr);
    br.eip= zreg(tsk,eip);
    copy_to_user(bret,&br,sizeof(br));

    return zt_yield;
}

/*
 * At this point, tsk should be about to execute an int 80.  We want
 * tsk to run until in ztrace_tsk_enters_syscall.
 */
int ztrace_syscall_enter(struct task_struct *tsk) {

	//zt_flags &= ZTRACE_ACTMASK;
	zt_flags=ZTRACE_SYSENTER_STOP;
	ztrace_yield_to(tsk);
	zt_flags=0;

	return zt_yield;
}

/*
 * SYSCALL_EXIT will stop after the syscall is reexecuted, before 
 * xusers are replayed.
 */
int ztrace_syscall_exit(struct task_struct *tsk) {

	//zt_flags &= ZTRACE_ACTMASK;
	//
	zt_flags=ZTRACE_SYSEXIT_STOP;
	ztrace_yield_to(tsk);
	zt_flags=0;

	return zt_yield;
}


/*
 * Run until a task blocks at a given eip and wchan, or until
 * we return from a system call at the given eip without blocking
 */
static int ztrace_block(struct task_struct * tsk,
				struct ztrace_rununtil_struct * rs,
				struct ztrace_breturn * bret) {
    if(tsk==current) {
    	printk("ztrace: block self!\n");
	return -1;
    }

    if(copy_from_user(&zt_rununtil, rs, sizeof(zt_rununtil))) {
    	return -EFAULT;
    }

    zt_flags = ZTRACE_BLOCK_STOP;

    ztrace_yield_to(tsk);

    zt_flags=0;

     {
                struct ztrace_breturn br;
                copy_branchdata(tsk, &br.ctr);
                br.eip= zreg(tsk,eip);
		br.wchan =  zt_wchan;
#if 0
# ifdef REPLAY_RECORD_ALL_PFCTRS
                printk("  hw:%llx i:%x pf:%x dev:%x eip:0x%x\n",
                                br.ctr.hw_branch,
                                br.ctr.hw_intr,
                                br.ctr.ex_noerror,
                                br.ctr.ex_error,
                                br.eip);
# else
		printk("  hw: %llx\n", br.ctr.hw_branch);
# endif
#endif
                copy_to_user(bret,&br,sizeof(br));
     }
     return zt_yield;
}

/*
 * Helper functions for ztrace_block and _rununtil
 */
static void ztrace_load_dr(struct task_struct *tsk, int reg, unsigned val) {
	struct thread_struct *ts;
	if (reg<0 || reg>7 || reg==4 || reg==5)
		return ;
	ts= &tsk->thread;
	ts->debugreg[reg]= val;
}

void ztrace_rununtil_set_bkpt(struct task_struct * tsk, unsigned int eip) {
        ztrace_load_dr(tsk, 7, ZTRC_DR7_L0 | ZTRC_DR7_LE |
                               ZTRC_DR7_RW0_EX | ZTRC_DR7_LEN0_0 );
        ztrace_load_dr(tsk, 0, eip); 
}

void ztrace_clear_bkpt(struct task_struct * tsk) {
	ztrace_load_dr(tsk, 7, 0 );
}

static int ztrace_rununtil(struct task_struct *tsk, 
			   struct ztrace_rununtil_struct * rs,
			   struct ztrace_breturn *bret) {
	if (tsk==current) {
		printk("ztrace: rununtiling self!\n");
		return -1;
	}

	if(copy_from_user(&zt_rununtil, rs, sizeof(zt_rununtil))) {
		return -EFAULT;
	}

	zreg(tsk,eflags) &= ~TRAP_FLAG; // FIXME this shouldn't be necessary
	/* FIXME -- do this someplace else */
	tsk->revirt_flags &= ~REPLAY_SSTEP;

	ztrace_rununtil_set_bkpt(tsk, zt_rununtil.eip);
	/*
	 * If we're already at the breakpoint, set the 'resume' flag, so
	 * we don't break until the next one.
	 */
	if(zt_rununtil.eip == zreg(tsk, eip)) {
#ifdef ZTRACE_DEBUG
		printk("ztrace_rununtil: setting EFLAGS_RF for task %d\n", 
				tsk->pid);
#endif
		zreg(tsk,eflags) |= X86_EFLAGS_RF;
	}

	//zt_flags &= ZTRACE_ACTMASK;
	zt_flags = ZTRACE_BKPT;
	if(! (zt_rununtil.flags & REPLAY_PROCESS_IN_UMODE))  {
		zt_flags |= ZTRACE_BKPT_IN_KERNEL;
	}
	/*
	 * For decode & break "single-step" over string instructions
	 */
	if(zt_rununtil.flags & REPLAY_SSTEP) {
		tsk->revirt_flags |= REPLAY_SSTEP;
	}

	ztrace_yield_to(tsk);

	//zt_flags &= ~(ZTRACE_BKPT | ZTRACE_BKPT_IN_KERNEL);
	zt_flags=0;
	if(zt_rununtil.flags & REPLAY_SSTEP) {
		tsk->revirt_flags &= ~REPLAY_SSTEP;
	}

	ztrace_clear_bkpt(tsk);

	{
		struct ztrace_breturn br;
		copy_branchdata(tsk, &br.ctr);
		br.eip= zreg(tsk,eip);
		copy_to_user(bret,&br,sizeof(br));
	}
	return zt_yield;
}

int ztrace_xstop(struct task_struct * tsk,struct ztrace_xuser_stop * addr) {
	int j;
	if((j=copy_from_user(&zt_xstop, addr, sizeof(zt_xstop)))) {
		return -EFAULT;
	}
	//zt_flags &= ZTRACE_ACTMASK;
	zt_flags = ZTRACE_XUSER_STOP;
	ztrace_yield_to(tsk);
	zt_flags=0;
	return zt_yield;
}


int ztrace_exit(struct task_struct * tsk) {
	//zt_flags &= ZTRACE_ACTMASK;
	zt_flags = ZTRACE_WAITEXIT;
	ztrace_yield_to(tsk);
	zt_flags=0;
	return zt_yield;
}

DECLARE_WAIT_QUEUE_HEAD(ztrace_debug);

void ztrace_debug_pause(void) {
	if(signal_pending(current)) {
		printk("ztrace_debug_pause: current->pending.signal=={%lx %lx} [zt_yield]\n",
				current->pending.signal.sig[0], 
				current->pending.signal.sig[1]);
		ztrace_yield_to(NULL);
	} else if(in_interrupt()){
		printk("ztrace_debug_pause: Called from interrupt!!!\n");
		ztrace_yield_to(NULL);
	} else {
	    interruptible_sleep_on(&ztrace_debug);
	}
}

EXPORT_SYMBOL_NOVERS(ztrace_debug_pause);

int ztrace_debug_wake(void) {
	wake_up(&ztrace_debug);
	return 0;
}

int ztrace_marker(int mark) {
	struct rlog_marker_entry me;
	do_gettimeofday(&me.tv);
	if(current->revirt_flags & REPLAY_LOG) {
		me.type = ENTRY_MARKER;
		me.mark=mark;
		revirt_write_log((char *)&me, me.type, sizeof(me));
	} else if(current->revirt_flags & REPLAY_REPLAY) {
		struct rlog_header rh;
		revirt_read_log(&rh);
		if(rh.type != ENTRY_MARKER) {
			printk("ztrace_marker: expecting type %d, got %d! [pause]\n",
				ENTRY_MARKER, rh.type);
			ztrace_debug_pause();
		}
	} 
	printk("ztrace_marker: pid %d, mark %d, %d.%06d sec\n",
			current->pid, mark, 
			(int)me.tv.tv_sec, 
			(int)me.tv.tv_usec);
	return 0;
}

/*************************
 * Random testing stuff...
 *************************/
#if 0
int ztrace_copyuser(struct task_struct * tsk, unsigned int dest, 
			unsigned int len) {
    char * buf;
    unsigned int i;

    buf=kmalloc(len, GFP_KERNEL);
    if(!buf)
       return -ENOMEM;
    for(i=0; i<len; i++)
    	buf[i]=(char) i;
//    printk("ztrace_copyuser: copy_to_user(0x%x, 0x%x, 0x%x)\n",
//    		dest, (unsigned int)buf, len);
    if(copy_to_user((void *)dest, buf, len))
    	return -EFAULT;
    kfree(buf);
    return 0;
}

int ztrace_xstoptest(struct task_struct * tsk, unsigned int dest, 
		unsigned int len) {
	char * d = (char *)dest;
	int r=0;

	if(current->revirt_flags & REPLAY_LOG) {
	    r|=put_user('L', d++);
	    schedule();
	    r|=put_user('o',d++) ;
	    schedule();
	    r|=put_user('g',d++);
	    schedule();
	    r|=copy_to_user(d, "ging", 4);
	    schedule();
	    d+=4;
	    r|=copy_to_user(d, " Stuff\0", 7);
	    schedule();
	    d+=7;
	} else {
	    r|=put_user('R',d++);
	    r|=put_user('e',d++);
	    r|=put_user('p',d++);
	    r|=copy_to_user(d, "layi", 4);
	    d+=4;
	    r|=copy_to_user(d, "ng!\0\0\0\0", 7);
	    d+=7;
	}
	if(r)
	    return -EFAULT;
	else
	    return 0;
}
#endif


int ztrace_smi(void) {
#if 0
	unsigned int lo, hi;
	unsigned int ctr1, ctr2, null;
	unsigned int ctr3, ctr4;
	/*
	 *  Set up the counter to count kernel stuff
	 */
	perfctr_enable_counters();

	/*
	 * Generate an SMI -- platform specific
	 */
	asm("mov $0x884, %dx ; in %dx, %al");
	//e=inb(0x884);
	
	/*
	 * Generate an SMI - platform generic
	 */
	err=(0x2) << 8;
	printk("apic_write(APIC_ICR (0x%x), 0x%x)\n", 
			APIC_ICR, err);
	apic_write(APIC_ICR, err);
	printk("returns\n");
	err=((0x2) << 8) | (1 << 18);
	printk("apic_write(APIC_ICR (0x%x), 0x%x)\n", 
			APIC_ICR, err);
	apic_write(APIC_ICR, err);
	printk("returns\n");

	perfctr_enable_counters();
#endif
	return 0;
}

void show_stack(unsigned long *);

int ztrace_printstack(struct task_struct * tsk) {
	unsigned long * esp;
	esp=(unsigned long *)tsk->thread.esp;
	printk("Stack for pid %d @ 0x%lx:\n", tsk->pid, (unsigned long)esp);
	show_stack(esp);
	return 0;
}

asmlinkage int sys_ztrace(int command,int pid,unsigned long addr, 
			unsigned long data, unsigned int len)
{
	int err;
	struct task_struct *tsk=NULL;
	int waitfl;
	unsigned long val;
	struct ztrace_state_struct state;

	switch(command) {
		case ZTRACE_MARKER:
		case ZTRACE_SMI:
		    tsk=NULL; 
		    break;
		default:
		    err= -ESRCH;
		    tsk= find_task_by_pid(pid);
		    if (!tsk) 
			goto out;
	}

	err= 0;
	waitfl= 0;
	/*
	 * FIXME -- we need to setup checks for some actions to make sure they
	 * don't trace themselves.
	 *
	 * I suppose we should also eventually do checks to make sure 
	 * only root can call these. =)
	 */
	switch(command) {
	  case ZTRACE_SET_OPT:
	  	err=ztrace_set_opt(tsk, addr, data);
		break;
	  /* Data control stuff */
	  case ZTRACE_GETREG:
	  	err= ztrace_get_reg(tsk,addr,&val);
	  	if (!err)
	  		err= put_user(val,(unsigned long*) data);
	  	break;
	  case ZTRACE_PUTREG:
		err=ztrace_put_reg(tsk, addr, data);
	  	break;
	  case ZTRACE_ATTACH:
	  	if(tsk!=current && tsk->revirt_tsk)
		    tsk->revirt_tsk->sched= current;
		else
		    err=-EPERM;
	  	break;
	  case ZTRACE_RESUME:
	  	tsk->revirt_flags = 0;
		data=1;
		zt_flags = 0;
		/* FALLTHRU */
	  case ZTRACE_DETACH:
	  	if(data && tsk->revirt_tsk)
		    tsk->revirt_tsk->sched=NULL;
		zreg(tsk,eflags) &= ~TRAP_FLAG;
		/* FIXME: do we need to worry about the lock here? */
		tsk->revirt_flags &= ~REPLAY_WAIT;
		tsk->revirt_flags &= ~REPLAY_SSTEP;
		break;
	  case ZTRACE_GETSTATE:
	  	err= ztrace_getstate(tsk,&state);
	  	if (!err)
			copy_to_user((void *)data,&state,sizeof(state));
	  	break;
	  case ZTRACE_PEEK:
	  	err = ztrace_peek(tsk, addr, (unsigned long *)data);
		break;

	  /* Flow control */
	  case ZTRACE_SSTEP_WAIT:
	  	waitfl=1;
		/* FALLTHRU */
	  case ZTRACE_SSTEP:
	  	err=ztrace_sstep(tsk,waitfl);
	  	break;
	  case ZTRACE_SYSCALL_ENTER:
	  	err=ztrace_syscall_enter(tsk);
	  	break;
	  case ZTRACE_SYSCALL_EXIT:
	  	err=ztrace_syscall_exit(tsk);
	  	break;
	  case ZTRACE_SIGNALQUEUE:
	  	err = ztrace_signalqueue(tsk, (void *)data);
		break;
	  case ZTRACE_RUNUNTIL:
	  	err= ztrace_rununtil(tsk,
				(struct ztrace_rununtil_struct *)addr,
				(struct ztrace_breturn *)data);
	  	break;
	  case ZTRACE_BLOCK:
	  	err = ztrace_block(tsk,
				(struct ztrace_rununtil_struct *)addr,
				(struct ztrace_breturn *)data);
	  	break;
		break;
	  case ZTRACE_XSTOP:
	  	err=ztrace_xstop(tsk,(struct ztrace_xuser_stop *)addr);
		break;
	  case ZTRACE_EXIT:
	  	err=ztrace_exit(tsk);
		break;
	  case ZTRACE_PERFOVF:
	  	err=ztrace_perf(tsk, 
				(s64 *)addr, 
				(struct ztrace_breturn *)data);
		break;
#if 0
	  case ZTRACE_COPYUSER:
	  	err = ztrace_copyuser(tsk, addr, data);
		break;
	  case ZTRACE_HANG:
	  	while(1);
		break;
	  case ZTRACE_XSTOPTEST:
	  	err=ztrace_xstoptest(tsk, addr, data);
		break;
#endif
	  /* Debugging help */
	  case ZTRACE_DEBUGWAKE:
	  	err=ztrace_debug_wake();
		break;
	  case ZTRACE_MARKER:
	  	err=ztrace_marker(pid);
		break;
	  case ZTRACE_CLEARBRANCH:
	  	clear_branchdata(tsk);
		err=0;
		break;
	  case ZTRACE_PRINTSTACK:
		err=ztrace_printstack(tsk);
		break;
	  /* Random testing stuff */
	  case ZTRACE_SMI:
		err=ztrace_smi();
		break;
	  default:
	  	err= -ENOSYS;
	}

out:
	return err;
}

