#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>

#define BUFSIZE (1024 * 1024) 
unsigned char buf[BUFSIZE];

#define NUMTYPES 24

int main(int argc, char * argv[]) {
	int fd;
	int ar[2], i, j, l;
	struct rlog_header rh;
	struct rlog_schedule_entry * le;
	char * names[] = { "SCHED", "FORK", "EXIT", "TOP", "SIGNAL", 
		"EXECPARM", 0, 0, 0, 0, 0, 0, 0, 0, 0, "PUT_USER", 
		"COPY_USER", "SYSRET", "CLEAR_USER", "COPY_USER_NODATA",
		"ENTRY_SIGFRAME", "ENTRY_RT_SIGFRAME", "ENTRY_MARKER",
		"ENTRY_SENDERIP" };

	unsigned records=0, bytes=0;

	unsigned type_count[NUMTYPES];
	unsigned type_data[NUMTYPES];
	unsigned syscall_count[256][12];
	unsigned signal_count[32];
	unsigned pidmap[12];
	int pidmapmax=0;
	unsigned copyuser_data[256];
	unsigned putuser_data[256];

	for(i=0; i<NUMTYPES; i++) {
		type_count[i]=type_data[i]=0;
	}

	for(i=0; i<256; i++) {
		copyuser_data[i]=putuser_data[i]=0;
		for(j=0; j<12; j++) {
			syscall_count[i][j]=0;
		}
	}

	for(i=0; i<32; i++) {
		signal_count[i]=0;
	}

	if(argc!=2) {
		fprintf(stderr, "Usage: %s logfile\n", argv[0]);
	}
	printf("Opening file %s\n", argv[1]);
	if((fd=open(argv[1], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	while((l=read(fd, &rh, sizeof(rh)))>0) {
		records++;
		bytes+=l + rh.len;
		if(!(records % 100000)) 
			fprintf(stderr, "%d records [%d bytes] read...\n", records, bytes);
		type_count[rh.type & 0x1F]++;
		type_data[rh.type & 0x1F] += rh.len;
		switch(rh.type) {
		    case ENTRY_TOP:
			break;
		    case ENTRY_SCHED:
			break;
		    case ENTRY_FORK:
			break;
		    case ENTRY_EXIT:
			break;
		    case ENTRY_SIGNAL:
			break;
		    case ENTRY_PUT_USER:
			break;
		    case ENTRY_COPY_USER:
			break;
		    case ENTRY_CLEAR_USER:
			break;
		    case ENTRY_SYSRET:
			break;
		    default:
		}
		if((l=read(fd, buf, rh.len))<rh.len) {
			if(l<0)
				perror("reading file");
			else
				fprintf(stderr, "Bad file format: wanted %d, got %d\n", rh.len, l);
			break;
		}

		if(1) {
			union {
				struct rlog_entry_signal signal;
				struct rlog_put_user_entry put_user;
				struct rlog_copy_user_entry copy_user;
				struct rlog_clear_user_entry clear_user;
				struct rlog_entry_sysret sysret;
				struct rlog_schedule_entry sched;
				struct rlog_entry_sigframe sigframe;
				struct rlog_entry_rt_sigframe rt_sigframe;
				struct rlog_marker_entry mark;
			} * le = (__typeof(le))buf;
			struct __kernel_siginfo *si;

		       switch(rh.type) {
			  case ENTRY_MARKER:
				printf("Mark %d %ld.%06d lo %d sys %d sig %d\n",
					le->mark.mark,
					le->mark.tv.tv_sec,
					le->mark.tv.tv_usec,
					bytes,
					type_count[ENTRY_SYSRET & 0xff],
					type_count[ENTRY_SIGNAL & 0xff]);
				break;
			  case ENTRY_SCHED:
			  case ENTRY_CLEAR_USER:
			  case ENTRY_FORK:
				break;
			  case ENTRY_PUT_USER:
			  	putuser_data[le->put_user.syscall] += 
						le->put_user.ssize;
			  	break;
			  case ENTRY_COPY_USER:
			  	copyuser_data[le->copy_user.syscall] += 
						le->copy_user.ssize;
				break;
			  case ENTRY_SYSRET:
				j=-1;
			  	for(i=0; i<pidmapmax ; i++) {
					if(pidmap[i]==le->sysret.pid) {
						j=i;
						break;
					}
				}
				if(j<0) {
					if(pidmapmax > 12) {
					  fprintf(stderr, "Maxed outpids...\n");
					  goto printout;
					}
					j=pidmapmax;
					pidmap[j]=le->sysret.pid;
					pidmapmax++;
				}
			  	syscall_count[le->sysret.syscall][0]++;
			  	syscall_count[le->sysret.syscall][j+1]++;
				break;
			  case ENTRY_SIGNAL:	
#if 0
				si=(struct __kernel_siginfo *)(buf + sizeof(struct rlog_entry_signal));
			  	signal_count[si->si_signo]++;
#endif
				break;
		      }
		} 
	}

    printout:

	printf("%d records\n", records);
	printf("Log types:\n");

#define PRNT(x) printf(" %s cnt: %d data:%d (%d in file)\n", names[x&0x1F], \
	type_count[x & 0x1F], type_data[x & 0x1F], 	\
	type_count[x & 0x1F] * sizeof(rh) + type_data[x & 0x1F]);

	PRNT(ENTRY_SCHED);
	PRNT(ENTRY_FORK);
	PRNT(ENTRY_EXIT);
	PRNT(ENTRY_TOP);
	PRNT(ENTRY_SIGNAL);
	PRNT(ENTRY_EXECPARM);
	PRNT(ENTRY_PUT_USER);
	PRNT(ENTRY_COPY_USER);
	PRNT(ENTRY_SYSRET);
	PRNT(ENTRY_CLEAR_USER);
	PRNT(ENTRY_COPY_USER_NODATA);
	PRNT(ENTRY_SIGFRAME);
	PRNT(ENTRY_MARKER);
	PRNT(ENTRY_SENDERIP);
#undef PRNT

#if 0
	printf(" ENTRY_FORK cnt: %d data:%d\n", type_count[ENTRY_FORK&0x1F], type_data[ENTRY_FORK&0x1F]);
	printf(" ENTRY_EXIT cnt: %d data:%d\n", type_count[ENTRY_EXIT&0x1F], type_data[ENTRY_EXIT&0x1F]);
	printf(" ENTRY_TOP cnt: %d data:%d\n", type_count[ENTRY_TOP&0x1F], type_data[ENTRY_TOP&0x1F]);
	printf(" ENTRY_SIGNAL cnt: %d data:%d\n", type_count[ENTRY_SIGNAL&0x1F], type_data[ENTRY_SIGNAL&0x1F]);
	printf(" ENTRY_EXECPARM cnt: %d data:%d\n", type_count[ENTRY_EXECPARM&0x1F], type_data[ENTRY_EXECPARM&0x1F]);
	printf(" ENTRY_PUT_USER cnt: %d data:%d\n", type_count[ENTRY_PUT_USER&0x1F], type_data[ENTRY_PUT_USER&0x1F]);
	printf(" ENTRY_COPY_USER cnt: %d data:%d\n", type_count[ENTRY_COPY_USER&0x1F], type_data[ENTRY_COPY_USER&0x1F]);
	printf(" ENTRY_SYSRET cnt: %d data:%d\n", type_count[ENTRY_SYSRET&0x1F], type_data[ENTRY_SYSRET&0x1F]);
	printf(" ENTRY_CLEAR_USER cnt: %d data:%d\n", type_count[ENTRY_CLEAR_USER&0x1F], type_data[ENTRY_CLEAR_USER&0x1F]);
	printf(" ENTRY_COPY_USER_NODATA cnt: %d data:%d\n", type_count[ENTRY_COPY_USER_NODATA&0x1F], type_data[ENTRY_COPY_USER_NODATA&0x1F]);
#endif
	printf("Signal count:\n");
	for(i=0; i<8; i++) {
	  printf("%2d:%6d %2d:%6d %2d:%6d %2d:%6d\n", 
	  	i, signal_count[i],
	  	i+8, signal_count[i+8],
	  	i+16, signal_count[i+16],
	  	i+24, signal_count[i+24]);
	}
	printf("Syscall: Count copy_user put_user\n");
	for(i=0; i<256; i++) {
	   printf("%3d: %10d %10d %10d\n", i, syscall_count[i][0], 
		copyuser_data[i], putuser_data[i]);
	}
       for(j=0; j<pidmapmax; j++) {
        printf("\n---Pid %d:%d:\n", j, pidmap[j]);
	for(i=0; i<256; i++) {
	   if(syscall_count[i][j+1])
		printf("%3d: %10d\n", i, syscall_count[i][j+1]);
	}
       }
}
