#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/flags.h>

/*
 * So asm/siginfo.h will compile...
 */
typedef long clock_t;

#include <asm/siginfo.h>

#define BUFSIZE (1024 * 1024) 
unsigned char buf1[BUFSIZE];
unsigned char buf2[BUFSIZE];

#ifndef REPLAY_RECORD_ALL_PERFCTRS
int main(void) {
	fprintf(stderr, "RECORD_ALL_PERFCTRS not set.\n");
	exit(1);
}
#else
void print_bc(struct revirt_ctrdata * le) {
	int i;
	printf(" x:0x%llx ix:0x%x p:0x%x d:0x%x\n",
		le->hw_branch, 
		le->hw_intr,
		le->ex_pf,
		le->ex_dev);
	if(le->hw_intr != le->ex_dev) {
		printf(" ---Interrupt counts do not match---\n");
	}
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
#warning listing exceptions
	for(i=0; i<REVIRT_NR_EX; i++) {
		if(le->exceptions[i]>0)
			printf("  [%d] 0x%x\n", i, le->exceptions[i]);
	}
#endif
	printf(" [ ");
	for(i=0; i<REVIRT_NR_PFCTRS; i++) {
		printf("%llx ", le->counter[i]);
	}
	printf("]\n");
}

struct rlog_entry_sysret * get_next_sysret(int fd, 
		struct rlog_header * rh, char * buf) {
	struct rlog_entry_sysret * ret = NULL;
	int l;
	do {
	    if((l=read(fd, rh, sizeof(*rh)))<0) {
		perror("read");
		exit(1);
	    }
	    if(l>0) {
		if((l=read(fd, buf, rh->len))<rh->len) {
			if(l<0) {
				perror("reading file");
				exit(1);
			} else {
				fprintf(stderr, "Bad file format: wanted %d, got %d\n", rh->len, l);
				exit(1);
			}
		}
	    } else {
		return NULL;
	    }
	} while(rh->type!=ENTRY_SYSRET);
	return (struct rlog_entry_sysret*)buf;
}

void print_stuff(struct rlog_header * rh, struct rlog_entry_sysret * s) {
	int i;
	printf("%d: pid %d syscall %d\n", 
			rh->id, s->pid, s->syscall);
	printf(" b: %llx i:%x p:%x d:%x br:%llx in:%llx ms:%llx ur:%llx [5]:%llx [6]:%llx [7]:%llx\n",
			s->ctr.hw_branch,
			s->ctr.hw_intr,
			s->ctr.ex_pf,
			s->ctr.ex_dev,
			s->ctr.counter[0],
//			s->ctr.counter[1],
			s->ctr.counter[2],
			s->ctr.counter[3],
			s->ctr.counter[4],
			s->ctr.counter[5],
			s->ctr.counter[6],
			s->ctr.counter[7]);
	for(i=0; i<REVIRT_NR_EX; i++) {
		if(s->ctr.exceptions[i]>0) {
			printf(" [%d]:%x",
				i, s->ctr.exceptions[i]);
		}
	}
	printf("\n");
}

int main(int argc, char * argv[]) {
	int fd1, fd2;
	int i, j, l, pause, last_id=0;
	int records=0;
	struct rlog_header rh1, rh2, lrh1, lrh2;
	struct rlog_entry_sysret * s1, *s2, l1, l2, d1, d2;
	struct siginfo * si;
	long long c1, c2;
	int d[8] = {0, 0, 0, 0, 0, 0,0,0};
	int dcount[8]={0,0,0,0,0,0,0,0};
	int delta, correct=0;

	bzero(&l1, sizeof(l1));
	bzero(&l2, sizeof(l2));
	bzero(&lrh1, sizeof(lrh1));
	bzero(&lrh2, sizeof(lrh2));

	if(argc!=3) {
		fprintf(stderr, "Usage: %s logfile1 logfile2\n", argv[0]);
		exit(1);
	} 

	printf("Opening file %s\n", argv[1]);
	if((fd1=open(argv[1], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	printf("Opening file %s\n", argv[2]);
	if((fd2=open(argv[2], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	while((s1=get_next_sysret(fd1, &rh1, buf1)) && 
	      (s2=get_next_sysret(fd2, &rh2, buf2))) {
		pause=0;
		records++;
		if(rh1.id-last_id > 50000) {
			printf("s1 id %u\n", rh1.id); 
			last_id=rh1.id;
		}
		if(s1->ctr.hw_branch < l1.ctr.hw_branch) {
			if(s2->ctr.hw_branch >= l2.ctr.hw_branch) {
				printf("s1 resets, but s2 doesnt!\n");
				pause=1;
			}
			//printf("Counters reset\n");
			for(i=0; i<8; i++)
				d[i]=0;
			d1=*s1;
			d2=*s2;
		} else {
			d1.pid=l1.pid;
			d1.syscall=l1.syscall;
#define D(d,s,l,x) (d.ctr.x = s->ctr.x - l.ctr.x)
			D(d1, s1, l1, hw_branch);
#define E(a,b,c) \
			D(a, b, c, hw_branch); \
			D(a, b, c, hw_intr); \
			D(a, b, c, ex_pf); \
			D(a, b, c, ex_dev); \
			for(i=0; i<8; i++) { \
				D(a, b, c, counter[i]); \
			} \
			for(i=0; i<REVIRT_NR_EX; i++) { \
				D(a, b, c, exceptions[i]); \
			} 
			E(d1, s1, l1);
			E(d2, s2, l2);

		}
		//printf("%d %d\n", rh1.id, rh2.id);
		if((s1->pid != s2->pid) 
		   || (s1->syscall != s2->syscall)) {
			printf("Hmm... logs seem different.\n");
			printf("%d: pid %d syscall %d\n%d: pid %d syscall %d\n",
					rh1.id, s1->pid, s1->syscall,
					rh2.id, s2->pid, s2->syscall);
			pause=1;
		}
		c1=s1->ctr.counter[0] - s1->ctr.hw_intr - s1->ctr.ex_pf
				- s1->ctr.ex_dev - s1->ctr.counter[7]/2;
		c2=s2->ctr.counter[0] - s2->ctr.hw_intr - s2->ctr.ex_pf
				- s2->ctr.ex_dev - s2->ctr.counter[7]/2;
		//printf("%llx %llx\n", c1, c2);
		if(c1-c2!=d[0]) {
			delta=(c1-c2) - d[0];
			dcount[0]++;
			d[0]=c1-c2;
			printf("branch_count drifts: %llx %llx (%lld) [%d]\n",
					c1, c2, c1-c2, delta);
			pause=1;
		}
		c1=s1->ctr.counter[2]; //- s1->ctr.hw_intr;
		c2=s2->ctr.counter[2]; //- s2->ctr.hw_intr;
		//printf("%llx %llx\n", c1, c2);
		if(c1-c2!=d[2]) {
			delta=(c1-c2) - d[2];
			dcount[2]++;
			d[2]=c1-c2;
#if 0
			printf("instr_retired drifts: %llx %llx (%lld) [%d]\n",
					c1, c2, c1-c2, delta);
			//pause=1;
#endif
		}
		c1=s1->ctr.counter[3] - 3*(s1->ctr.hw_intr + s1->ctr.ex_dev)
			- s1->ctr.ex_pf;
		c2=s2->ctr.counter[3] - 3*(s2->ctr.hw_intr + s2->ctr.ex_dev)
			- s2->ctr.ex_pf;
		//printf("%llx %llx\n", c1, c2);
		if(c1-c2!=d[3]) {
			delta=(c1-c2) - d[3];
			dcount[3]++;
			d[3]=c1-c2;
#if 0
			printf("memory_stores drifts: %llx %llx (%lld) [%d]\n",
					c1, c2, c1-c2, delta);
			//pause=1;
#endif
		}

		c1=s1->ctr.counter[4];
		c2=s2->ctr.counter[4];
		//printf("%llx %llx\n", c1, c2);
		if(c1-c2!=d[4]) {
			delta=(c1-c2) - d[4];
			dcount[4]++;
			d[4]=c1-c2;
#if 0
			printf("uops drifts: %llx %llx (%lld) [%d]\n",
					c1, c2, c1-c2, delta);
			pause=1;
#endif
		}
		c1=s1->ctr.counter[6];
		c2=0;
		if(c1-c2 != d[6]) {
			delta=(c1-c2) - d[6];
			dcount[6]++;
			d[6]=c1-c2;
			printf("s1 smi: %llx %llx (%d) [%d]\n",
					c2, c1, d[6], delta);
			//pause=1;
		}
		c1=s2->ctr.counter[6];
		c2=0;
		if(c1-c2 != d[7]) {
			delta=(c1-c2) - d[7];
			dcount[7]++;
			d[7]=c1-c2;
			printf("s2 smi: %llx %llx, (%d) [%d]\n",
					c2, c1, d[7], delta);
			//pause=1;
		}
		c1=s1->ctr.counter[6];
		c2=s1->ctr.counter[7]/2;
		if(c1-c2 != d[1]) {
			delta=(c1-c2) - d[1];
			dcount[1]++;
			d[1]=(c1-c2);
			printf("s1: ack & smi differ: %llx %llx (%d) [%d]\n",
					c1, c2, d[1], delta);
			//pause=1;
		} else if (d1.ctr.counter[7]>0) {
			correct++;
		}
		c1=s2->ctr.counter[6];
		c2=s2->ctr.counter[7]/2;
		if(c1-c2!=d[5]) {
			delta=(c1-c2) - d[5];
			dcount[5]++;
			d[5]=(c1-c2);
			printf("s2: ack & smi differ: %llx %llx (%d) [%d]\n",
					c1, c2, d[5], delta);
			//pause=1;
		} else if (d2.ctr.counter[7]>0) {
			correct++;
		}
		if(pause) {
			printf("l1|");
			print_stuff(&lrh1, &l1);
			printf("l2|");
			print_stuff(&lrh2, &l2);
			printf("s1|");
			print_stuff(&rh1, s1);
			printf("s2|");
			print_stuff(&rh2, s2);
			printf("d1|");
			print_stuff(&rh1, &d1);
			printf("d2|");
			print_stuff(&rh2, &d2);
			getchar();
		}
		l1=*s1;
		l2=*s2;
		lrh1=rh1;
		lrh2=rh2;
	}
	printf("All done!  %d records examined.\n", records);
	printf("branch_retired drifts: %d\n", dcount[0]);
	printf("instr_retired drifts: %d\n", dcount[2]);
	printf("memory_store drifts: %d\n", dcount[3]);
	printf("uop_retired drifts: %d\n", dcount[4]);
	printf("logging smis: %d\n", dcount[6]);
	printf("replay smis: %d\n", dcount[7]);
	printf("smi int / ack drifts: %d\n", dcount[1]);
	printf("smi int / ack correlations: %d\n", correct);
	return dcount[6]+dcount[7];
}
		
int printdata(char *buf, int l) {	
	int i, j;
		for(i=0; i<l; i+=16) {
		    printf(" %08x:", i);
		    for(j=i; j<i+16 && j<l; j++) {
			if (!(j%4) && j>0)
				printf(" ");
			printf("%02x ", (unsigned int)buf[j]);
                    }
		    if(j<i+16) {
			for(; j<i+16; j++) {
				if(!j%4)
					printf("  ");
				printf("   ");
			}
		    }
		    printf("  ");
		    for(j=i; j<i+16; j++) {
			if(isprint(buf[j]))
				fputc(buf[j], stdout);
			else
				fputc('.', stdout);
		    }
		    printf("\n");
		}
}
#endif
