/*
 * $Id: umapic.c,v 1.1.1.1 2003/05/27 21:45:37 root Exp $
 */
#include <linux/config.h>

#if defined(__SMP__) || defined(CONFIG_SMP)

#include <asm/io.h>
#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>
#include <asm/apicdef.h>
#include <linux/assert.h>

#define APIC_SET(x,y)	(*((volatile unsigned long *)(APIC_BASE + x)) = y)
#define APIC_GET(x)	*((volatile unsigned long *)(APIC_BASE + x))
#define	CPU	(GET_APIC_ID(*(unsigned long *)(APIC_BASE+APIC_ID)))

#define UMAPIC_BUS_HZ	60000000
#define UMAPIC_EMUL_HZ	20
#define CALIBRATE_TIME	1000000

/* read per-cpu tsc */
extern void umcpu_tsc_get(unsigned long *, unsigned long *);

/* read tsc of hosting CPU */
#define rdtsc_real(low, high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))

/* XXX HSHOEXER */
unsigned long start_tsc_low[32] = { -1, /* ... */ };
unsigned long start_tsc_high[32] = { -1, /* ... */ };
static unsigned long tsc_to_bus = 0;		/* CPU ticks per bus tick */

unsigned char umapic_ports[2] = { -1, /* ... */ };	/* IMCR ports */

int
umapic_timer_base(void)
{
	int base;
	unsigned long value;
	int mask;

	value = APIC_GET(APIC_TDCR);

	/* The bits 0,1 and 3 define the timer base. */
	mask = 0;
	mask |= (value & 0x3);
	mask |= ((value & 0x8) >> 1);

	switch(mask) {
	case 0:
		base = 2;
		break;

	case 1:
		base = 4;
		break;

	case 2:
		base = 8;
		break;

	case 3:
		base = 16;
		break;

	case 4:
		base = 32;
		break;

	case 5:
		base = 64;
		break;

	case 6:
		base = 128;
		break;

	case 7:
		base = 1;
		break;

	default:
		assert(0);
	}

	return base;
}

unsigned long
umapic_read(unsigned long reg)
{
	char buffer[64];
	unsigned long value;
	unsigned long delta;
	unsigned long cur_tsc_low;
	unsigned long cur_tsc_high;

	extern void sub64(unsigned long *, unsigned long *, unsigned long *, 
			  unsigned long *);

	switch(reg) {
	case APIC_TMCCT:
		
		umcpu_tsc_get(&cur_tsc_low, &cur_tsc_high);
		sub64(&cur_tsc_low, &cur_tsc_high, 
				&start_tsc_low[CPU], &start_tsc_high[CPU]);

		delta = cur_tsc_low / tsc_to_bus;
		delta /= umapic_timer_base();
		delta %= APIC_GET(APIC_TMICT);
		value = APIC_GET(APIC_TMICT) - delta;
		break;

	default:
		value = APIC_GET(reg);
	}

	sprintf(buffer, "umapic_read:  CPU#%ld reg=0x%04lx val=0x%08lx\n", CPU, 
			reg, value);
	lotol_write(2, buffer, strlen(buffer));

	return value;
}

/*
 * This is the stub for the APIC emulation.  Anything that can be done
 * locally, is done locally.  Everything else is sent to the APIC emulation
 * and served there.
 */
void
umapic_write(unsigned long reg, unsigned long val)
{
	char buffer[64];
	int ret;
	
	sprintf(buffer, "umapic_write: CPU#%ld reg=0x%04lx val=0x%08lx", CPU, 
			reg, val);
	lotol_write(2, buffer, strlen(buffer));

	/* These registers are read-only */
	switch(reg) {
	case APIC_LVR:
	case APIC_ARBPRI:
	case APIC_PROCPRI:
	case APIC_ESR:
	case APIC_TMCCT:
		sprintf(buffer, "\tcanceled (read-only)\n");
		lotol_write(2, buffer, strlen(buffer));
		break;

	/* These are read/write (execept EOI which is write-only) */
	case APIC_ID:
		sprintf(buffer, "\tLocal APIC ID register set to 0x%08lx\n",
				val);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_TASKPRI:
		sprintf(buffer, "\tTPR set to 0x%08lx\n", val & APIC_TPRI_MASK);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_EOI:
		sprintf(buffer, "\tEOI: cpu=%ld %ld\n", CPU, FD_TO_LAPIC0 + CPU);
		lotol_write(2, buffer, strlen(buffer));

		sprintf(buffer, "%ld:%ld\n", reg, val);
		ret = lotol_write(FD_TO_LAPIC0 + CPU, buffer, strlen(buffer));
		assert(ret == strlen(buffer));
		break;

	case APIC_LDR:
		sprintf(buffer,"\tLDR set to 0x%08lx\n",GET_APIC_LOGICAL_ID(val));
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_DFR:
		switch (((val >> 24) & 0xFF)) {
		case 0xFF:
			sprintf(buffer, "\tDFR set to flat model\n");
			lotol_write(2, buffer, strlen(buffer));
			break;

		case 0x0:
			sprintf(buffer, "\tDFR set to clustered model\n");
			lotol_write(2, buffer, strlen(buffer));
			break;

		default:
			sprintf(buffer, "\tDFR: illegal model\n");
			lotol_write(2, buffer, strlen(buffer));
			break;
		}

		APIC_SET(reg, val);
		break;

	case APIC_SPIV:
		if (val & APIC_SPIV_APIC_ENABLED) {
			sprintf(buffer, "\tSVR: APIC enabled\n");
			lotol_write(2, buffer, strlen(buffer));
		} else {
			sprintf(buffer, "\tSVR: APIC disabled\n");
			lotol_write(2, buffer, strlen(buffer));
		}

		APIC_SET(reg, val);
		break;

	case APIC_ICR:
		/* Writing to ICR1 always triggers an interrupt */
		sprintf(buffer, "\tICR1: cpu=%ld %ld\n", CPU, FD_TO_LAPIC0 + CPU);
		lotol_write(2, buffer, strlen(buffer));

		/* First of all set delivery state to pending */
		val |= APIC_ICR_BUSY;
		APIC_SET(reg, val);

		sprintf(buffer, "%ld:%ld\n", reg, val);
		ret = lotol_write(FD_TO_LAPIC0 + CPU, buffer, strlen(buffer));
		assert(ret == strlen(buffer));
		break;

	case APIC_ICR2:
		sprintf(buffer, "\tICR2: Destination: ID#%ld\n", 
				GET_APIC_DEST_FIELD(val));
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_LVTT:
		sprintf(buffer, "\tLVTT set to 0x%08lx\n", val);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_LVTPC:
		sprintf(buffer, "\tnot yet implemented\n");
		lotol_write(2, buffer, strlen(buffer));
		break;

	case APIC_LVT0:
		sprintf(buffer, "\tLVT - LINT0 set to 0x%08lx\n", val);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_LVT1:
		sprintf(buffer, "\tLVT - LINT1 set to 0x%08lx\n", val);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	case APIC_LVTERR:
		sprintf(buffer, "\tnot yet implemented\n");
		lotol_write(2, buffer, strlen(buffer));
		break;

	case APIC_TMICT:
		/* Writing to TMICT starts the timer */
		sprintf(buffer, "\tTMICT set to 0x%08lx\n", val);
		lotol_write(2, buffer, strlen(buffer));

		umcpu_tsc_get(&start_tsc_low[CPU], &start_tsc_high[CPU]);

		APIC_SET(reg, val);
		APIC_SET(APIC_TMCCT, val);
		break;

	case APIC_TDCR:
		sprintf(buffer, "\tTDCR set to 0x%08lx\n", val);
		lotol_write(2, buffer, strlen(buffer));

		APIC_SET(reg, val);
		break;

	default:
		/* ISR, TMR  and IRR (0x100-0x280) are also read-only */
		if (APIC_ISR <= reg && reg < APIC_ESR) {
			sprintf(buffer, "\tcanceled (read-only)\n");
			lotol_write(2, buffer, strlen(buffer));
			break;;
		}

		/* This should not happen */
		panic("illegal access!\n");
	}

	return;
}

unsigned char
umapic_inb(unsigned short port)
{
	return umapic_ports[port];
}

void
umapic_outb(unsigned char value, unsigned short port)
{
	switch(port) {
	case 0:
		umapic_ports[port] = value;
		break;

	case 1:
		if (umapic_ports[0] == 0x70) {
			umapic_ports[1] = value;
		} else {
			assert(0);
		}
		break;

	default:
		assert(0);
	}

}

void
umapic_timerint(void)
{
	if (!lapic->lvt_timer.mask) {
		lapic->lvt_timer.delivery_status = 1;
	}
}

/*
 * Return mask of an interrupt in the local vector table.
 */
int
umapic_test_lint(int irq)
{
	int ret;

	if (lapic->lvt_lint0.vector == irq) {
		ret = lapic->lvt_lint0.mask;

	} else if (lapic->lvt_lint1.vector == irq) {
		ret = lapic->lvt_lint1.mask;

	} else {
		ret = 0;
	}

	return ret;
}

int
umapic_test_irq(void)
{
	unsigned int irq;
	unsigned int offset;
	unsigned int mask;

	for (irq = 255; 16 <= irq; irq--) {
		/* Check timer vector */
		if (lapic->lvt_timer.vector == irq) {
			/* There is a pending interrupt */
			if (lapic->lvt_timer.delivery_status) {
				return 1;
			}
		}

		/* XXX check lints */

		offset = irq >> 5;
		mask = 1 << (irq & 0x1F);

		/* XXX handle TPR correctly! */

		if (lapic->irr[offset].bitfield & mask) {
			return 1;
		}
	}

	return 0;
}

int
umapic_ack_irq(void)
{
	unsigned int irq;
	unsigned int offset;
	unsigned int mask;

	for (irq = 255; 16 <= irq; irq--) {
		offset = irq >> 5;
		mask = 1 << (irq & 0x1F);

		/* XXX handle TPR correctly! */

		if(lapic->irr[offset].bitfield & mask) {
			lapic->irr[offset].bitfield &= ~mask;
			lapic->isr[offset].bitfield |= mask;
			break;
		}
	}

	/* There is an interrupt pending at the timer */
	if (lapic->lvt_timer.delivery_status) {
		if (irq < lapic->lvt_timer.vector) {
			irq = lapic->lvt_timer.vector;
		}
	}

	/* XXX check lints */

	return irq;
}

void
umapic_calibrate(void)
{
	struct lotol_timeval tvstart, tvend, tvtmp;
	unsigned long startlow, starthigh;
	unsigned long endlow, endhigh;
	int ret;

	extern void sub64(unsigned long*, unsigned long*, unsigned long *, 
			  unsigned long *);

	ret = lotol_gettimeofday(&tvtmp, NULL);
	tvstart.tv_sec = tvtmp.tv_sec;
	tvstart.tv_usec = tvtmp.tv_usec;
	tvend.tv_sec = tvtmp.tv_sec + CALIBRATE_TIME / 1000000;
	tvend.tv_usec = tvtmp.tv_usec + CALIBRATE_TIME % 1000000;

	/* normalize microseconds */
	while (1000000 <= tvend.tv_usec) {
		tvend.tv_sec++;
		tvend.tv_usec -= 1000000;
	}

	/* calibrate */
	rdtsc_real(startlow, starthigh);
	do {
		ret = lotol_gettimeofday(&tvtmp, NULL);
		assert(0 <= ret);
	} while (tvtmp.tv_sec < tvend.tv_sec || tvtmp.tv_usec < tvend.tv_usec);
	rdtsc_real(endlow, endhigh);

	sub64(&endlow, &endhigh, &startlow, &starthigh);
	assert(endhigh == 0);

	/* tsc ticks per timer ticks */
	tsc_to_bus = endlow / UMAPIC_BUS_HZ;

	return;
}

void
umapic_init(void)
{ 
	int ret;
	struct lotol_itimerval it;
	static int calibrated = 0;

	/* FIXME HSHOEXER: fixed mapping to 0x7fffe000 */
	ret = (unsigned long) lotol_mmap((char *)APIC_BASE, PAGE_SIZE,
		  LOTOL_PROT_READ | LOTOL_PROT_WRITE | LOTOL_PROT_EXEC,
		  LOTOL_MAP_SHARED | LOTOL_MAP_FIXED, FD_APIC_MEMORY,
		  0);
	assert(ret == APIC_BASE);

	/* Reset IMCR */
	umapic_ports[0] = 0;
	umapic_ports[1] = 0;

	/* Do calibration only once */
	if (!calibrated) {
		umapic_calibrate();
		calibrated = 1;
	}

	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 1000000 / UMAPIC_EMUL_HZ;
	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = 1000000 / UMAPIC_EMUL_HZ;

	ret = lotol_setitimer(LOTOL_ITIMER_REAL, &it, NULL);
	assert(0 <= ret);

	return;
}
#else
unsigned char
umapic_inb(unsigned short port)
{
	return 0;
}

void
umapic_outb(unsigned char value, unsigned short port)
{
	return;
}

void
umapic_init(void)
{
	return;
}
#endif /* defined(__SMP__) || defined(CONFIG_SMP) */
