/*
 * $Id: umcore.c,v 1.2 2003/07/02 22:05:53 cvs Exp $
 */

#define DEBUG	0

#if DEBUG
#define cfprint(fmt, arg...)	{ char buf[64]; sprintf(buf, fmt, ##arg); lotol_write(1, buf, strlen(buf)); }
#else
#define cfprint(fmt, arg...)	{ }
#endif


#include <linux/assert.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>
#include <asm/pgtable.h>
#include <asm/ptrace.h>


#define NTRAPS	256		/* FIXME */
#define FIRST_EXTERNAL_VECTOR	0x20	/* see arch/i386/kernel/irq.h */

#define __CLI	0
#define __STI	1


/*
#define rdtsc_real(low,high) \
	__asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
*/

#if 1
static unsigned long long uml_tsc;
# define rdtsc_real(low,high) \
    do {				\
	uml_tsc+=90000;			\
	low=uml_tsc & 0x00000000ffffffffLL; \
	high= uml_tsc >> 32 ; 		\
    } while (0)

#else
# define rdtsc_real(low,high) \
  do { \
    struct timeval tv; \
    unsigned long long tsc; \
    lotol_gettimeofday(&tv,NULL); \
    tsc = tv.tv_usec + ((unsigned long long) tv.tv_sec) * 1000000; \
    tsc *= 1000; \
    low = tsc & 0x00000000ffffffffL; \
    high = tsc >> 32; \
  } while (0)
#endif

extern unsigned char umpic_test_irq(void);
extern unsigned char umpic_ack_irq(void);
#ifdef CONFIG_SMP
extern unsigned char umapic_test_irq(void);
extern unsigned char umapic_ack_irq(void);
#endif

struct stack {
	void *pretcode;
	int sig;
	void *pinfo;
	void *puc;

	/* siginfo is totally broken in the Linux kernel so it is totally useless for us
	 * I tried to use it for SIGIO to get the file descriptor.
	 * This file descriptor is set in fs/fcntl.c and overwritten in send_sig_info 
	 * kernel/signal.c.
	 * This was for 2.2.8 and 2.4.18 is also broken ( I tried it ... )
	 * --tg 17:20 2002-06-24
	 */

	struct { /* 128 byte */
		int si_signo;
		int si_errno;
		int si_code;

		int pad[(128 / sizeof(int)) - 3];
	} siginfo;

	/* struct ucontext { */
		unsigned long uc_flags;
		void *uc_link;
		struct /* stack_t */ {
			void *ss_sp;
			int ss_flags;
			int ss_size;
		} uc_stack;
		/* struct sigcontext { */
			unsigned int gs;
			unsigned int fs;
			unsigned int xes;
			unsigned int xds;
			unsigned long edi;
			unsigned long esi;
			unsigned long ebp;
			unsigned long esp;
			unsigned long ebx;
			unsigned long edx;
			unsigned long ecx;
			unsigned long eax;
			unsigned long trapno;
			unsigned long orig_eax;
			unsigned long eip;
			unsigned int xcs;                                                               unsigned long eflags;
			unsigned long esp_at_signal;
			unsigned int xss;
			void *fpstate;
			unsigned long oldmask;
			unsigned long cr2;
		/* } uc_mcontext; */
		struct /* sigset_t */ {
			unsigned long int __val[2];
		} uc_sigmask;
	/* } uc; */
	
	/*
	 * rest of structure is unknown!!
	 * (depends on kernel version)
	 *
	 * => don't use sizeof(struct stack)
	 * => don't use something like `*regs' or `regs + 1'!!
	 */

	struct /* _fpstate */ {
		unsigned long cw;
		/* ... */
	} _fpstate;
	char retcode[8];
	char dummy2[4];	/* 8-Byte aligned */
};

struct umcpu {
	/* special registers */
	unsigned long cr0;
	unsigned long cr1;
	unsigned long cr2;
	unsigned long cr3;
	unsigned long cr4;

	/* segment registers */
	unsigned long ss;
	unsigned long cs;
	unsigned long ds;
	unsigned long es;
	unsigned long fs;
	unsigned long gs;

	/* time stamp counter (of host!) */
	unsigned long tsc_low;
	unsigned long tsc_high;

	unsigned long esp0;
	int irqflag;
	unsigned long idt_limit;
	unsigned long idt_base;
	char stack[SZ_CPU - 17 * 4];	/* pad to SZ_CPU */
};

#define UMCPU	((struct umcpu *) VA_CPU)


/*
 * Fix a general protection resulting from a
 * - inb/inw/inl outb/outw/outl instruction
 * - cli/sti instruction
 *
 * Note:
 * we might execute in user mode (after doing ioperm/iopl)
 * or in kernel mode (without any ioperm/iopl)
 */
static int
fix_general_protection(struct stack *regs, unsigned long error_code)
{
	extern unsigned char inb(unsigned short port);
	extern void insb(unsigned short port, void *addr, unsigned long count);
	extern unsigned short inw(unsigned short port);
	extern void insw(unsigned short port, void *addr, unsigned long count);
	extern unsigned long inl(unsigned short port);
	extern void insl(unsigned short port, void *addr, unsigned long count);
	extern void outb(unsigned char value, unsigned short port);
	extern void outsb(unsigned short port, const void *addr, unsigned long count);
	extern void outw(unsigned short value, unsigned short port);
	extern void outsw(unsigned short port, const void *addr, unsigned long count);
	extern void outl(unsigned long value, unsigned short port);
	extern void outsl(unsigned short port, const void *addr, unsigned long count);
	unsigned char osp;
	unsigned char inst;
	unsigned char addr;


	if (user_mode(regs)) {
		if (((regs->eflags >> 12) < 3)
		 && (0)) {	/* FIXME VOSSI */
			/* iopl level too low */
			/* and io_bitmap set */
			printk("WARNING: %s got general protection (no iopl)\n",
					current->comm);
			return 0;
		}
	}

	inst = *(unsigned char *) regs->eip;
	if (inst == 0x66) {
		/* operand size prefix */
		osp = 1;
		inst = *(unsigned char *) (regs->eip + 1);
		addr = *(unsigned char *) (regs->eip + 2);
	} else {
		osp = 0;
		addr = *(unsigned char *) (regs->eip + 1);
	}

	switch (inst) {
	case 0x6C:	/* ins m8, dx */
	case 0x6D:	/* ins m16, dx */
	case 0x6E:	/* outs dx, m8 */
	case 0x6F:	/* outs dx, m16 */
		/* FIXME VOSSI */
		printk("%s: inst=0x%02x\n", __FUNCTION__, inst);
		return 0;

	case 0xE4:	/* in al, imm8 */
		regs->eax &= 0xffffff00;
		regs->eax |= inb((unsigned short) addr);
		regs->eip += 2;
		return 1;

	case 0xE5:	/* in (e)ax, imm8 */
		if (osp) {
			regs->eax &= 0xffff0000;
			regs->eax |= inw((unsigned short) addr);
		} else {
			regs->eax = inl((unsigned short) addr);
		}
		regs->eip += osp + 2;
		return 1;

	case 0xE6:	/* out imm8, al */
		outb((unsigned char) regs->eax, (unsigned short) addr);
		regs->eip += 2;
		return 1;

	case 0xE7:	/* out imm8, (e)ax */
		if (osp)
			outw((unsigned short) regs->eax, (unsigned short) addr);
		else
			outl((unsigned long) regs->eax, (unsigned short) addr);
		regs->eip += osp + 2;
		return 1;

	case 0xEC:	/* in al, dx */
		regs->eax &= 0xffffff00;
		regs->eax |= inb((unsigned short) regs->edx);
		regs->eip += 1;
		return 1;

	case 0xED:	/* in (e)ax, dx */
		if (osp) {
			regs->eax &= 0xffff0000;
			regs->eax |= inw((unsigned short) regs->edx);
		} else {
			regs->eax = inl((unsigned short) regs->edx);
		}
		regs->eip += osp + 1;
		return 1;

	case 0xEE:	/* out dx, al */
		outb((unsigned char) regs->eax, (unsigned short) regs->edx);
		regs->eip += 1;
		return 1;

	case 0xEF:	/* out dx, (e)ax */
		if (osp)
			outw((unsigned short) regs->eax, (unsigned short) regs->edx);
		else
			outl((unsigned long) regs->eax, (unsigned short) regs->edx);
		regs->eip += osp + 1;
		return 1;

	case 0xFA:	/* cli */
		UMCPU->irqflag = 0;
		regs->eip += 1;
		return 1;

	case 0xFB:	/* sti */
		UMCPU->irqflag = 1;
		regs->eip += 1;
		return 1;

	case 0x0F:      /* rdtsc */
          // STK added rdtsc fault handling
	  if(*((char *) (regs->eip+1)) == 0x31) {
	    rdtsc_real(regs->eax, regs->edx);
	    regs->eip += 2;
	    return 1;
	  } else {
	    return 0;
	  }	  
	default:
		/* FIXME VOSSI */
		printk("%s: inst=0x%02x at 0x%08lx\n", __FUNCTION__,
				inst, regs->eip);
		return 0;
	}
}


/*
 * regs->orig_eax:
 *	bit 0 == 0 means no page found, 1 means protection fault
 *	bit 1 == 0 means read, 1 means write
 *	bit 2 == 0 means kernel, 1 means user-mode
 */
static int
fix_page_fault(struct stack *regs, unsigned long error_code)
{
	extern char _text;
	extern char _etext;
	unsigned long addr;
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;
	unsigned long va;
	unsigned long flags;
	unsigned long pa;
	unsigned long prot;
	unsigned long ret;
	unsigned long fd;
	unsigned long offset;


	addr = regs->cr2;

	if (! (error_code & 4)
	 && (unsigned long) &_text <= addr && addr < (unsigned long) &_etext) {
		printk("WARNING: %s writes to 0x%lx at 0x%lx\n",
				current->comm, addr, regs->eip);
		ret = lotol_mprotect((char *) (addr & PAGE_MASK), PAGE_SIZE,
			LOTOL_PROT_READ | LOTOL_PROT_WRITE | LOTOL_PROT_EXEC);
		assert(0 <= ret);
		return 1;
	}

	pgd = (pgd_t *) __va(UMCPU->cr3) + (addr >> PGDIR_SHIFT);
	if (pgd_none(*pgd))
		return 0;	/* page directory not present */
	pmd = pmd_offset(pgd, addr);
	if (pmd_none(*pmd))
		return 0;	/* page directory not present */
	pte = pte_offset(pmd, addr);
	if (pte_none(*pte))
		return 0;	/* page not present */

	if (!pte_present(*pte))
		return 0;	/* page not present */
	if (!pte_write(*pte) && (error_code & 2))
		return 0;	/* no write permission */
	if (!pte_read(*pte) && (error_code & 4))
		return 0;	/* not in supervisor mode */

	va = addr & PAGE_MASK;
	flags = LOTOL_MAP_FIXED | LOTOL_MAP_SHARED;
	pa = pte_val(*pte) & PAGE_MASK;
	prot = LOTOL_PROT_READ | LOTOL_PROT_EXEC;
	if (pte_write(*pte))
		prot |= LOTOL_PROT_WRITE;

	if (PA_VIDEO_MEMORY_LOW <= pa
		&& pa < PA_VIDEO_MEMORY_LOW + SZ_VIDEO_MEMORY_LOW) {
		fd = FD_VIDEO_MEMORY;
		offset = pa - PA_VIDEO_MEMORY_LOW;

	} else
	if (PA_VIDEO_MEMORY_HIGH <= pa
		&& pa < PA_VIDEO_MEMORY_HIGH + 2*1024*1024) {
		/* video memory size is unknown at compile-time */
		/* should be re-written determining video size at run-time */
		/* FIXME VOSSI */
		fd = FD_VIDEO_MEMORY;
		offset = pa - PA_VIDEO_MEMORY_HIGH;

	} else
	if (PA_IO_APIC <= pa
		&& pa < PA_IO_APIC + SZ_IO_APIC) {
		/* FIXME HSHOEXER */
		fd = FD_IOAPIC_MEMORY;
		offset = pa - PA_IO_APIC;

	} else
	if (PA_LOCAL_APIC <= pa
		&& pa < PA_LOCAL_APIC + SZ_LOCAL_APIC) {
		/* FIXME HSHOEXER */
		fd = FD_APIC_MEMORY;
		offset = pa - PA_LOCAL_APIC;
	} else {
		/* MUST BE DEFAULT as size is unknown at compile-time */
		/* should be re-written determining memory size at run-time */
		/* FIXME VOSSI */
		fd = FD_MAIN_MEMORY;
		offset = pa - PA_MAIN_MEMORY_LOW;
		assert(offset == pa);
	}

	ret = (unsigned long) lotol_mmap((char *) va, PAGE_SIZE, prot,
			flags, fd, offset);
	assert(ret == va);

	return 1;
}


extern void umcore_scan_irqs_int3(void);
void
umcore_scan_irqs(void)
{
	cfprint("umcore_scan_irqs: called()\n");

	if (UMCPU->irqflag
	 && umpic_test_irq()) {
		asm(
		"	int3\n"
		"umcore_scan_irqs_int3:\n"
		);
	}

	cfprint("umcore_scan_irqs: return\n");
}


extern void umcore_iret_int3(void);
asm(
"umcore_iret: .globl umcore_iret\n"
"	int3\n"
"umcore_iret_int3:"
);


void
umcore_sighandler(int sig, void *si, void *uc)
{
	struct stack *regs = (struct stack *) ((char *) &sig - sizeof(void *));
	int vec;


	cfprint("umcore_sighandler: called()\n");

	/* sanity checks */
	assert((char *) UMCPU->stack < (char *) regs
	    && (char *) regs <= (char *) UMCPU->stack + sizeof(UMCPU->stack));
	assert((char *) regs < (char *) regs->pretcode
	    && (char *) regs->pretcode < (char *) UMCPU->stack + sizeof(UMCPU->stack));
	assert(sig == regs->sig);
	assert(si == regs->pinfo);
	assert(uc == regs->puc);
	assert(regs->pinfo == &regs->siginfo);
	assert(regs->puc == &regs->uc_flags);
	assert(regs->fpstate == (void *) 0
	    || regs->fpstate == &regs->_fpstate);
	
	/* vec must be on alternate stack */
	assert((char *) UMCPU->stack < (char *) &vec
	    && (char *) &vec < (char *) UMCPU->stack + sizeof(UMCPU->stack));
	/* mustn't be re-entered on alternate stack */
	if ((char *) UMCPU->stack <= (char *) regs->esp
	 && (char *) regs->esp <= (char *) UMCPU->stack + sizeof(UMCPU->stack)) {
		char buf[128];
		sprintf(buf, "regs->eip=0x%08lx regs->esp=0x%08lx, regs->sig=%d\n",
				regs->eip, regs->esp, regs->sig);
		lotol_write(1, buf, strlen(buf));
	}
	assert((char *) regs->esp < (char *) UMCPU->stack
	    || (char *) UMCPU->stack + sizeof(UMCPU->stack) < (char *) regs->esp);


	if (regs->sig == LOTOL_SIGUSR2
	 && regs->eip == (unsigned long) umcore_iret_int3) {

		/* --------------------------------- */
		/* return from exception / interrupt */
		/* --------------------------------- */

		unsigned long esp0;
		unsigned long eflags;
		unsigned long eip;
		unsigned long esp;

		cfprint("umcore_sighandler: iret\n");

		/*
		 * get esp0
		 */
		esp0 = regs->esp;

		/*
		 * pop registers from stack
		 */
		/* eip */
		eip = *(unsigned long *) esp0;
		esp0 += 4;

		/* cs */
		/* ignored - same as current segment */
		esp0 += 4;

		/* eflags */
		eflags = *(unsigned long *) esp0;
		esp0 += 4;
		
		if (eip < __PAGE_OFFSET) {
			/*
			 * returning to user mode
			 */
			/* esp */
			esp = *(unsigned long *) esp0;
			esp0 += 4;

			/* ss */
			/* ignored - same as current segment */
			esp0 += 4;
		} else {
			esp = -1;	/* not used */
		}

		/*
		 * set new registers
		 */
		regs->eflags = eflags | 1 << 9;
		UMCPU->irqflag = (eflags >> 9) & 1;
		regs->eip = eip;

		if (regs->eip < __PAGE_OFFSET) {
			regs->esp = esp;
			UMCPU->esp0 = esp0;
		} else {
			regs->esp = esp0;
		}

		vec = -1;
		goto rescan_irqs;
	}

	if (regs->sig == LOTOL_SIGSEGV && regs->trapno == 13) {
		int saved_irqflag;

		saved_irqflag = UMCPU->irqflag;
		UMCPU->irqflag = 0;

		if (fix_general_protection(regs, regs->orig_eax)) {
			/* inb/outb might raise interrupt */
			/* cli/sti might change interrupt enable flag */
			vec = -1;
			UMCPU->irqflag = saved_irqflag;
			goto rescan_irqs;
		}

		UMCPU->irqflag = saved_irqflag;
	}
	if (regs->sig == LOTOL_SIGSEGV && regs->trapno == 14) {
		unsigned long error_code;
		int saved_irqflag;

		saved_irqflag = UMCPU->irqflag;
		UMCPU->irqflag = 0;

		if (regs->eip < __PAGE_OFFSET) {
			error_code = regs->orig_eax | 4;
		} else {
			error_code = regs->orig_eax & ~4;
		}
		if (fix_page_fault(regs, error_code)) {
			/* interrupts possible? FIXME */
			vec = -1;
			UMCPU->irqflag = saved_irqflag;
			goto rescan_irqs;
		}

		UMCPU->irqflag = saved_irqflag;
	}

	switch (regs->sig) {
	case LOTOL_SIGSEGV:
		assert(regs->trapno == 4	/* overflow */
		    || regs->trapno == 5	/* bounds */
		    || regs->trapno == 7	/* device_not_available */
		    || regs->trapno == 8	/* double_fault */
		    || regs->trapno == 10	/* invalid_TSS */
		    || regs->trapno == 13	/* general_protection */
		    || regs->trapno == 14	/* page_fault */
		    || regs->trapno == 17	/* alignment_check */
		    || regs->trapno == 18	/* machine_check */
		    || regs->trapno == 19);	/* simd_coprocessor_error */
						/*	see below */
		vec = regs->trapno;
		break;

	case LOTOL_SIGFPE:
		assert(regs->trapno == 0	/* divide_error */
		    || regs->trapno == 9	/* coprocessor_segment_overrun*/
		    || regs->trapno == 16	/* coprocessor_error */
		    || regs->trapno == 19);	/* simd_coprocessor_error */
						/*	see above */
		vec = regs->trapno;
		break;

	case LOTOL_SIGBUS:
		assert(regs->trapno == 11	/* segment_not_present */
		    || regs->trapno == 12);	/* stack_segment */
		vec = regs->trapno;
		break;

	case LOTOL_SIGILL:
		assert(regs->trapno == 6);	/* invalid_op */
		vec = regs->trapno;
		break;

	case LOTOL_SIGPIPE:
		vec = 1;			/* debug */
		break;

	case LOTOL_SIGUSR1:
		vec = 0x80;			/* int $0x80 */
		break;

	case LOTOL_SIGUSR2:
		if (regs->eip == (unsigned long) umcore_scan_irqs_int3) {
			vec = -1;		/* re-scan irqs */
		} else {
			vec = 3;		/* int3 */
		}
		break;

	case LOTOL_SIGALRM: {
		extern void umpit_interrupt(void);
		extern void umrtc_interrupt(void);
#ifdef CONFIG_SMP
		extern void umapic_timerint(void);
#endif
		int saved_irqflag;

		saved_irqflag = UMCPU->irqflag;
		UMCPU->irqflag = 0;

		umpit_interrupt();
		umrtc_interrupt();
#ifdef CONFIG_SMP
		umapic_timerint();
#endif

		UMCPU->irqflag = saved_irqflag;

		vec = -1;
		break;
	    }
	case LOTOL_SIGHUP:
		/* FIXME */
		assert(0);
		vec = -1;
		break;

	case LOTOL_SIGIO: {
		extern void umkbd_interrupt(void);
		extern void umeth_interrupt(void);
		int saved_irqflag;

		saved_irqflag = UMCPU->irqflag;
		UMCPU->irqflag = 0;

		umeth_interrupt();
		umkbd_interrupt();

		UMCPU->irqflag = saved_irqflag;

		vec = -1;
		break;
	    }
	case LOTOL_SIGURG: {
		int saved_irqflag;

		saved_irqflag = UMCPU->irqflag;
		UMCPU->irqflag = 0;

		UMCPU->irqflag = saved_irqflag;

		vec = -1;
		break;
	    }
	case LOTOL_SIGVTALRM:
		/* FIXME */
		assert(0);
		vec = -1;
		break;

	default: {
		char buf[32];
		sprintf(buf, "got signal %d, trapno %d\n",
		(int) regs->sig, (int) regs->trapno);
		lotol_write(1, buf, strlen(buf));
		assert(0);
		vec = -1;
		/*NOTREACHED*/
	    }
	}

rescan_irqs:
	if (vec == -1 && UMCPU->irqflag) {
		/* scan pic irqs */
		if (umpic_test_irq()) {
			vec = umpic_ack_irq();
			assert(0x20 <= vec);
		}

#ifdef CONFIG_SMP
		/* scan apic irqs */
		if (umapic_test_irq()) {
			vec = umapic_ack_irq();
			assert(0x20 <= vec);
		}
#endif
	}

	if (vec == -1) {

		/* ------------------------------- */
		/* simulator (real) interrupt only */
		/* ------------------------------- */

		/* nothing more to do... */

		cfprint("umcore_sighandler: real interrupt only\n");
		return;
	}

	{
		/* ------------------------- */
		/* new exception / interrupt */
		/* ------------------------- */

		unsigned short desc0, desc1, desc2, desc3;
		unsigned long esp0;
		unsigned long eflags;
		unsigned short cs;
		unsigned long eip;
		unsigned long esp;

		cfprint("umcore_sighandler: exception/interrupt %d\n",
				vec);

		assert(0 <= vec && vec < NTRAPS);

		/*
		 * get interrupt/exception descriptor
		 */
		/* check idt_limit FIXME VOSSI */
		desc0 = *(unsigned short *) (UMCPU->idt_base + vec * 8 + 0);
		desc1 = *(unsigned short *) (UMCPU->idt_base + vec * 8 + 2);
		desc2 = *(unsigned short *) (UMCPU->idt_base + vec * 8 + 4);
		desc3 = *(unsigned short *) (UMCPU->idt_base + vec * 8 + 6);

		/*
		 * get esp0
		 */
		if (regs->eip < __PAGE_OFFSET) {
			/* coming from user mode */
			esp0 = UMCPU->esp0;
		} else {
			/* coming from kernel mode */
			esp0 = regs->esp;
		}
		assert(__PAGE_OFFSET <= esp0 && esp0 < __UML_ADDR_LIMIT);

		/*
		 * push old registers to stack
		 */
		if (regs->eip < __PAGE_OFFSET) {
			/* ss */
			esp0 -= 4;
			*(unsigned long *) esp0 = 0x2b;

			/* esp */
			esp0 -= 4;
			*(unsigned long *) esp0 = regs->esp;
		}

		/* eflags */
		esp0 -= 4;
		*(unsigned long *) esp0
			= (regs->eflags & ~(1 << 9)) | (UMCPU->irqflag << 9);

		/* cs */
		esp0 -= 4;
		if (regs->eip < __PAGE_OFFSET) {
			*(unsigned long *) esp0 = 0x23;
		} else {
			*(unsigned long *) esp0 = 0x10;
		}

		/* eip */
		esp0 -= 4;
		*(unsigned long *) esp0 = regs->eip;

		/* error_code */
		switch (vec) {
		case 0: /* divide_error */
		case 1: /* debug */
		case 2: /* nmi */
		case 3: /* int3 */
		case 4: /* overflow */
		case 5: /* bounds */
		case 6: /* invalid_op */
		case 9: /* coprocessor_segment_overrun */
		case 15: /* spurious_interrupt_bug */
		case 16: /* coprocessor_error */
		case 18: /* reserved */
			/* no error code */
			break;

		case 7: /* device_not_available */
			assert(0);	/* who knows? FIXME */
			break;

		case 8: /* double_fault */
		case 10: /* invalid_TSS */
		case 11: /* segment_not_present */
		case 12: /* stack_segment */
		case 13: /* general_protection */
		case 17: /* alignment_check */
			panic("Got general protection fault I can't handle");
			assert(0);	/* who knows? FIXME */
			break;

		case 14: /* page_fault */
			esp0 -= 4;
			if (regs->eip < __PAGE_OFFSET) {
				*(unsigned long *) esp0 = regs->orig_eax | 4;
			} else {
				*(unsigned long *) esp0 = regs->orig_eax & ~4;
			}
			UMCPU->cr2 = regs->cr2;
			break;

		default:
			/* no error code */
			assert(0x20 <= vec);
			break;
		}

		/*
		 * get new registers
		 */
		eflags = regs->eflags;
		if (! (desc2 & 0x0100)) {
			/* disable interrupts */
			eflags &= ~(1 << 9);
		}
		cs = desc1;
		eip = (desc0 << 0) | (desc3 << 16);
		/* ss = ? FIXME VOSSI */
		esp = esp0;

		/*
		 * set new registers
		 */
		regs->eflags = eflags | 1 << 9;
		UMCPU->irqflag = (eflags >> 9) & 1;
		/* UMCPU->cs = cs; FIXME VOSSI */
		regs->eip = eip;
		/* UMCPU->ss = ss; FIXME VOSSI */
		regs->esp = esp;

		return;
	}
}

void
umcpu_invlpg(unsigned long page)
{
	int ret;

	ret = lotol_munmap((char *) (page & 0xffff000), 4096);
	assert(0 <= ret);
}

void
umcpu_clts(void)
{
	/* FIXME VOSSI */
}

void
umcore_hlt(void)
{
	lotol_sigset_t mask;

	lotol_sigemptyset(&mask);
	lotol_rt_sigsuspend(&mask, LOTOL__NSIG / 8);
}

unsigned long
umcore_save_flags(void)
{
	return UMCPU->irqflag;
}

void
umcore_restore_flags(unsigned long flags)
{
	assert(flags == 0 || flags == 1);

	UMCPU->irqflag = flags;
}

void
umcore_sti(void)
{
	UMCPU->irqflag = 1;
}

void
umcore_cli(void)
{
	UMCPU->irqflag = 0;
}

void
umcpu_cr0_set(unsigned long val)
{
	UMCPU->cr0 = val;	/* FIXME */
}

unsigned long
umcpu_cr0_get(void)
{
	return UMCPU->cr0;
}

#if 0	/* not used FIXME */
void
umcpu_cr1_set(unsigned long val)
{
	UMCPU->cr1 = val;	/* FIXME */
}
#endif	/* not used FIXME */

void
umcpu_cr2_set(unsigned long val)
{
	UMCPU->cr2 = val;	/* FIXME */
}

unsigned long
umcpu_cr2_get(void)
{
	return UMCPU->cr2;
}

void
umcpu_cr3_set(unsigned long val)
{
	int ret;

//	ret = lotol_munmap(0, 0x70000000);
	ret = lotol_munmap(0, __PAGE_OFFSET);
	assert(0 <= ret);

	UMCPU->cr3 = val;
}

static int my_page_fixup(unsigned long addr) {
  pgd_t *pgd;
  pmd_t *pmd;
  pte_t *pte;
  unsigned long va;
  unsigned long flags;
  unsigned long pa;
  unsigned long prot;
  unsigned long ret;
  unsigned long fd;
  unsigned long offset;
  
  
  /*
  if (! (error_code & 4)
      && (unsigned long) &_text <= addr && addr < (unsigned long) &_etext) {
    printk("WARNING: %s writes to 0x%lx at 0x%lx\n",
	   current->comm, addr, regs->eip);
    ret = lotol_mprotect((char *) (addr & PAGE_MASK), PAGE_SIZE,
			 LOTOL_PROT_READ | LOTOL_PROT_WRITE | LOTOL_PROT_EXEC);
    assert(0 <= ret);
    return 1;
  }
  */
  
  pgd = (pgd_t *) __va(UMCPU->cr3) + (addr >> PGDIR_SHIFT);
  if (pgd_none(*pgd))
    return 0;	/* page directory not present */
  pmd = pmd_offset(pgd, addr);
  if (pmd_none(*pmd))
    return 0;	/* page directory not present */
  pte = pte_offset(pmd, addr);
  if (pte_none(*pte))
    return 0;	/* page not present */
    
  //if (!pte_present(*pte))
  //  return 0;	/* page not present */
  //if (!pte_write(*pte) && (error_code & 2))
  //  return 0;	/* no write permission */
  //if (!pte_read(*pte) && (error_code & 4))
  //  return 0;	/* not in supervisor mode */
  

  va = addr & PAGE_MASK;
  flags = LOTOL_MAP_FIXED | LOTOL_MAP_SHARED;
  pa = pte_val(*pte) & PAGE_MASK;
  prot = LOTOL_PROT_READ | LOTOL_PROT_EXEC;
  if (pte_write(*pte))
    prot |= LOTOL_PROT_WRITE;
  
  if (PA_VIDEO_MEMORY_LOW <= pa
      && pa < PA_VIDEO_MEMORY_LOW + SZ_VIDEO_MEMORY_LOW) {
    fd = FD_VIDEO_MEMORY;
    offset = pa - PA_VIDEO_MEMORY_LOW;
    
  } else
    if (PA_VIDEO_MEMORY_HIGH <= pa
	&& pa < PA_VIDEO_MEMORY_HIGH + 2*1024*1024) {
      /* video memory size is unknown at compile-time */
      /* should be re-written determining video size at run-time */
      /* FIXME VOSSI */
      fd = FD_VIDEO_MEMORY;
      offset = pa - PA_VIDEO_MEMORY_HIGH;
      
    } else
      if (PA_IO_APIC <= pa
	  && pa < PA_IO_APIC + SZ_IO_APIC) {
	/* FIXME HSHOEXER */
	fd = FD_IOAPIC_MEMORY;
	offset = pa - PA_IO_APIC;
	
      } else
	if (PA_LOCAL_APIC <= pa
	    && pa < PA_LOCAL_APIC + SZ_LOCAL_APIC) {
	  /* FIXME HSHOEXER */
	  fd = FD_APIC_MEMORY;
	  offset = pa - PA_LOCAL_APIC;
	} else {
	  /* MUST BE DEFAULT as size is unknown at compile-time */
	  /* should be re-written determining memory size at run-time */
	  /* FIXME VOSSI */
	  fd = FD_MAIN_MEMORY;
	  offset = pa - PA_MAIN_MEMORY_LOW;
	  assert(offset == pa);
	}
  
  ret = (unsigned long) lotol_mmap((char *) va, PAGE_SIZE, prot,
				   flags, fd, offset);
  assert(ret == va);
  
  return 1;
}

// STK this is so we can keep track of the current cr3 and
// guestpgd
static struct guestpgd_struct *currpgd = NULL;

struct guestpgd_struct *umcpu_currpgd_get(void) {
  return currpgd;
}

void
umcpu_cr3_switch(unsigned long new_cr3, struct guestpgd_struct * nextpgd)
{

#ifdef CONFIG_UMLINUX_FAST_CTX
  unsigned long addr;
  UMCPU->cr3 = new_cr3;
  currpgd = nextpgd;
  lotol_switchguest(nextpgd->pgd_id);
  /*
  lotol_munmap(0,0x70000000);
  for(addr=0;addr<0x70000000;addr+=4096) {
    my_page_fixup(addr);
  }
  */
#else
  umcpu_cr3_set(new_cr3);
#endif

}

unsigned long
umcpu_cr3_get(void)
{
	return UMCPU->cr3;
}

#if 0	/* not used FIXME */
void
umcpu_cr4_set(unsigned long val)
{
	UMCPU->cr4 = val;	/* FIXME */
}
#endif	/* not used FIXME */

unsigned long
umcpu_cr4_get(void)
{
	return UMCPU->cr4;
}

void
umcpu_lidt(unsigned long idt)
{
	UMCPU->idt_limit = *(unsigned short *) (idt + 0);
	UMCPU->idt_base = *(unsigned long *) (idt + 2);
}

unsigned long
umcpu_ds_get(void)
{
	return UMCPU->ds;
}

void
umcpu_ds_set(unsigned long ds)
{
	UMCPU->ds = ds;
}

unsigned long
umcpu_es_get(void)
{
	return UMCPU->es;
}

void
umcpu_es_set(unsigned long es)
{
	UMCPU->es = es;
}

unsigned long
umcpu_fs_get(void)
{
	return UMCPU->fs;
}

void
umcpu_fs_set(unsigned long fs)
{
	UMCPU->fs = fs;
}

unsigned long
umcpu_gs_get(void)
{
	return UMCPU->gs;
}

void
umcpu_gs_set(unsigned long gs)
{
	UMCPU->gs = gs;
}

void
umcpu_tsc_get(unsigned long *tsc_low, unsigned long *tsc_high)
{
	unsigned long cur_tsc_low;
	unsigned long cur_tsc_high;

	/* XXX put this in something like umlib.c */
	extern void sub64(unsigned long *, unsigned long *,
			  unsigned long *, unsigned long *);

	rdtsc_real(cur_tsc_low, cur_tsc_high);

	sub64(&cur_tsc_low, &cur_tsc_high, &UMCPU->tsc_low, &UMCPU->tsc_high);

	*tsc_low = cur_tsc_low;
	*tsc_high = cur_tsc_high;
}

void
umcpu_tsc_low_get(unsigned long *tsc_low)
{
	unsigned long tsc_high;

	umcpu_tsc_get(tsc_low, &tsc_high);
}

void
umcpu_tsc_high_get(unsigned long *tsc_high)
{
	unsigned long tsc_low;

	umcpu_tsc_get(&tsc_low, tsc_high);
}

void
umcpu_tsc_highlow_get(unsigned long long *tsc)
{
	unsigned long tsc_low;
	unsigned long tsc_high;

	umcpu_tsc_get(&tsc_low, &tsc_high);

	*tsc = tsc_high;
	*tsc <<= 32;
	*tsc |= tsc_low;
}

void
umcpu_msr_write(int msr, unsigned long val_low, unsigned long val_high)
{
	switch(msr) {
	case 0x10:
		/* asume reset of tsc */
		assert(val_low == 0);

		rdtsc_real(UMCPU->tsc_low, UMCPU->tsc_high);
		break;

	default: {
		char buf[64];
		sprintf(buf, "%d wrote to msr 0x%x\n", lotol_getpid(), msr);
		lotol_write(2, buf, strlen(buf));
		 }
	}
}

void
umcpu_msr_read(int msr, unsigned long *val_low, unsigned long *val_high)
{
	switch(msr) {
	case 0x10:
		/* read tsc (that's the same as "rdtsc"!) */
		umcpu_tsc_get(val_low, val_high);
		break;

	default: {
		char buf[64];
		sprintf(buf, "%d read from msr 0x%x\n", lotol_getpid(), msr);
		lotol_write(2, buf, strlen(buf));
		 }
	}
}


void
umcore_init(void)
{
	int sig;
	int ret;

	assert(sizeof(struct umcpu) == SZ_CPU);

	/*
	 * setup signal stack
	 */
	{
		lotol_stack_t ss;

		ss.ss_sp = UMCPU->stack;
		ss.ss_flags = 0;
		ss.ss_size = sizeof(UMCPU->stack);
		ret = lotol_sigaltstack(&ss, (lotol_stack_t *) 0);
		assert(0 <= ret);
	}

	/*
	 * setup signal handler
	 */
	for (sig = 1; sig <= LOTOL__NSIG; sig++) {
		struct lotol_sigaction sa;

		if (sig != LOTOL_SIGALRM	/* interrupts */
		 && sig != LOTOL_SIGHUP
		 && sig != LOTOL_SIGIO
		 && sig != LOTOL_SIGURG
		 && sig != LOTOL_SIGVTALRM
		 && sig != LOTOL_SIGFPE		/* exceptions */
		 && sig != LOTOL_SIGPIPE
		 && sig != LOTOL_SIGUSR2
		 && sig != LOTOL_SIGSEGV
		 && sig != LOTOL_SIGILL
		 && sig != LOTOL_SIGBUS
		 && sig != LOTOL_SIGUSR1) {	/* system call */
			continue;	/* FIXME */
		}

		sa.sa_sigaction = umcore_sighandler;
		sa.sa_flags = LOTOL_SA_SIGINFO | LOTOL_SA_ONSTACK ;
		lotol_sigemptyset(&sa.sa_mask);
		lotol_sigaddset(&sa.sa_mask, LOTOL_SIGALRM);
		lotol_sigaddset(&sa.sa_mask, LOTOL_SIGIO);
		lotol_sigaddset(&sa.sa_mask, LOTOL_SIGHUP);
		lotol_sigaddset(&sa.sa_mask, LOTOL_SIGURG);
		lotol_sigaddset(&sa.sa_mask, LOTOL_SIGVTALRM);

		ret = lotol_rt_sigaction(sig, &sa, (struct lotol_sigaction *) 0,
				LOTOL__NSIG / 8);
		assert(0 <= ret);
	}

	/*
	 * setup signal mask
	 */
	{
		lotol_sigset_t mask;

		lotol_sigemptyset(&mask);
		ret = lotol_sigprocmask(SIG_SETMASK, &mask, (lotol_sigset_t *) 0);
		assert(0 <= ret);
	}

	/*
	 * init registers
	 */
	rdtsc_real(UMCPU->tsc_low, UMCPU->tsc_high);
}
