#include "umide.h"
/* $Id: umide.c,v 1.1.1.1 2003/05/27 21:45:37 root Exp $ */


/* used for debuging and unsupported GPCMD's*/
static void
report_command(int nr)
{
	unsigned int lo = 0, mid, hi = ARY_LEN(umide_gpcmd_texts);
	const char *s = NULL;

	while (lo < hi) {
		mid = (lo + hi) / 2;
		if (umide_gpcmd_texts[mid].packet_command == ide[nr].buf[0]) {
			s = umide_gpcmd_texts[mid].text;
			break;
		}
		if (ide[nr].buf[0] < umide_gpcmd_texts[mid].packet_command)
			hi = mid;
		else
			lo = mid + 1;
	}
	{
		/* FIXME: put this into uml_?printf functions */
		char buf[90];
		sprintf(buf,
			 "%28s : 0x%02x% 4d% 4d% 4d% 4d% 4d% 4d% 4d% 4d% 4d% 4d% 4d\n",
			 s, ide[nr].buf[0], ide[nr].buf[1], ide[nr].buf[2],
			 ide[nr].buf[3], ide[nr].buf[4], ide[nr].buf[5],
			 ide[nr].buf[6], ide[nr].buf[7], ide[nr].buf[8],
			 ide[nr].buf[9], ide[nr].buf[10], ide[nr].buf[11]);
		lotol_write_wrapper(1, buf, strlen(buf));
	}
}

static void
umide_gen_irq(int nr)
{
	if ((ide[nr].control & 2) == 0) {
		void call_isa_irq(int irq);
		call_isa_irq(umide_irq[nr]);
	}
}

static void
umide_restore(int nr)
{				/* 0x10 */
	assert(ide[nr].command == WIN_RESTORE);
	ide[nr].status = READY_STAT;
	ide[nr].command = 0x00;
	UMIDE_GEN_IRQ;
}

static void
umide_read(int nr)
{				/* 0x20 */
	int unit = (ide[nr].select >> 4) & 1;
	unsigned long sector;
	unsigned long cyl;
	unsigned long head;
	long pos;
	unsigned long count;
	int ret;
	loff_t res;
	unsigned char readStore[SECTOR_SIZE*2];
	unsigned char *readBuf;

	// STK check it out to make sure the buffer is sector aligned
	unsigned long addr = (unsigned long) readStore;
	addr += SECTOR_SIZE;
	addr &= ~(SECTOR_SIZE-1);
	readBuf = (unsigned char *) addr;
      

	/* FIXME VOSSI */
	sector = (unsigned long) ide[nr].sector - 1;
	assert(0 <= sector && sector < ide[nr].unit[unit].phys_sectors);
	head = (unsigned long) ide[nr].select & 0x0f;
	assert(0 <= head && head < ide[nr].unit[unit].phys_heads);
	cyl = (unsigned long) ide[nr].lcyl + (unsigned long) ide[nr].hcyl * 256;
	assert(0 <= cyl && cyl < ide[nr].unit[unit].phys_cyls);
	pos = (cyl * ide[nr].unit[unit].phys_heads + head)
	    * ide[nr].unit[unit].phys_sectors + sector;
	
	ret = lotol__llseek_wrapper(
			ide[nr].unit[unit].fd,
			(pos >> 23) &0xffffffff,
			(pos << 9)  &0xffffffff,
			&res,
			SEEK_SET);
	assert( -1 != ret );

	count = 512;
	//ret = lotol_read_wrapper(ide[nr].unit[unit].fd, ide[nr].buf, count);
	ret = lotol_read_wrapper(ide[nr].unit[unit].fd, readBuf, count);
	memcpy(ide[nr].buf,readBuf,count);

	assert(ret < 0 || ret == count);
	if (ret < 0) {
	  printk("did not read the ide disk, ret = %d, errno = %d\n",ret,errno);
	  printk("res = %lld, count = %ld buf = 0x%08x\n",res,count,
		 (unsigned int) ide[nr].buf);
		/* fault injective */
		ide[nr].status = ERR_STAT;
		ide[nr].error = 1 << 6;	/* UNC = Uncorrectable data error 
					   Schmidt page 59/60 */
	} else {
		ide[nr].status = READY_STAT;
	}

	ide[nr].tail = 0;
	ide[nr].head = count;
	ide[nr].count = count;
	ide[nr].status |= DRQ_STAT;

	UMIDE_GEN_IRQ;
}

static void
umide_read_end(int nr)
{
	int unit = (ide[nr].select >> 4) & 1;

	ide[nr].nsector--;
	ide[nr].sector++;
	if (ide[nr].sector > ide[nr].unit[unit].phys_sectors) {
		ide[nr].sector = 1;
		ide[nr].select = (ide[nr].select & 0xf0)
		    | (((ide[nr].select & 0x0f) + 1) & 0x0f);

		if ((ide[nr].select & 0x0f) == ide[nr].unit[unit].phys_heads) {
			ide[nr].select = ide[nr].select & 0xf0;
			ide[nr].lcyl++;

			if (ide[nr].lcyl == 0)
				ide[nr].hcyl++;
		}
	}
	if (0 < ide[nr].nsector)
		umide_read (nr);
	else {
		ide[nr].status = READY_STAT;
		ide[nr].command = 0x00;
	}
}

static void
umide_write(int nr)
{				/* 0x30 */
	ide[nr].tail = 0;
	ide[nr].head = 0;
	ide[nr].count = 512;
	ide[nr].status = READY_STAT | DRQ_STAT | SEEK_STAT;
}

static void
umide_write_end(int nr)
{
	int unit = (ide[nr].select >> 4) & 1;
	unsigned long sector;
	unsigned long cyl;
	unsigned long head;
	long pos;
	unsigned long count;
	int ret;
	loff_t res;
	unsigned char writeStore[SECTOR_SIZE*2];
	unsigned char *writeBuf;

	// STK check it out to make sure the buffer is sector aligned
	unsigned long addr = (unsigned long) writeStore;
	addr += SECTOR_SIZE;
	addr &= ~(SECTOR_SIZE-1);
	writeBuf = (unsigned char *) addr;

	sector = (unsigned long) ide[nr].sector - 1;
	assert(0 <= sector && sector < ide[nr].unit[unit].phys_sectors);
	head = (unsigned long) ide[nr].select & 0x0f;
	assert(0 <= head && head < ide[nr].unit[unit].phys_heads);
	cyl = (unsigned long) ide[nr].lcyl + (unsigned long) ide[nr].hcyl * 256;
	assert(0 <= cyl && cyl < ide[nr].unit[unit].phys_cyls);
	pos = (cyl * ide[nr].unit[unit].phys_heads + head)
	    * ide[nr].unit[unit].phys_sectors + sector;

	ret = lotol__llseek_wrapper(
			ide[nr].unit[unit].fd,
			(pos >> 23) &0xffffffff,
			(pos << 9)  &0xffffffff,
			&res,
			SEEK_SET);
	assert ( -1 != ret );

	count = 512;
	memcpy(writeBuf,ide[nr].buf,count);
	//ret = lotol_write_wrapper(ide[nr].unit[unit].fd, ide[nr].buf, count);
	ret = lotol_write_wrapper(ide[nr].unit[unit].fd, writeBuf, count);
	assert(ret < 0 || ret == count);	/* ignore errors => fault inject */

	ide[nr].nsector--;
	ide[nr].sector++;
	if (ide[nr].unit[unit].phys_sectors < ide[nr].sector) {
		ide[nr].sector = 1;
		ide[nr].select = (ide[nr].select & 0xf0)
		    | (((ide[nr].select & 0x0f) + 1) & 0x0f);

		if ((ide[nr].select & 0x0f) == ide[nr].unit[unit].phys_heads) {
			ide[nr].select = ide[nr].select & 0xf0;
			ide[nr].lcyl++;
			if (ide[nr].lcyl == 0)
				ide[nr].hcyl++;
		}
	}
	if (0 < ide[nr].nsector) {
		umide_write (nr);
		UMIDE_GEN_IRQ;
	} else {
		ide[nr].status = READY_STAT | SEEK_STAT;
		ide[nr].command = 0x00;
		UMIDE_GEN_IRQ;
	}
}

static void
umide_specify(int nr)
{				/* 0x91 */
	assert(ide[nr].command == WIN_SPECIFY);
	ide[nr].status = READY_STAT;
	ide[nr].command = 0x00;
	UMIDE_GEN_IRQ;
}

static void
probe_drives(int nr)
{
	/* probe master drive */
	char drive = nr * 2 + UNIT0;

	ide[nr].active |= (1 << (UNIT0 + 2));	/* set probed flag */

	if (-1 != lotol_lseek_wrapper(FD_HD_CONFIG(drive), 0, SEEK_SET))
		/* is device active? */
		ide[nr].active |= (1 << UNIT0);

	/* probe slave drive */
	drive = nr * 2 + UNIT1;

	ide[nr].active |= (1 << (UNIT1 + 2));	/* set probed flag */

	if (-1 != lotol_lseek_wrapper(FD_HD_CONFIG(drive), 0, SEEK_SET))
		/* is device active? */
		ide[nr].active |= (1 << UNIT1);
}

/*
 * See if the file is at least "size" sectors.  Return 1 if file is at least
 * "size" sectors, 0 otherwise.
 */
static int testSize(int fd, loff_t size) {
    int status;
    char buf[SECTOR_SIZE * 2];
    char *buf1;

    size *= 512;

    /* used sector-aligned buffer */
    buf1 = (char *) (((unsigned int) buf + SECTOR_SIZE) & ~(SECTOR_SIZE - 1));

    /* printf("%d, %d\n", (int) (size>>32), (int) (size & 0xffffffff)); */

    status = lotol__llseek_wrapper(fd, (int) (size >> 32), (int) (size & 0xffffffff),
                        &size, SEEK_SET);
    if (status < 0) {
      //printf("%lld bytes: no\n", size);
	return(0);
    }

    status = lotol_read(fd, buf1, SECTOR_SIZE);

    if (status == SECTOR_SIZE) {
      //printf("%lld bytes: yes\n", size);
    } else {
      //printf("%lld bytes: no\n", size);
    }
    /*
    */

    return(status == SECTOR_SIZE);
}


static void
umide_identify(int nr)
{				/* 0xec */
	int fd, ret, unit = (ide[nr].select >> 4) & 1;
	char drive = nr * 2 + unit;
	char config[10], model[20];
	loff_t res;
	uint   len;
	struct hd_driveid *id = (struct hd_driveid *) ide[nr].buf;

	if (0 != lotol_lseek_wrapper(FD_HD_CONFIG(drive), 0, SEEK_SET)) { /* not configured */
		ide[nr].identify = 0;
		if ((1 == ((ide[nr].active >> (2 + UNIT0)) & 1))
		    || (1 == ((ide[nr].active >> (2 + UNIT1)) & 1))
		    )		/* only generate irq if controller exists */
			UMIDE_GEN_IRQ;
		return;
	}

	memset(id, 0, sizeof(*id));

	/* get config */
	if ((fd = FD_HD_CONFIG(drive)) < 0)
		panic("hd%c/config: can't open: errno=%d\n", HDX, errno);

	if ((ret = lotol_lseek_wrapper(fd, 0, SEEK_SET)) < 0)
		panic("hd%c/config: can't lseek: errno=%d\n", HDX, errno);

	if (-1 == (ret = lotol_read_wrapper(fd, config, sizeof(config) - 1)))
		panic("hd%c/config: can't read: errno=%d\n", HDX, errno);

	config[ret - 1] = '\0';

	if ((0 == strncmp(config, "disk", sizeof("disk"))) &&
			(ide[nr].command == WIN_IDENTIFY))
		id->config = ide_disk << 8;
	else if ((0 == strncmp(config, "cdrom", sizeof("cdrom"))) &&
			(ide[nr].command == WIN_PIDENTIFY))
		id->config = ide_cdrom << 8;
	else {			/* we only support disk _and_ cdrom */
		ide[nr].identify = 0;
		if ((1 == ((ide[nr].active >> (2 + UNIT0)) & 1))
		    || (1 == ((ide[nr].active >> (2 + UNIT1)) & 1))
		    )		/* only generate irq if controller exists */
			UMIDE_GEN_IRQ;
		return;
	}

	/* get model */
	if ((fd = FD_HD_MODEL(drive)) < 0)
		panic("hd%c/model: can't open: errno=%d\n", HDX, errno);

	if ((ret = lotol_lseek_wrapper(fd, 0, SEEK_SET)) < 0)
		panic("hd%c/model: can't lseek: errno=%d\n", HDX, errno);

	if (-1 == (ret = lotol_read_wrapper(fd, model, sizeof(model) - 1)))
		panic("hd%c/model: can't read: errno=%d\n", HDX, errno);

#if 0 /* granola: memset(id->model, 0, 20); read spec ... has to be padded with \0
	 ATTENTION! 40 chars ... read the spec
	 */
	 for ( char *p = model; *p != 20; *p++ ){ }
#endif

	/* open media */
	if ((fd = FD_HD_MEDIA(drive)) < 0)
		panic("hd%c/media: can't open: errno=%d\n", HDX, errno);

	if ( -1 == lotol__llseek_wrapper(fd, 0, 0, &res, SEEK_END))
		panic("hd%c/media: can't lseek: errno=%d\n", HDX, errno);

	  
	  
	//pmchen, get the size of a raw disk
	if(res == 0) {
	  loff_t low, high;
	  /*
	   * Keep doubling size until it's larger than the file
	   */
	  for (low = 1; testSize(fd, low*2); low *= 2);
	  high = low * 2;
	  
	  /*
	   * Find the exact size
	   */
	  while (high - low > 1) {
	    loff_t try = low + (high - low) /2;
	    if (testSize(fd, try)) {
	      low = try;
	    } else {
	      high = try;
	    }
	  }
	  res = (low + 1) * SECTOR_SIZE;
	}
	
	
	
	  assert(res != 0);	/* media size can't be zero! --tg 16:14 2002-06-05 */
	  
	  
	  if ( id->config == ide_cdrom << 8 ){ /* this is also verified by lilo in umide.c */
	    assert( !(res & 0x3ff )); /* FIXME support only CDROMS with UMIDE_BS */
	  } else { /* disk */
	    assert( !(res & 0xff ));    /* res % 512 must be zero */
	  }
	  
	  len = res >> 9;
	  
	  ide[nr].unit[unit].phys_sectors = (63 < len) ? 63 : len;
	  len = (len + ide[nr].unit[unit].phys_sectors - 1) /
	    ide[nr].unit[unit].phys_sectors;
	  ide[nr].unit[unit].phys_heads = (15 < len) ? 15 : len;
	  len = len / ide[nr].unit[unit].phys_heads;
	  ide[nr].unit[unit].phys_cyls = len;
	  
	  ide[nr].unit[unit].fd = fd;
	  
	  /* gepadded auf 20 Zeichen byteorder ABCD -> BADC */
	  strcpy(id->model, "MUiLun xDI E         ");
	  id->sectors = ide[nr].unit[unit].phys_sectors;
	  id->heads = ide[nr].unit[unit].phys_heads;
	  id->cyls = ide[nr].unit[unit].phys_cyls;
	  id->cur_sectors = ide[nr].unit[unit].phys_sectors;
	  id->cur_heads = ide[nr].unit[unit].phys_heads;
	  id->cur_cyls = ide[nr].unit[unit].phys_cyls;
	  
	  ide[nr].tail = 0;
	  ide[nr].head = 512;
	  ide[nr].count = 512;
	  ide[nr].status = READY_STAT | DRQ_STAT;
	  
	  ide[nr].identify = 1;
	  
	  UMIDE_GEN_IRQ;
}

static void
umide_identify_end(int nr)
{
	if (!ide[nr].identify)
		return;		/* we don't exist */
	ide[nr].status = READY_STAT;
	ide[nr].command = 0x00;
}

static void
umide_gpcmd(int nr)
{				/* 0xA0 */
	ide[nr].nsector = 0x1;	/* set i'command or data' delete rest */
	ide[nr].status = DRQ_STAT;	/* set command or data delete rest */
	ide[nr].count = 12;	/* gimme ATAPI COMMAND */
	ide[nr].head = 0;
	/* no irq here */
}

static void
umide_gpcmd_mode_sense_10(int nr)
{				/* 0x5a */

	struct atapi_capabilities_page *cap =
		(struct atapi_capabilities_page *) ide[nr].buf;

	/* Datenlaenge 24 byte && Seite 0x2a */
	assert((ide[nr].buf[8] == 0x18) && (ide[nr].buf[2] == 0x2a));

	memset(cap, 0, sizeof(*cap));

	cap->header.mode_data_length = htons(8 + 24 - 2);
	cap->header.medium_type = 1;

	cap->page_length = 20;

	cap->cd_r_read = 1;
	cap->cd_rw_read = 1;
	cap->method2 = 1;

	cap->audio_play = 1;
	cap->mode2_form1 = 1;
	cap->mode2_form2 = 1;

	cap->cdda = 1;
	cap->isrc = 1;
	cap->upc = 1;

	cap->lock = 1;
	cap->eject = 1;
	cap->mechtype = mechtype_tray;

	cap->separate_volume = 1;
	cap->separate_mute = 1;
	cap->disc_present = 1;

	cap->maxspeed = htons(7056);
	cap->n_vol_levels = htons(255);
	cap->buffer_size = htons(256);

	ide[nr].nsector = 0x2;	/* set IO && delete RESET */
	ide[nr].status = DRQ_STAT;
	ide[nr].tail = 0;
	ide[nr].count = ide[nr].lcyl = 8 + 24;
	ide[nr].pcdone = 2;	/* prepare for packet command with data transfer */

	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_read_toc_pma_atip(int nr)
{				/* 0x43 */
	struct atapi_toc_header *hdr = (struct atapi_toc_header *) ide[nr].buf;
	struct atapi_toc_entry *entr_1 =
		(struct atapi_toc_entry *) &ide[nr].buf[4];
	struct atapi_toc_entry *entr_2 =
		(struct atapi_toc_entry *) &ide[nr].buf[12];
	int ret = 0;
	int unit = (ide[nr].select >> 4) & 1;
	loff_t res;

	assert(0x02 == ide[nr].buf[1]);	/* REPORT MSF DATA! */

	ide[nr].count = ide[nr].lcyl = ide[nr].buf[8];	/* angeforderte Laenge */

	if (0x40 == ide[nr].buf[9])	/* Multi Saisson Anfrage mit control 0x40 */
		hdr->toc_length = htons(10);
	else
		hdr->toc_length = htons(18);	/* (4-2) + 2*8 */

	memset(entr_1, 0, 2 * sizeof(*entr_1));	/* toc_length already written */

	hdr->first_track = hdr->last_track = 1;

	entr_1->control = entr_2->control = 4;
	entr_1->adr = entr_2->adr = 1;

	entr_1->track = 1;
	entr_2->track = 170;

	entr_1->addr.msf.second = 2;

	ret = lotol__llseek_wrapper(ide[nr].unit[unit].fd, 0, 0, &res, SEEK_END);

	assert( -1 != ret );

	lba_to_msf(res >> UMIDE_BS, &entr_2->addr.msf.minute,
		    &entr_2->addr.msf.second, &entr_2->addr.msf.frame);

	ide[nr].nsector = 0x2;	/* set IO && delete RESET */
	ide[nr].status = DRQ_STAT;
	ide[nr].tail = 0;
	ide[nr].pcdone = 2;	/* prepare for packet command with data transfer */

	UMIDE_GEN_IRQ;

}

static void
umide_gpcmd_read_10(int nr)
{				/* 0x28 */
	struct lba {
		unsigned lba;
	}  *tmp_lba        = (struct lba *) &ide[nr].buf[2];

	struct count {
		unsigned short count;
	}    *tmp_count    = (struct count *) &ide[nr].buf[7];

	int ret            = 0;
	int unit           = (ide[nr].select >> 4) & 1;
	unsigned int lba   = ntohl(tmp_lba->lba);
	unsigned int count = ntohs(tmp_count->count) << UMIDE_BS;
	loff_t res;

	assert(ide[nr].buf[1] == 0);	/* no relative addressing */

	if (UMIDE_MAX < count) {	/* we can only transfer 63488 bytes at once */
		ide[nr].lcyl = (UMIDE_MAX >> 16) & 0xff;
		ide[nr].hcyl = (UMIDE_MAX >> 8) & 0xff;
		ide[nr].count = UMIDE_MAX;
		ide[nr].m2t = count - UMIDE_MAX;
	} else {
		ide[nr].lcyl = (count >> 16) & 0xff;
		ide[nr].hcyl = (count >> 8) & 0xff;
		ide[nr].count = count;
	}

	ret = lotol__llseek_wrapper(
			ide[nr].unit[unit].fd,
			(lba >> (32 - UMIDE_BS)) &0xffffffff,
			(lba << UMIDE_BS)        &0xffffffff,
			&res,
			SEEK_SET);

	assert( -1 != ret );

	ret =
	    lotol_read_wrapper(ide[nr].unit[unit].fd, ide[nr].buf,
				ide[nr].count);
	assert(-1 != ret);

	ide[nr].nsector = 0x2;	/* set IO && delete RESET */
	ide[nr].status = DRQ_STAT;
	ide[nr].tail = 0;
	ide[nr].pcdone = 2;	/* prepare for packet command with data transfer */

	UMIDE_GEN_IRQ;
}

static void
umide_continue_gpcmd_read_10(int nr)
{				/* 0x28 */
	int ret = 0;
	int unit = (ide[nr].select >> 4) & 1;
	if (UMIDE_MAX < ide[nr].m2t) {
		ide[nr].lcyl = (UMIDE_MAX >> 16) & 0xff;
		ide[nr].hcyl = (UMIDE_MAX >> 8) & 0xff;
		ide[nr].count = UMIDE_MAX;
		ide[nr].m2t -= UMIDE_MAX;
	} else {
		ide[nr].lcyl = (ide[nr].m2t >> 16) & 0xff;
		ide[nr].hcyl = (ide[nr].m2t >> 8) & 0xff;
		ide[nr].count = ide[nr].m2t;
		ide[nr].m2t = 0;	/* we are done. */
	}

	ret =
	    lotol_read_wrapper(ide[nr].unit[unit].fd, ide[nr].buf,
				ide[nr].count);
	assert(-1 != ret);

	ide[nr].tail = 0;
	ide[nr].nsector = 0x2;	/* set IO && delete RESET */
	ide[nr].status = DRQ_STAT;
	ide[nr].pcdone = 2;	/* prepare for packet command with data transfer */

	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_end(int nr)
{
	ide[nr].pcdone = 0;	/* no more to do */
	ide[nr].nsector = 0x3;	/* set C/D && I/o */
	ide[nr].status = READY_STAT;

	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_read_cdvd_capacity(int nr)
{				/* 0x25 */
	struct cap {
		unsigned lba;
		unsigned len;
	}  *cap = (struct cap *) ide[nr].buf;
	int ret;
	int unit = (ide[nr].select >> 4) & 1;
	loff_t res;

	ret = lotol__llseek_wrapper(ide[nr].unit[unit].fd, 0, 0, &res, SEEK_END);
	assert(-1 != ret);

	/* FIXME: check if res % (1 << UMIDE_BS) == 0 */

	cap->lba = htonl(res >> UMIDE_BS);
	cap->len = htonl(1 << UMIDE_BS);

	ide[nr].count = ide[nr].lcyl = 8;

	ide[nr].nsector = 0x2;	/* set io && clear reset */
	ide[nr].status = DRQ_STAT;
	ide[nr].tail = 0;
	ide[nr].pcdone = 2;	/* prepare for packet command with data transfer */

	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_seek(int nr)
{				/* 0x2b */
	ide[nr].nsector = 0x3;	/* set C/D && I/o */
	ide[nr].status = READY_STAT | SEEK_STAT;
	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_prevent_allow_medium_removal(int nr)
{				/* 0x1e */
	ide[nr].nsector = 0x3;	/* set C/D && I/o */
	ide[nr].status = READY_STAT;
	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_test_unit_ready(int nr)
{				/* 0x00 */
	ide[nr].nsector = 0x3;	/* set C/D && I/o */
	ide[nr].status = READY_STAT;
	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_read_disc_info(int nr)
{				/* 0x51 */
	ide[nr].nsector = 0x3;	/* set C/D && I/o */
	ide[nr].status = 0;	/* not implemented */
	ide[nr].tail = 0;
	UMIDE_GEN_IRQ;
}

static void
umide_gpcmd_step(int nr)
{				/* 0xA0 */

	if (1 == ide[nr].pcdone) {
		if (ide[nr].m2t)	/* more to transfer? */
			umide_continue_gpcmd_read_10(nr);

		umide_gpcmd_end(nr);
		return;
	}

	switch (ide[nr].buf[0]) {
	case GPCMD_MODE_SENSE_10:	/* 0x5a */
		umide_gpcmd_mode_sense_10(nr);
		break;

	case GPCMD_SEEK:	/* 0x2b */
		umide_gpcmd_seek(nr);
		break;

	case GPCMD_PREVENT_ALLOW_MEDIUM_REMOVAL:	/* 0x1e */
		umide_gpcmd_prevent_allow_medium_removal(nr);
		break;

	case GPCMD_TEST_UNIT_READY:	/* 0x00 */
		umide_gpcmd_test_unit_ready(nr);
		break;

	case GPCMD_READ_TOC_PMA_ATIP:	/* 0x43 */
		umide_gpcmd_read_toc_pma_atip(nr);
		break;

	case GPCMD_READ_CDVD_CAPACITY:	/* 0x25 */
		umide_gpcmd_read_cdvd_capacity(nr);
		break;

	case GPCMD_READ_DISC_INFO:	/* 0x51 */
		umide_gpcmd_read_disc_info(nr);
		break;

	case GPCMD_READ_10:	/* 0x28 */
		umide_gpcmd_read_10(nr);
		break;

	default:
		report_command(nr);
		assert(0);
	}
}

static void
umide_step(int nr)
{
	switch (ide[nr].command) {
	case WIN_RESTORE:	/* 0x10 */
		assert(0);
		break;
	case WIN_READ:		/* 0x20 */
		umide_read_end(nr);
		break;
	case WIN_WRITE:	/* 0x30 */
		umide_write_end(nr);
		break;
	case WIN_SPECIFY:	/* 0x91 */
		assert(0);
		break;
	case WIN_IDENTIFY:	/* 0xEC */
	case WIN_PIDENTIFY:	/* 0xA1 */
		umide_identify_end(nr);
		break;
	case WIN_PACKETCMD:	/* 0xA0 */
		umide_gpcmd_step(nr);
		break;
	default:
		printk("unknown step 0x%02x\n", ide[nr].command);

	}
}

unsigned char
umide_inb(int nr, unsigned short port)
{
	unsigned char value;
	int unit = (ide[nr].select >> 4) & 1;

	if (1 != ((ide[nr].active >> unit) & 1))	/* device present? */
		return value = 0;

	switch (port) {
	case 0:		/* data port */
		assert(ide[nr].tail < sizeof(ide[nr].buf));
		assert(0 < ide[nr].count);
		value = ide[nr].buf[ide[nr].tail];
		ide[nr].tail++;
		ide[nr].count--;
		if (ide[nr].count == 0) {
			if (ide[nr].pcdone == 2)
				ide[nr].pcdone = 1;
			umide_step(nr);
		}
		break;
	case 1:		/* error register */
		value = ide[nr].error;
		break;
	case 2:		/* sector count register */
		value = ide[nr].nsector;
		break;
	case 3:		/* sector number register */
		value = ide[nr].sector;
		break;
	case 4:		/* cylinder low register */
		value = ide[nr].lcyl;
		break;
	case 5:		/* cylinder high register */
		value = ide[nr].hcyl;
		break;
	case 6:		/* drive/head register */
		value = ide[nr].select | 0xa0;
		break;

	case 7:		/* status register */
		value = ide[nr].status | READY_STAT;
		break;
	case 8:		/* altstatus register */
		value = ide[nr].status | READY_STAT;
		/* ide[nr].status &= ~ERR_STAT; */
		break;
	default:
		assert(0);
	 /*NOTREACHED*/}
	return value;
}

static void
umide_report_unknown_step(int value)
{
	char buf[80];
	sprintf(buf,
		 "unknown step: 0x%02x please report to <sithglan@stud.uni-erlangen.de>\n",
		 value);
	lotol_write_wrapper(1, buf, strlen(buf));
}

static void
umide_dummy(int nr)
{
	ide[nr].status = READY_STAT;
	ide[nr].command = 0x00;
	UMIDE_GEN_IRQ;
}

void
umide_insb(int nr, unsigned short port, unsigned char *addr, unsigned int count)
{
	assert(0 < ide[nr].count);
	memcpy(addr, &ide[nr].buf[ide[nr].tail], count);
	ide[nr].tail  += count;
	ide[nr].count -= count;

	if (!ide[nr].count) {

		if (ide[nr].pcdone == 2)
			ide[nr].pcdone = 1;

		umide_step(nr);
	}
}

void
umide_insw(int nr, unsigned short port, unsigned char *addr,
	    unsigned int count)
{
	umide_insb(nr, port, (unsigned char *) addr, count * 2);
}

void
umide_outb(int nr, unsigned char value, unsigned short port)
{

	switch (port) {
	case 0:		/* data port */
#if 0 /* debugging facility */
		printk("##tg## umide_outb->(%d;%d;%d)\n",
				ide[nr].head, ide[nr].count, value);
#endif

		assert(ide[nr].head < sizeof(ide[nr].buf));
		assert(0 < ide[nr].count);
		ide[nr].buf[ide[nr].head] = value;
		ide[nr].head++;
		ide[nr].count--;
		if (ide[nr].count == 0)
			umide_step(nr);
		break;
	case 1:		/* write precompensation */
		/* FIXME */
#if 0 /* --ignore that --tg 21:26 2002-05-20 */ 
		assert(0);
#endif
		break;
	case 2:		/* sector count register */
		ide[nr].nsector = value;
		break;
	case 3:		/* sector number register */
		ide[nr].sector = value;
		break;
	case 4:		/* cylinder low register */
		ide[nr].lcyl = value;
		break;
	case 5:		/* cylinder high register */
		ide[nr].hcyl = value;
		break;
	case 6:		/* drive/head register */
		ide[nr].select = value & ~0xa0;
		break;
	case 7:		/* command register */
		ide[nr].command = value;
		switch (value) {
		case WIN_RESTORE:	/* 0x10 */
			umide_restore(nr);
			break;
		case WIN_READ:	/* 0x20 */
			umide_read(nr);
			break;
		case WIN_WRITE:	/* 0x30 */
			umide_write(nr);
			break;
		case WIN_SPECIFY:	/* 0x91 */
			umide_specify(nr);
			break;
		case WIN_IDENTIFY:	/* 0xEC */
		case WIN_PIDENTIFY:	/* 0xA1 */
			umide_identify(nr);
			break;
		case WIN_PACKETCMD:	/* 0xA0 */
			umide_gpcmd(nr);
			break;
		case WIN_STANDBYNOW1:	/* 0xE0 */
		case WIN_STANDBYNOW2:	/* 0x94 */
		case WIN_SLEEPNOW1:	/* 0xE6 */
		case WIN_SLEEPNOW2:	/* 0x99 */
		case WIN_SMART:	/* 0xB0 */
		case WIN_IDLEIMMEDIATE:	/* 0xe1 -> force drive to become ready */
			umide_dummy(nr);
			/* FIXME --tg */
			break;
		default:
			/* FIXME */
			umide_report_unknown_step(value);
			umide_dummy(nr);
			break;
		}
		break;
	case 8:		/* device control register */
		ide[nr].control = value;
		break;
	default:
		assert(0);
		/* NOTREACHED --tg 21:26 2002-05-20 */
	}
}

void
umide_outsb(int nr, unsigned short port, unsigned char *addr,
	     unsigned long count)
{
	assert(0 < ide[nr].count);
	memcpy(&ide[nr].buf[ide[nr].head], addr, count);
	ide[nr].head  += count;
	ide[nr].count -= count;

	if (!ide[nr].count)
		umide_step(nr);
}

void
umide_outsw(int nr, unsigned short port, void *addr, unsigned long count)
{
	umide_outsb(nr, port, addr, count * 2);
}

void
umide_init(void)
{
	int i;
	for (i = 0; i < UMIDE_BUSES; i++)
		probe_drives(i);
}
