/*
 * $Id: umrtc.c,v 1.1.1.1 2003/05/27 21:45:37 root Exp $
 */

/*
 * This is not a driver but a simulator reacting on in/out requests to
 * I/O addresses in range 0x70 - 0x7f.
 */

#include <asm/lotol.h>
#include <linux/assert.h>
#include <linux/mc146818rtc.h>


static unsigned char reg = -1;
static unsigned char rtc_freq_select = 0x00;
static unsigned char rtc_control = RTC_24H;
static unsigned char rtc_intr_flags = 0x00;

/*
 * At byte 15 of the CMOS RAM the BIOS expects a shutdown code to distinguish
 * between warm/hard resets, etc.  This one is need -- only for convenience --
 * by the SMP code.
 */
#define RTC_SHUTCODE	15


static inline void umktime(unsigned long t, int *year, int *month, int *day, int *hour, int *min, int *sec)
{
	/* Find a faster conversion function! FIXME VOSSI */
	const int secspermin = 60;
	const int secsperhour = 60 * secspermin;
	const int secsperday = 24 * secsperhour;
	static unsigned char dayspermonth[12] = {
		31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
	};
	unsigned int daysperyear;

	*year = 1970;
	*month = 0;
	*day = 0;
	*hour = 0;
	*min = 0;
	*sec = 0;
	daysperyear = (*year % 4 == 0) ? 366 : 365;
	while (daysperyear * secsperday <= t) {
		assert(365 <= daysperyear && daysperyear <= 366);
		t -= daysperyear * secsperday;
		(*year)++;
		daysperyear = (*year % 4 == 0) ? 366 : 365;
	}
	assert(1970 <= *year);
	dayspermonth[1] = (*year % 4 == 0) ? 29 : 28;
	while (dayspermonth[*month] * secsperday <= t) {
		t -= dayspermonth[*month] * secsperday;
		(*month)++;
	}
	(*month)++;	/* 0->1, 1->2, ... */
	assert(1 <= *month && *month <= 12);
	while (24 * 60 * 60 <= t) {
		t -= 24 * 60 * 60;
		(*day)++;
	}
	(*day)++;	/* 0->1, 1->2, ... */
	assert(1 <= *day && *day <= 31);
	while (60 * 60 <= t) {
		t -= 60 * 60;
		(*hour)++;
	}
	assert(0 <= *hour && *hour <= 23);
	while (60 <= t) {
		t -= 60;
		(*min)++;
	}
	assert(0 <= *min && *min <= 59);
	*sec = t;
	assert(0 <= *sec && *sec <= 59);

#if 0	/* only for debugging... */
	printk("date=%d-%d-%d %02d:%02d:%02d\n",
			*year, *month, *day, *hour, *min, *sec);
#endif
}

unsigned char umrtc_inb(unsigned short addr)
{
	unsigned char value;
	struct lotol_timeval tv;
	int sec, min, hour, day, month, year;
	int ret;


	switch (addr) {
	case 0x0070:
		/* correct? FIXME VOSSI */
		value = (unsigned long) reg;
		break;

	case 0x0071:
		if (reg == RTC_SECONDS
		 || reg == RTC_MINUTES
		 || reg == RTC_HOURS
		 || reg == RTC_DAY_OF_MONTH
		 || reg == RTC_MONTH
		 || reg == RTC_YEAR) {
			ret = lotol_gettimeofday(&tv, (void *) 0);
			assert(0 <= ret);
			umktime(tv.tv_sec, &year, &month, &day, &hour, &min, &sec);
			year -= 1900;	/* FIXME VOSSI */
			if (! (rtc_control & RTC_DM_BINARY) || RTC_ALWAYS_BCD) {
				BIN_TO_BCD(year);
				BIN_TO_BCD(month);
				BIN_TO_BCD(day);
				BIN_TO_BCD(hour);
				BIN_TO_BCD(min);
				BIN_TO_BCD(sec);
			}
		}
		if (reg == RTC_FREQ_SELECT) {
			ret = lotol_gettimeofday(&tv, (void *) 0);
			assert(0 <= ret);
		}
		switch (reg) {
		case RTC_SECONDS:
			value = sec;
			break;

		case RTC_MINUTES:
			value = min;
			break;

		case RTC_HOURS:
			value = hour;
			break;

		case RTC_DAY_OF_MONTH:
			value = day;
			break;

		case RTC_MONTH:
			value = month;
			break;

		case RTC_YEAR:
			value = year;
			break;

		case RTC_DAY_OF_WEEK:
			value = 0x00;	/* correct? FIXME VOSSI */
			break;

		case RTC_SECONDS_ALARM:
		case RTC_MINUTES_ALARM:
		case RTC_HOURS_ALARM:
			value = 0xff;	/* FIXME VOSSI */
			break;

		case RTC_FREQ_SELECT:
			value = 0;
			if (tv.tv_usec < 100 || 999900 < tv.tv_usec)
				value |= RTC_UIP;
			value |= rtc_freq_select & ~RTC_UIP;
			break;

		case RTC_CONTROL:
			value = rtc_control;
			break;

		case RTC_INTR_FLAGS:
			/* rtc_intr_flags register is cleared by read */
			/* and set by interrupt */
			value = rtc_intr_flags;
			rtc_intr_flags = 0x00;
			break;

		case RTC_VALID:
			value = RTC_VRT;
			break;

		default:
			/* FIXME VOSSI */
			printk("%s: reg=%d returning 0\n", __FUNCTION__, reg);
			value = 0;
			break;
		}
#if 0	/* only for debugging */
		printk("%s: read 0x%02x from register %d\n",
				__FUNCTION__, value, reg);
#endif
		break;

	default:
		/* FIXME VOSSI */
		panic("%s: addr=0x%04x\n", __FUNCTION__, addr);
	}

	return value;
}

unsigned short umrtc_inw(unsigned short addr)
{
	/* correct? FIXME VOSSI */
	unsigned short high;
	unsigned short low;

	low = umrtc_inb(addr);
	addr += 1;
	high = umrtc_inb(addr);
	
	return (high << 8) | (low << 0);
}

void umrtc_outb(unsigned char value, unsigned short addr)
{
	assert(0x70 <= addr && addr <= 0x71);

	switch (addr) {
	case 0x0070:
		reg = value;
		break;

	case 0x0071:
		switch (reg) {
		case RTC_SECONDS:
		case RTC_MINUTES:
		case RTC_HOURS:
		case RTC_DAY_OF_WEEK:
		case RTC_DAY_OF_MONTH:
		case RTC_MONTH:
		case RTC_YEAR:
			/* FIXME VOSSI */
			printk("WARNING: %s writes to RTC register %d.\n",
					current->comm, reg);
			break;

		case RTC_SECONDS_ALARM:
		case RTC_MINUTES_ALARM:
		case RTC_HOURS_ALARM:
			/* FIXME VOSSI */
			panic("%s: reg=%d\n", __FUNCTION__, reg);
			break;

		case RTC_FREQ_SELECT:
			rtc_freq_select &= RTC_UIP;
			rtc_freq_select |= value & ~RTC_UIP;
			break;

		case RTC_CONTROL:
			rtc_control = value;
			break;

		case RTC_INTR_FLAGS:
			/* shouldn't be written?!? FIXME VOSSI */
			panic("%s: reg=%d\n", __FUNCTION__, reg);
			break;

		case RTC_VALID:
			/* shouldn't be written?!? FIXME VOSSI */
			panic("%s: reg=%d\n", __FUNCTION__, reg);
			break;

		case RTC_SHUTCODE:
			/* Can be ignored. */
			break;

		default:
			/* FIXME VOSSI */
			panic("%s: reg=%d\n", __FUNCTION__, reg);
		}

#if 0	/* only for debugging */
		printk("%s: wrote 0x%02x to register %d\n", __FUNCTION__,
				value, reg);
#endif
		break;

	default:
		/* FIXME VOSSI */
		panic("%s: addr=0x%04x\n", __FUNCTION__, addr);
	}
}

void umrtc_outw(unsigned short value, unsigned short addr)
{
	umrtc_outb((unsigned char) (value & 0xff), addr);
	addr++;
	umrtc_outb((unsigned char) (value >> 8), addr);
}

void umrtc_interrupt(void)
{
	/* see arch/i386-linux/kernel/irq.h */
	/* and drivers/char/rtc.c */
#define RTC_IRQ	8
	extern void call_isa_irq(unsigned int);

	if ((rtc_control & RTC_AIE) && 0) {
		/* FIXME VOSSI */
	}
	if ((rtc_control & RTC_PIE) && 0) {
		/* FIXME VOSSI */
	}
	if ((rtc_control & RTC_UIE) && !(rtc_intr_flags & RTC_UF)) {
		rtc_intr_flags |= RTC_IRQF | RTC_UF;

		call_isa_irq(RTC_IRQ);
	}
}

void umrtc_init(void)
{
}
