/*
 * $Id: config.c,v 1.1.1.1 2003/05/27 21:45:53 root Exp $
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "lilo.h"
#include "config.h"


struct config config;


static char *
config_read_file(char *path)
{
	static char buf[2048];
	int fd;
	int ret;


	fd = open(path, O_RDONLY);
	if (fd < 0) {
		return (char *) 0;
	}

	ret = read(fd, buf, sizeof(buf));
	assert(0 <= ret && ret < sizeof(buf));
	buf[ret] = '\0';

	ret = close(fd);
	assert(0 <= ret);

	return buf;
}


void
config_read(void)
{
	char *buf;
	int i;

	/*
	 * memory config
	 */
	buf = config_read_file("memory");
	if (buf == (char *) 0) {
		fprintf(stderr, "%s: %s: %s.\n", progname,
				"memory", strerror(errno));
		exit(1);
	}

	config.mem_size = strtoul(buf, (char **) 0, 16);

	/*
	 * cpu config
	 */
	for (i = 0; i < MAX_CPUS; i++) {
		char path[64];

		sprintf(path, "cpu%d", i);
		buf = config_read_file(path);
		if (buf == (char *) 0) {
			config.cpu[0].available = 0;
			continue;
		}
		config.cpu[0].available = 1;
	}

	/*
	 * keyboard
	 */
	buf = config_read_file("keyboard");
	if (buf == (char *) 0) {
		fprintf(stderr, "%s: %s: %s.\n", progname,
				"keyboard", strerror(errno));
		exit(1);
	}

	sscanf(buf, "%lx:%*x\n", &config.keyboard.ipaddr);
	sscanf(buf, "%*x:%hx\n", &config.keyboard.ipport);

	/*
	 * display
	 */
	buf = config_read_file("video");
	if (buf == (char *) 0) {
		fprintf(stderr, "%s: %s: %s.\n", progname,
				"video", strerror(errno));
		exit(1);
	}

	config.display.video_size = 2 * 1024 * 1024;	/* FIXME VOSSI */
	sscanf(buf, "%lx:%*x:%*x\n", &config.display.ipaddr);
	sscanf(buf, "%*x:%hx:%*x\n", &config.display.ipport);
	sscanf(buf, "%*x:%*x:%d\n", &config.display.video_mode);

	/*
	 * ethernet
	 */
	for (i = 0; i < MAX_ETH_CARDS; i++) {
		char path[64];

		sprintf(path, "eth%d", i);
		buf = config_read_file(path);
		if (buf == (char *) 0) {
			config.eth[i].available = 0;
			continue;
		}
		config.eth[i].available = 1;
	}

	/*
	 * ide
	 */
	for (i = 0; i < MAX_HDS; i++) {
		char path[64];

		sprintf(path, "hd%c/config", 'a' + i);
		buf = config_read_file(path);
		if (buf == (char *) 0) {
			config.hd[i].available = 0;
			continue;
		}
		config.hd[i].available = 1;
	}

#if 0	/* FIXME VOSSI */
	/*
	 * floppies
	 */
	for (i = 0; i < MAX_FDS; i++) {
		char path[64];

		sprintf(path, "fd%c/config", '0' + i);
		buf = config_read_file(path);
		if (buf == (char *) 0) {
			config.fd[i].available = 0;
			continue;
		}
		config.fd[i].available = 1;
	}
#endif	/* FIXME VOSSI */
}
