/*
 * $Id: lilo.c,v 1.1.1.1 2003/05/27 21:45:53 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#define assert(expr)	\
if (! (expr)) { \
	char buf[256]; \
	sprintf(buf, "%s:%d: %s: %s failed\n", __FILE__, __LINE__, __PRETTY_FUNCTION__, #expr); \
	lotol_write(1, buf, strlen(buf)); \
}

#include <sys/types.h>
#include <elf.h>
#include <stdio.h>
#include <string.h>

/* can't use <asm/page.h> for assembler code */
/* #include <asm/page.h> */
#define PAGE_SIZE	4096

#include "lotol.h"
#include "lotol_fdesc.h"

#include "lilo.h"
#include "config.h"


/* see arch/i386-linux/kernel/head.c */
#define PARAM (MEM_BASE + 0x90000)
/* ... */
#define CL_MAGIC_ADDR (*(unsigned short *) (PARAM+0x20))
#define CL_OFFSET (*(unsigned short *) (PARAM+0x22))
/* ... */
#define MOUNT_ROOT_RDONLY (*(unsigned short *) (PARAM+0x1F2))
#define RAMDISK_FLAGS (*(unsigned short *) (PARAM+0x1F8))
/* ... */
#define LOADER_TYPE (*(unsigned char *) (PARAM+0x210))
#define KERNEL_START (*(unsigned long *) (PARAM+0x214))
#define INITRD_START (*(unsigned long *) (PARAM+0x218))
#define INITRD_SIZE (*(unsigned long *) (PARAM+0x21c))
/* ... */
#define COMMAND_LINE ((char *) (PARAM+2048))
#define COMMAND_LINE_SIZE 256

#define CL_MAGIC 0xA33F


char **
lilo_conf_read(void)
{
	static char config[4096];
	static char *argv[128];
	int argc;
	int fd;
	int ret;
	char *p;

	/*
	 * read config
	 */
	fd = lotol_open("lilo.conf", LOTOL_O_RDONLY, 0);
	if (fd < 0) {
		char buf[256];
		sprintf(buf, "%s: %s: %s.\n", progname,
				"lilo.conf", strerror(errno));
		lotol_write(2, buf, strlen(buf));
		lotol__exit(1);
	}

	ret = lotol_read(fd, config, sizeof(config));
	assert(0 <= ret && ret < sizeof(config));
	config[ret] = '\0';
	if (strchr(config, '\n') != (char *) 0) {
		*strchr(config, '\n') = '\0';
	}

	ret = lotol_close(fd);
	assert(0 <= ret);

	/*
	 * split config into pieces
	 */
	argc = 0;
	for (p = config; *p != '\0'; ) {
		argv[argc++] = p;
		while (*p != ' ' && *p != '\0') {
			p++;
		}
		while (*p == ' ') {
			*p = '\0';
			p++;
		}
	}
	argv[argc] = (char *) 0;

	return argv;
}


/* convert macro into string */
#define STR2(x)	#x
#define STR(x)	STR2(x)

/* static */ void
do_map_code_begin(void)
{
	/*
	 * will receive
	 * - mem_base in %eax
	 * - mem_size in %ebx
	 * - current pointer in %ecx
	 *
	 * We don't have current pointer any more FIXME VOSSI
	 */
	asm(
	"	pushl $" STR(PAGE_SIZE) "\n"		/* offset */
	"	pushl $" STR(FD_MAIN_MEMORY) "\n"	/* fd */
	"	pushl $" STR(LOTOL_MAP_SHARED|LOTOL_MAP_FIXED) "\n"	/* flags */
	"	pushl $" STR(LOTOL_PROT_READ|LOTOL_PROT_WRITE|LOTOL_PROT_EXEC) "\n" /* prot */
	"	subl $" STR(PAGE_SIZE) ", %%ebx\n"	/* len */
	"	pushl %%ebx\n"
	"	pushl $" STR(MEM_BASE+PAGE_SIZE) "\n"	/* addr */
	"	movl %%esp, %%ebx\n"
	"	movl $" STR(__NR_mmap) ", %%eax\n"
	"	int $0x80\n"
	"	addl $6*4, %%esp\n"

	"	int $3\n"
	: /* no output */
	: /* no input */
	: "memory"
	);
}
/* static */ void
do_map_code_end(void)
{
}

/* static */ void
do_unmap_code_begin(void)
{
	/*
	 * will receive
	 * - mem_base in %eax
	 * - mem_size in %ebx
	 * - current pointer in %ecx
	 *
	 * We don't have current pointer any more FIXME VOSSI
	 */
	asm(
	"	movl %%ecx, %%edx\n"		/* save current pointer */

	"	movl $" STR(MEM_BASE+PAGE_SIZE) ", %%ebx\n"	/* addr */
	"	movl %%edx, %%ecx\n"				/* len */
	"	addl $" STR(PAGE_SIZE) ", %%ecx\n"
	"	subl %%ebx, %%ecx\n"
	"	movl $" STR(__NR_munmap) ", %%eax\n"
	"	int $0x80\n"

	"	movl %%edx, %%ebx\n"				/* addr */
	"	addl $" STR(PAGE_SIZE) "*2, %%ebx\n"
	"	movl $0x80000000, %%ecx\n"			/* len */
	"	subl %%ebx, %%ecx\n"
	"	movl $" STR(__NR_munmap) ", %%eax\n"
	"	int $0x80\n"

	"	int $3\n"
	: /* no output */
	: /* no input */
	: "memory"
	);
}
/* static */ void
do_unmap_code_end(void)
{
}

#undef STR2
#undef STR

void
(*do_load(char **argv))(int)
{
	int sys_fd;
	Elf32_Ehdr elf_hdr;
	Elf32_Phdr elf_phdr;
	int p;
	char *initrd;
	int initrd_fd;
	unsigned long initrd_start;
	unsigned long initrd_end;
	int i;
	void (*func)(int);
	int ret;


	/* ========================================= */
	/* write trampoline code to address MEM_BASE */
	/* ========================================= */

	memcpy((char *) MEM_BASE +    0, do_map_code_begin,
			(unsigned long) do_map_code_end
			- (unsigned long) do_map_code_begin);

	memcpy((char *) MEM_BASE + PAGE_SIZE / 2, do_unmap_code_begin,
			(unsigned long) do_unmap_code_end
			- (unsigned long) do_unmap_code_begin);

#if 0	/* security hole!!! FIXME VOSSI */
	ret = mprotect((char *) MEM_BASE, PAGE_SIZE, PROT_READ | PROT_EXEC);
	assert(0 <= ret);
#endif	/* FIXME VOSSI */

	assert(*(int *) do_map_code_begin == *(int *) MEM_BASE);
	assert(*(int *) do_unmap_code_begin == *(int *) (MEM_BASE + PAGE_SIZE / 2));

	/* ===================== */
	/* load system to memory */
	/* ===================== */

	/* assuming ELF binary (FIXME!) */

	sys_fd = lotol_open(argv[0], LOTOL_O_RDONLY, 0);
	if (sys_fd < 0) {
		char buf[256];
		sprintf(buf, "%s: %s: %s.\n", progname,
				argv[0], strerror(errno));
		lotol_write(2, buf, strlen(buf));
		lotol__exit(1);
	}
	assert(0 <= sys_fd);

	/* read elf header */
	ret = lotol_read(sys_fd, (void *) &elf_hdr, sizeof(elf_hdr));
	assert(ret == sizeof(elf_hdr));

#if 0
	printf("e_type: %d\n", elf_hdr.e_type);
	printf("e_machine: %d\n", elf_hdr.e_machine);
	printf("e_version: %d\n", elf_hdr.e_version);
	printf("e_entry: %p\n", elf_hdr.e_entry);
	printf("e_phoff: %d\n", elf_hdr.e_phoff);
	printf("e_shoff: %d\n", elf_hdr.e_shoff);
	printf("e_flags: %d\n", elf_hdr.e_flags);
	printf("e_ehsize: %d\n", elf_hdr.e_ehsize);
	printf("e_phentsize: %d\n", elf_hdr.e_phentsize);
	printf("e_phnum: %d\n", elf_hdr.e_phnum);
	printf("e_shentsize: %d\n", elf_hdr.e_shentsize);
	printf("e_shnum: %d\n", elf_hdr.e_shnum);
	printf("e_shstrndx: %d\n", elf_hdr.e_shstrndx);
#endif

	/* read elf program header */
	for (p = 0; p < elf_hdr.e_phnum; p++) {
		ret = lotol_lseek(sys_fd, elf_hdr.e_phoff + p * elf_hdr.e_phentsize, 0);
		assert(ret == elf_hdr.e_phoff + p * elf_hdr.e_phentsize);

		ret = lotol_read(sys_fd, (void *) &elf_phdr, sizeof(elf_phdr));
		assert(ret == sizeof(elf_phdr));

#if 0
		printf("p_type: %d\n", elf_phdr.p_type);
		printf("p_offset: 0x%x\n", elf_phdr.p_offset);
		printf("p_vaddr: %p\n", elf_phdr.p_vaddr);
		printf("p_paddr: %p\n", elf_phdr.p_paddr);
		printf("p_filesz: 0x%x\n", elf_phdr.p_filesz);
		printf("p_memsz: 0x%x\n", elf_phdr.p_memsz);
		printf("p_flags: 0x%x\n", elf_phdr.p_flags);
		printf("p_align: 0x%x\n", elf_phdr.p_align);
#endif

		ret = lotol_lseek(sys_fd, elf_phdr.p_offset, 0);
		assert(ret == elf_phdr.p_offset);

		ret = lotol_read(sys_fd, (char *) elf_phdr.p_vaddr, elf_phdr.p_filesz);
		if (ret < 0) {
			char buf[256];
			sprintf(buf, "%s: %s: %s.\n", progname,
					argv[0], strerror(errno));
			lotol_write(2, buf, strlen(buf));
			lotol__exit(1);
		}
		if (ret < elf_phdr.p_filesz) {
			char buf[256];
			sprintf(buf, "%s: %s: %s.\n", progname,
					argv[0], "Short read");
			lotol_write(2, buf, strlen(buf));
			lotol__exit(1);
		}
		assert(ret == elf_phdr.p_filesz);

		if (elf_phdr.p_filesz != elf_phdr.p_memsz) {
			memset((char *) elf_phdr.p_vaddr + elf_phdr.p_filesz, 0,
					elf_phdr.p_memsz - elf_phdr.p_filesz);
		}
	}

	ret = lotol_close(sys_fd);
	assert(0 <= ret);

	/* ===================== */
	/* load initrd to memory */
	/* ===================== */

	for (i = 0; ; i++) {
		if (argv[i] == (char *) 0) {
			/* no initrd */
			initrd = NULL;
			break;
		}
		if (strncmp(argv[i], "initrd=", 7) == 0) {
			initrd = argv[i] + 7;
			break;
		}
	}
	if (initrd == NULL) {
		initrd_start = 0;
		initrd_end = 0;

	} else {
		initrd_fd = lotol_open(initrd, LOTOL_O_RDONLY, 0);
		if (initrd_fd < 0) {
			char buf[256];
			sprintf(buf, "%s: %s: %s.\n", progname,
					initrd, strerror(errno));
			lotol_write(2, buf, strlen(buf));
			lotol__exit(1);
		}

		initrd_start = config.mem_size / 2; /* FIXME */
		initrd_end = initrd_start;
		for (;;) {
			int len;

			len = lotol_read(initrd_fd, (char *) MEM_BASE + initrd_end,
					1024);
			if (len < 0) {
				char buf[256];
				sprintf(buf, "%s: %s: %s.\n",
						progname, initrd,
						strerror(errno));
				lotol_write(2, buf, strlen(buf));
				lotol__exit(1);
			}
			if (len == 0) {
				break;
			}
			initrd_end += len;
		}

		ret = lotol_close(initrd_fd);
		assert(0 <= ret);
	}

	/* ============= */
	/* set parameter */
	/* ============= */

	for (i = 0; ; i++) {
		if (argv[i] == (char *) 0) {
			MOUNT_ROOT_RDONLY = 1;
			break;
		}
		if (strcmp(argv[i], "read-only") == 0) {
			MOUNT_ROOT_RDONLY = 1;
			break;
		} else if (strcmp(argv[i], "read-write") == 0) {
			MOUNT_ROOT_RDONLY = 0;
			break;
		}
	}
	RAMDISK_FLAGS = 0;

	LOADER_TYPE = 1;	/* must be != 0 */
	/* KERNEL_START = 0;	FIXME */
	INITRD_START = initrd_start;
	INITRD_SIZE = initrd_end - initrd_start;

	CL_MAGIC_ADDR = CL_MAGIC;
	CL_OFFSET = (char *) COMMAND_LINE - (char *) PARAM;
	strcpy(COMMAND_LINE, "BOOT_IMAGE=");
	strcat(COMMAND_LINE, argv[0]);
	for (i = 1; argv[i] != (char *) 0; i++) {
		strcat(COMMAND_LINE, " ");
		strcat(COMMAND_LINE, argv[i]);
	}

	func = (void (*)(int)) elf_hdr.e_entry;

	return func;
}

void
sigurg_handler(int sig, void *si, void *uc)
{
	assert(sig == LOTOL_SIGURG);
	return;
}

void
(*do_wait(int cpu))(int)
{
	void (*func)(int);
	lotol_sigset_t mask;
	struct lotol_sigaction sa;
	int ret;
	int eip;

	sa.sa_sigaction = sigurg_handler;
	sa.sa_flags = LOTOL_SA_SIGINFO;
	lotol_sigemptyset(&sa.sa_mask);
	lotol_sigaddset(&sa.sa_mask, LOTOL_SIGURG);

	ret = lotol_rt_sigaction(LOTOL_SIGURG, &sa, 
			(struct lotol_sigaction *) 0, LOTOL__NSIG / 8);
	assert(0 <= ret);

	lotol_sigemptyset(&mask);

	ret = lotol_rt_sigsuspend(&mask, LOTOL__NSIG / 8);
	assert(ret < 0 && errno == LOTOL_EINTR);

	/* Get EIP */
	eip = (*(unsigned short *)(MEM_BASE + INIT_HIGH) << 4) 
		+ (*(unsigned short *)(MEM_BASE + INIT_LOW));

	eip += MEM_BASE;

	func = (void (*)(int)) eip;

	return func;
}
