/*
 * $Id: umide.c,v 1.1.1.1 2003/05/27 21:45:53 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#define __USE_FILE_OFFSET64 1
#define __USE_LARGEFILE64 1

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <string.h>

#define __USE_LARGEFILE64 1
#define __USE_FILE_OFFSET64 1

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "external_constants.h"
#include "lotol_fdesc.h"

#include "lilo.h"

void
umide_init(void)
{
	char drive;
	char path[32];
	char config[10]; /* contains disk _or_ cdrom */
	char name[] = "hd.";
	int fd;
	int ret;
	off64_t res;

	for (drive = 0; drive < MAX_HDS; drive++) {
		sprintf(name, "hd%c", drive + 'a');
		sprintf(path, "%s/model",  name);
		dfprintf(stderr,"path=%s\n",path);

		/*
		 * model
		 */
		fd = open(path, O_RDONLY);
		if (fd < 0 && errno != ENOENT) {
			fprintf(stderr, "%s: can't open: %s.\n",
					path, strerror(errno));
			exit(1);
		}

		if (fd < 0) {
			/* no such drive */
			continue;
		}

		assert(0 <= fd && fd < FD_LOTOL_MIN);

		ret = dup2(fd, FD_HD_MODEL(drive));
		assert(ret == FD_HD_MODEL(drive));

		ret = close(fd);
		assert(0 <= ret);

		/*
		 * config
		 */
		sprintf(path, "%s/config", name);
		fd = open(path, O_RDONLY);
		assert(0 <= fd && fd < FD_LOTOL_MIN);

		memset(&config, 0, sizeof(config));
		ret = read(fd, config, sizeof(config) - 1); /* disk or cdrom? */
		assert(0 < ret && ret < sizeof(config));
		config[ret - 1] = '\0'; /* XXX trim of newline */

		ret = dup2(fd, FD_HD_CONFIG(drive));
		assert(ret == FD_HD_CONFIG(drive));

		ret = close(fd);
		assert(0 <= ret);

		/*
		 * media
		 */
		sprintf(path, "%s/media", name);

		if (strncmp(config, "disk",  sizeof("disk")) == 0) {
			fd = open(path, O_RDWR | O_LARGEFILE, 0);

		} else if (strncmp(config, "cdrom", sizeof("cdrom")) == 0) {
			fd = open(path, O_RDONLY | O_LARGEFILE, 0);
		} else {
			fprintf(stderr, "config must be disk _or_ cdrom (config = '%s')\n"
					"aborting ...\n", config);
			assert(0);
		}

		if (fd < 0) {
			fprintf(stderr, "%s: can't open: %s.", path,
					strerror(errno));
			assert(0);
		}

		/* should use llseek FIXME VOSSI */
		res = lseek(fd, 0, SEEK_END);
		if (res < 0) {
			fprintf(stderr, "%s: lseek: %s.", path,
					strerror(errno));
			assert(0);
		}

		if (strncmp(config, "disk", sizeof("disk")) == 0 && (res & 0xff)) {
			fprintf(stderr, "%s: disksize must be a multiple of 512\n", path);
			assert(0);

		} else if (strncmp(config, "cdrom", sizeof("cdrom")) == 0 && (res & 0x3ff)) {
			fprintf(stderr, "%s: cdromsize must be a multiple of 2048\n", path);
			assert(0);
		}

		ret = dup2(fd, FD_HD_MEDIA(drive));
		assert(ret == FD_HD_MEDIA(drive));

		ret = close(fd);
		assert(0 <= ret);
	}
}

void
umide_shutdown(void)
{
	int drive;

	for (drive = 0; drive < MAX_HDS; drive++) {
		(void) close(FD_HD_MODEL(drive));
		(void) close(FD_HD_CONFIG(drive));
		(void) close(FD_HD_MEDIA(drive));
	}
}
