#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <revirt/ztrace.h>
#include <asm/unistd.h>

#include <revirt/wraw.h>

#define BUFSIZE 1024 * 1024 * 10

#define O_LARGEFILE 0100000

#define DISKSIZE ((loff_t)1024 * 3381651)
#define DMAPSIZE  ((unsigned int)OFF2BYTE(DISKSIZE))
#define DMAPBYTES (DMAPSIZE*sizeof(unsigned int))

#define dprintf(x...)

char  * buf, buf_area[BUFSIZE+4096];

#if 0
int llseek(unsigned int fd, unsigned long offset_high, 
	    unsigned long offset_low, loff_t * result, unsinged int whence) {
    return syscall(__NR__llseek, fd, offset_high, offset_low, result, whence);
}
#endif

unsigned long long lseek64(int fd, unsigned long long offset, unsigned int whence);

unsigned int * datamap;

int copyblock(int bfd, int cfd, loff_t offset, loff_t tlen) {
	loff_t result;
	int r;
	loff_t len;

#if 0
	if(len > BUFSIZE) {
		fprintf(stderr, "len to long: %lld, BUFSIZE %d\n", 
				len, BUFSIZE);
		exit(1);
	}
#endif
	printf("[0x%llx, 0x%x]\n", offset, len);
//	if(lseek64(bfd, hi, lo, & result, SEEK_SET)) {
	while(tlen > 0) {
	 len = (tlen > BUFSIZE) ? BUFSIZE : tlen;
	 if(tlen > len) {
		 printf("Warning: tlen %llx bigger than buffer, doing %llx (offset %llx)\n",
				 tlen, len, offset);
	 }
	 if((result = lseek64(bfd, offset, SEEK_SET))<0) {
		perror("lseek bfd");
		exit(1);
	 }
	 if(result != offset) {
		fprintf(stderr, "result: 0x%llx offset: 0x%llx\n",
			result, offset);
		exit(1);
	 }
	 if((r=read(bfd, buf, len))<0) {
		perror("read bfd");
		exit(1);
	 } else if(r != len) {
	     fprintf(stderr, "Warning: read returned %d, wanted %d\n",
				r, len);
	 }
//	if(lseek64(cfd, hi, lo, & result, SEEK_SET)) {
	 if((result=lseek64(cfd, offset, SEEK_SET)) <0) {
		perror("lseek bfd");
		exit(1);
	 }
	 if(result != offset) {
		fprintf(stderr, "result: 0x%llx offset: 0x%llx\n",
			result, offset);
		exit(1);
	 }
	 if((r=write(cfd, buf, len))<0) {
		perror("write cfd");
		exit(1);
	 } else if(r != len) {
	    fprintf(stderr, "Warning: write returned %d, wanted %d\n",
				r, len);
	 }
	 tlen -= len;
	 offset += len;
	}
	return 0;
}

int main(int argc, char * argv []) {
	int bfd, cfd, lfd, c;
	unsigned int i;
	unsigned long li, lm, bi, bm, ni, nm;
	long long total=0;
	loff_t start, end, len;
	if(argc != 4) {
		fprintf(stderr, "Usage: %s writelist backup copy\n",
				argv[0]);
		exit(1);
	}

	if((datamap=malloc(DMAPBYTES+sizeof(unsigned int)*2))==NULL) {
		fprintf(stderr, "Couldn't allocate datamap[%d]\n",
				DMAPBYTES+sizeof(int));
		exit(1);
	}

	if((lfd=open(argv[1], O_RDONLY))<0){
		perror("Opening datamap file");
		exit(1);
	}
	if((c=read(lfd, &datamap[1], DMAPBYTES))<0) {
		perror("Reading datamap\n");
		exit(1);
	}
	close(lfd);
	datamap[0]=datamap[DMAPSIZE+1]=0;  /* Encase in zeros */

	if((bfd=open(argv[2], O_RDONLY | O_LARGEFILE))<0){
		perror("Opening backup file");
		exit(1);
	}
	if((cfd=open(argv[3], O_WRONLY | O_LARGEFILE))<0){
		perror("Opening copy file");
		exit(1);
	}

	/* Need proper alignment for I/O to raw devices */
	buf =  buf_area + 4096;
	buf = (char *) (  (unsigned int)buf & 0xFFFFF800 );

#define Bit2Byte(x) ((x) >> 5)
#define Bit2Bitmask(x) (1 << ((x) & 31))
#define Bit2Off(x) ((x) << (12))

	for(i=32; i < (DISKSIZE >> 12)+32; i++) {
		/*
		 * i-1 and i+1 are OK, because we padded map around with 0s
		 */
		li=Bit2Byte(i-1);
		lm=Bit2Bitmask(i-1);
		bi=Bit2Byte(i);
		bm=Bit2Bitmask(i);
		ni=Bit2Byte(i+1);
		nm=Bit2Bitmask(i+1);

		dprintf("bit %d: li %d lm 0x%x bi %d bm 0x%x ni %d nm 0x%x\n",
				i-32, li, lm, bi, bm, ni, nm);

		if(datamap[bi] & bm) {
			if(!(datamap[li] & lm)) {
			    start=Bit2Off(i-32);
			    dprintf("bit %d set, bit %d not set; start=%lld\n",
						i-32, i-33, start);
			}
			if(!(datamap[ni] & nm)) {
			    end=Bit2Off(i-31);
			    /* 
			     * Partitions aren't guaranteed to be on page
			     * boundaries (in fact, disk geometry may make it
			     * hard to do so), so we account for the fact.
			     */
			    if(end > DISKSIZE)
				    end=DISKSIZE;
			    len=end-start;
			    total+=len;
			    dprintf("bit %d set, bit %d not set; end=%lld\n",
					    i-32, i-31, end);
			    copyblock(bfd, cfd, start, len);
			}
		}
	}
	printf("%lld bytes copied\n", total);
	close(cfd);
	close(bfd);
	return 0;
}
