#include <asm/ptrace.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include "readlog.h"
#include "printlog.h"
#include "decode.h"

#define ZTRACE_BREAKPOINT_METHOD
#define ZTRACE_OVERFLOW_METHOD
#define ZTRACE_FAST_REPSTRING_HANDLING
//#define SET_NETWORK_THROTTLE 1

/* Buffer to deal with overflow imprecision */
#define OVF_BUFFER (0x100)

#define RECORD_MAX(x, y) \
	if((x) > (y)) { \
	    (y)=(x); \
	}

struct {
    int max_ovf_error;
    s64 max_bc;
    int max_intr;
    int max_pf;
    s64 max_real;
    int nr_breakpoint;
    int max_breakpoint;
    int ztrace_call[32];
    int replen[8];

    int nr_repstring, cnt_repstring, nr_repstring_sstep,
	cnt_repstring_sstep;
} rstat;

int logcount[5]= {0, 0, 0, 0, 0};


#define try(x) do { if (!(x)) die(__FILE__,__LINE__,#x);  } while(0)

int debuglevel=0;

#ifndef ZTRACE_SSTEP_METHOD 
# ifndef ZTRACE_BREAKPOINT_METHOD
#  error "define a method first"
# endif
#endif

#define DEBUG0 1
#define DEBUG1 (1 << 1)
#define DEBUG2 (1 << 2)
#define DEBUG3 (1 << 3)
#define THROTTLE (1 << 4)

struct task {
	short was_taken;
//	short syscall_pending;
	int has_run;
	int pid;
	s64 counter;
	s64 last_count;
	int last_data;
	unsigned last_eip;
	int stop_eip;
	struct rlog_schedule_entry lastlog;
};

struct task tasklist[10];
int mapl;
int stop_request=0;

#define STATE_HISTORY 100
#define STATEREC(x) (staterec[(x)%STATE_HISTORY])
int staterec_index=0, staterec_lastlog=0;
struct ztrace_state_struct staterec[STATE_HISTORY];

int writestaterec(void) {
    int i, fd;
    if((fd = creat("staterec", 0644))<0) {
	perror("creat");
	return -1;
    }
    for(i=staterec_index; i<staterec_index + STATE_HISTORY; i++) {
	if( write(fd, (char *)&STATEREC(i), sizeof(struct ztrace_state_struct))
		    	<0) {
		perror("write");
		return -1;
	}
    }
    return 0;
}

void sigusr1(int sig) {
	fprintf(stderr, "zR: Writing staterec\n");
	writestaterec();
}

void sigint(int sig) {
	static int c=0;
	struct ztrace_state_struct zt_state;
	if(c) {
		fprintf(stderr, "Second interrupt received, dying.\n");
		exit(1);
	} else {
		c=1;
#if 0
		stop_request=1;
		printf("Interrupt received, putting in stop request.\n");
#else
		/* UMLinux specific!  Host-based needs last proc run */
		if(ztrace(ZTRACE_GETSTATE, tasklist[0].pid,
			0, &zt_state)) {
			perror("getstate");
		}
		if(zt_state.revirt_flags & REPLAY_SYSCALL) {
		 printf("Stop request received, but still in syscall.  Calling SYSCALL_EXIT...\n");
		 ztrace(ZTRACE_SYSCALL_EXIT,tasklist[0].pid,0,0);
		} 
		printf("Stop request received, exiting\n");
		exit(1);
		
#endif
	}
}

int debug_pause(void) {
	fprintf(stderr, "Pausing with SIGSTOP\n");
	fflush(stdout);
	kill(getpid(), SIGSTOP);
}

void die(char *file,int line,char *ass) {
	fprintf(stderr, "file %s, line %d: assertion %s failed",
		file,line,ass);
	perror("system error is");
	*((int*)0)= 0;
}



void new_task(int pid) {
	try(mapl<10);
	tasklist[mapl]= (struct task) {.was_taken= 0, .pid= pid, 
					.counter= 0, .has_run= 0 };
	mapl++;
	if(debuglevel & DEBUG0)
		printf("New task %d\n",pid);

}

void unmap_task(struct task *tsk) {
	try(mapl>0);
	if(debuglevel & DEBUG0)
		printf("No longer tracing task %d\n",tsk->pid);
	/*
  	 * move last task in list to deleted task's space
	 */
	mapl--;
	*tsk= tasklist[mapl];
}

struct task *find_by_pid(int pid) {
	int i;
	for(i= 0;i<mapl;i++)
		if (tasklist[i].pid==pid)
			return tasklist+i;
	fprintf(stderr, "Cannot find pid %d\n", pid);
	try(0);
	return 0;
}

int singlestep_task(struct task *task) {
	struct ztrace_state_struct state; 
	unsigned eax;
	int ret;

	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);

	if(state.regs.eip == task->stop_eip) {
		task->last_count = task->counter;
		task->last_data = state.regs.eip;
	}

	task->was_taken= branch_taken(state.code, state.regs.eflags,state.regs.ecx);

	/*
	 * We shouldn't need to do this int 80 enter/exit thing anymore,
	 * unless we really want to stop in the middle of the kernel.
	 * ztrace_tsk_exits_syscall() will yield if SSTEP is set, so int 80
	 * should respond to SSTEP like you'd expect.
	 */
	if(is_syscall(state.code)) {
		fprintf(stderr, "Calling ZTRACE_SSTEP for system call\n");
		fprintf(stdout, "Calling ZTRACE_SSTEP for system call\n");
	}

	ret=ztrace(ZTRACE_SSTEP_WAIT,task->pid,0,0);
	if(ret==ZTRACE_SIGNALQUEUE) {
		if(debuglevel & DEBUG1)
		  printf(" sstep interrupted by signal!\n");
		if(task->was_taken) {
			task->was_taken=0;
			if(debuglevel & DEBUG1)
			  printf(" clearing task->was_taken\n");
		}
	}
	if (task->was_taken) {
		return 1;
	} else
		return 0;	
}

#if 0
/*
 * This was added because 'schedule_entry' used to mean just that-- the task
 * was scheduled out; but sometimes a task was scheduled out multiple times
 * before exec() even completed; so in zr we tossed out logs that looked
 * like they didn't do anything.  But as soon as we get one 'real' log, do
 * them all.
 *
 * FIXME This works now because you can't get scheduled without an interrupt
 * or exception of some type.  But if we ever throw these away to save space,
 * (the P4 with retired_branch_type doesn't need them) watch out.  If the 
 * first signal happens before the first branch, we'll skip it (in error).
 */
int null_log(struct revirt_ctrdata *ctr) {
	static int done=1;
	if(ctr->hw_branch)
		done=0;
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	if(ctr->hw_intr | ctr->ex_pf)
		done=0;
#endif
	return done;
}
#endif

/*
 * Single-step through repstring until we get to the next instruction
 *
 * Well, OK, now we decode the instruction, set a bkpt for the next eip, and
 * let it run.  
 */
void repstring_step(struct task * task) {
	struct ztrace_state_struct state;
	unsigned long this_eip;
	struct ztrace_rununtil_struct ru;
	struct ztrace_breturn bbuf;
	int d;
	s64 count, xret;
	rstat.nr_repstring_sstep++;
	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
#ifdef SSTEP_THRU_REPSTRING 
	this_eip = state.regs.eip;
	while(1) {
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if(this_eip) {
			if(this_eip!=state.regs.eip)
				break;
		}
#if 0
 		else
			this_eip=state.regs.eip;
#endif
		if(singlestep_task(task)) {
			fprintf(stderr, "Wierd... singlestep_task returns non-zero\n");
			debug_pause();
		}
		rstat.cnt_repstring_sstep++;
	}
	d=state.regs.eip - this_eip;
	if(d >0 && d < 8) {
		rstat.replen[d]++;
	} else
		rstat.replen[0]++;
#else /* sstep_thru_repstring */
	/*
 	 * At this point, it's already established that we've got
	 * a repstring; we just need to know how long it is.  For now
	 * we assume that the actual opcode is only one byte long; 
  	 * the rest comes from prefixes (including the 'rep' prefix)
	 */
	count=ADJ_BC(state.ctr);
	xret = parse_prefixes(state.code, &d);
	this_eip = state.regs.eip + d + 1;
	if(debuglevel & DEBUG1) {
		printf("parse_prefixes returns xret = 0x%x, offset = %d\n", 
				xret, d);
		printf("repstring_step: stepping from eip %x to %x\n",
			state.regs.eip, this_eip);
		if(debuglevel & DEBUG2)
			debug_pause();
	}
	ru.flags = REPLAY_SSTEP;
	ru.eip = this_eip;
	ru.wchan = 0;
	ru.count = count;
	 if(ztrace(ZTRACE_RUNUNTIL, task->pid, (unsigned long)&ru, &bbuf)<0) {
		perror("ztrace_rununtil");
		exit(1);
	 }
	
	 xret=ADJ_BC(bbuf.ctr);
	 if(xret - count > 0) {
	 	printf("decode & breakpoint returned %d!\n", xret);
		//debug_pause();
	 }
#endif
}

/*
 * Step through a repz instruction until current ecx == logged ecx (data).
 */
void handle_repstring(struct task *task,unsigned int data) {
	struct ztrace_state_struct state;
	unsigned long this_eip=0, eflags=0;
	long ecx=0;
	unsigned long real_ecx=0, current_ecx;
	long delta;
	rstat.nr_repstring++;

	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	real_ecx=state.regs.ecx;
	this_eip=state.regs.eip;
	eflags = state.regs.eflags;

	/*
	 * FIXME -- both of these methods have the disadvantage that they
	 * modify ecx.  This is fine if we're only ever running repstring
	 * instructions; but what if the repstring is interrupted by a 
	 * guest page fault?  The pf handler might see a different ecx...
	 */
#ifdef ZTRACE_FAST_REPSTRING_HANDLING
	/*
	 * FIXME -- is this a problem with 16 bit addr-size attributes?
	 */
	ecx = state.regs.ecx;
	delta = ecx - data ;
#if 0
	fprintf(stderr, "handle_repstring: eip 0x%x ecx %d data %d delta %d\n",
			this_eip, ecx, data, delta);
#endif
	ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)delta);
	/*
	 * Step through the instruction
	 */
	repstring_step(task);
	/*
	 * Calculate 'real' value of ecx...
	 */
	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	real_ecx = ecx - (delta - state.regs.ecx);

	/*
	 * Check to see if everything's ok...
	 */
	if(state.regs.ecx != 0) {
		fprintf(stderr, "hmm, repstring_step, ecx != 0?\n");
	}

	if(real_ecx != data) {
		fprintf(stderr, "real_ecx != data: ecx %d delta %d state.regs.ecx %d data %d real_ecx %d\n",
			ecx, delta, state.regs.ecx, data, real_ecx);
		fprintf(stdout, "real_ecx != data: ecx %d delta %d state.regs.ecx %d data %d real_ecx %d\n",
			ecx, delta, state.regs.ecx, data, real_ecx);
		debug_pause();
	}

	/*
	 * And reset our EIP and ECX back to normal
	 */
	if(eflags != state.regs.eflags) {
#if 0
		fprintf(stderr, "eflags went from %x to %x.  Changing back?\n",
				eflags, state.regs.eflags);
#endif
		if( ztrace(ZTRACE_PUTREG, task->pid, EFL, (void *)eflags) ) {
			perror("ztrace_putreg: eflags");
		}
	}
	if( ztrace(ZTRACE_PUTREG, task->pid, EIP, (void *)this_eip) ) {
		perror("ztrace_putreg: eip");
	}
	if(ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)real_ecx) ) {
		perror("ztrace_putreg: ecx");
	}
#else
	while(1) {
		/*
		 * Pentum 4 has fast string instructions, which if > a
		 * certain amount jump down by 128 (or something like that)
		 * rather than one at a time if greater than a certain
		 * amount.
		 * 
		 * So we make it small (2) so that it only goes up / down
		 * by 1.  We maintain real_ecx in parallel, moving it up
		 * or down as we see state.regs.ecx move up or down from 2.
		 */
		ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)2);
		/* FIXME -- This getstate is redundant. */
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if(state.regs.ecx != 2) {
		    fprintf(stderr, "ZTRACE_PUTREG didn't change ecx from %d to 2.\n",
			    	state.regs.ecx);
		    debug_pause();
		}

		/*
		 * Make sure we haven't drifted or anything wierd
		 * like that...
		 */
		if(this_eip) {
		    if(this_eip!=state.regs.eip) {
			fprintf(stderr, "Wrong eip! want %x, got %x\n",
				this_eip, state.regs.eip);
			debug_pause();
		    }
		}

		/*
		 * Are we there yet?
		 */
		if (real_ecx == data)
			break;
		else if(debuglevel & DEBUG1)
			printf(" repstring: ecx==%d [%d].  want %d\n",
					real_ecx, state.regs.ecx, data);

		ecx=state.regs.ecx;
		if(singlestep_task(task)) {
		   fprintf(stderr, "That's wierd... singlestep_task returns non-zero.\n");
		   debug_pause();
		}
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		switch(state.regs.ecx - ecx) {
		   case 1:
		       fprintf(stderr, "ecx goes up!\n");
		       real_ecx+=1;
		       break;
		   case -1:
		       real_ecx-=1;
		       	break;
		   default:
			fprintf(stderr, "ecx jumps from %d to %d!\n",
					ecx, state.regs.ecx);
			real_ecx += (state.regs.ecx-ecx);
			//debug_pause();
		}
		rstat.cnt_repstring++;
	}
	ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)real_ecx);
#endif
}

int parse_prefixes(unsigned char * code, int * cstart);


/*
 * We should only stop for system calls if we want to -- i.e., 
 * REPLAY_SSTEP should be set *only* when single-stepping through
 */
s64 breakpoint(struct task *task,unsigned eip, int flags, unsigned long wchan, 
			s64 real_count) {
	struct ztrace_breturn bbuf;
	struct ztrace_state_struct state;
	struct ztrace_rununtil_struct ru;
	s64 ret,xret;

	ret= 0;

	/* 
 	 * First, if we're already at the bkpt in userland, step once.  
	 * If this is a string instruction, step through to the end.
	 *
	 * If we're in the kernel, the fact that we're already at the
	 * breakpoint means that the bkpt is at our return address.
	 * If we want to stop right there, single-step will confound
	 * us.
	 */
        ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	/* 
	 * UMLinux tailoring: for zero branch counts, don't do single-stepping.
	 * This allows us to go to usermode and back again -- useful if we
	 * need to deliver two signals in a row.
	 */
        if(state.regs.eip==eip 
		&& state.revirt_flags & REPLAY_PROCESS_IN_UMODE
		&& real_count > 0) {

		if(debuglevel & DEBUG1) 
			printf("Single-stepping over ip...\n");
                task->was_taken= 0;
                //task->syscall_pending= 0;

		/*
		 * Step over this eip.  If a repstring, decode & bkpt; 
		 * otherwise, just use the TRAP flag.
		 */
		if(is_repstring(state.code)) {
		    if(debuglevel & DEBUG1)
			printf(" instruction is repstr, stepping through...\n");
		    repstring_step(task);
		} else {
                	ret=singlestep_task(task);
		}
#if 0
                if (task->syscall_pending) {
			printf("Calling ZTRACE_SYSCALL_EXIT\n");
                        ztrace(ZTRACE_SYSCALL_EXIT,task->pid,0,0);
		} 
#endif
        }

	/* 
	 * FIXME -- I'm pretty sure this shouldn't be here... it should be
	 * next to the 'XUSER' stop stuff.
	 *
	 * Anyway, if the process blocked during logging, wait for it
	 * to block in the same place...
	 */
	if(wchan) {
	 ru.eip=eip;
	 ru.wchan=wchan;
	 ru.flags=flags;

	 if(debuglevel & DEBUG1)
		printf("Waiting for block 0x%x:0x?%x\n", eip, wchan);

	 if(ztrace(ZTRACE_BLOCK, task->pid, (unsigned long)&ru, &bbuf)<0) {
	 	perror("ztrace_block");
		exit(1);
	 }
	} else {
	 /*
	  * Now set the breakpoint and run.
	  */
	 if(debuglevel & DEBUG1)
		printf("Setting breakpoint for 0x%x\n", eip);

	 ru.eip=eip;
	 ru.flags=flags;
	 ru.wchan=0;
	 ru.count = real_count;
	 if(ztrace(ZTRACE_RUNUNTIL, task->pid, (unsigned long)&ru, &bbuf)<0) {
		perror("ztrace_rununtil");
		exit(1);
	 }
	}

	xret = ADJ_BC(bbuf.ctr);
	if (bbuf.eip!=eip) {
		printf("eips differ. new=%x, orig= %x\n",bbuf.eip, eip);
		try(bbuf.eip==eip);
	}
	if(debuglevel & DEBUG1) {
		printf("task %d returns with eip= %x bc= %x\n",task->pid,
			bbuf.eip, xret);
#ifdef REPLAY_RECORD_ALL_PERFCTRS
		printf("returned counts: hw=%x i=%x pf=%x def=%x\n",
			bbuf.ctr.hw_branch, bbuf.ctr.hw_intr, bbuf.ctr.ex_pf,
			 bbuf.ctr.ex_dev);
#else
		printf("returned counts: hw=%x\n", bbuf.ctr.hw_branch);
#endif
	}
	ret = xret;

	return ret;
}

void do_schedule_entry(struct rlog_schedule_entry *log) {
	struct task *task;
	s64 real_count;
	int iterations;
	struct ztrace_state_struct state;
	int errors=0, flag_doone=0;


	if(debuglevel & DEBUG0) {
		 printf("do_schedule_log: %d\n", log->pid);
		print_schedlog(log);
	}
	task= find_by_pid(log->pid);

#if 0
	/*
	 * Skip to the first log with a branch...
	 */
	if (null_log(&log->ctr)) {
		if(debuglevel & DEBUG0) 
			printf("null log, ignoring\n");
		return ;
	}
#endif

	/*
	 * If we've never run this pid before, attach to it
	 */
	if(!task->has_run) {
	    try(!ztrace(ZTRACE_ATTACH, task->pid,0,0));
	    if(debuglevel & DEBUG0) 
		printf("Attached to pid %d\n", task->pid);
	    task->has_run = 1;
	}

	real_count = ADJ_BC(log->ctr);

	/*
	 * Log some statistics...
	 */
	RECORD_MAX(real_count, rstat.max_real);
	RECORD_MAX(log->ctr.hw_branch, rstat.max_bc);
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	RECORD_MAX(log->ctr.hw_intr, rstat.max_intr);
	RECORD_MAX(log->ctr.ex_pf, rstat.max_pf);
#endif

	if(debuglevel & DEBUG1) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	    printf("real count: %llx [%llx %x %x]\n",
			real_count,
			log->ctr.hw_branch,
			log->ctr.hw_intr,
			log->ctr.ex_pf);
#else
	    printf("real count: %llx [%llx]\n",
			real_count);
#endif
	}

	if(real_count < 0) {
		printf("Warning: real_count < 0, treating as zero...\n");
		errors++;
		real_count=0;
	}

	/*
 	 * Do some sanity checks, see if we can get away with skipping
	 * this log.
	 */
	if(real_count == 0 
	   && log->eip == task->lastlog.eip 
	   && (log->flags == task->lastlog.flags)) {

	    if(log->flags & REPLAY_PROCESS_IN_UMODE) {
		if( log->data == task->lastlog.data) {
			if(debuglevel & DEBUG0)
			  printf("Zero count, no change in eip, same flags," \
					" same ecx. Treating as null...\n");
			if(debuglevel & DEBUG2) debug_pause();
			return;
		} else if(debuglevel & DEBUG2) {
			fprintf(stderr, "WARNING: only ecx changes; " \
				"suspect multiple interruption of string" \
				" instruction\n");
			errors++;
		}
	    } else { // log->flags & REPLAY_PROCESS_IN_UMODE
		if(log->nrxuser == task->lastlog.nrxuser && 
		   log->edi == task->lastlog.edi) {
			if(debuglevel & DEBUG0)
			   printf("Still in kernel, same nrxuser & edi."\
				" Treating as null.\n");
			return;
		}
	    }
	}
	task->counter= 0;
	task->last_count=0;
	task->last_data=0;

	task->last_eip = task->stop_eip = log->eip;

	ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
	STATEREC(staterec_index).debugreg[1]=logcount[0];
	STATEREC(staterec_index).debugreg[2]=0;
	staterec_index++;


	/* 
	 * Each of these "methods" should stop with the next eip as prescribed
	 * in the log, and the # of branches.  If the next eip is on the 
	 * return from a syscall, they should stop on the entry to
	 * the syscall.  After that we decide whether to stop in the middle
	 * of the syscall or go to the end.
	 */

#ifdef ZTRACE_SSTEP_METHOD
#error ZTRACE_SSTEP_METHOD no longer supported
#endif

#ifdef ZTRACE_OVERFLOW_METHOD
    	/*
	 * First get as close as possible using performance counter overflow
	 * then we step closer.
	 */
	if(real_count > OVF_BUFFER) {
	    s64 taken;
	    s64 b1, b2, ocount;
	    struct ztrace_breturn bret;

	    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	    b1=state.ctr.hw_branch;

	    ocount = real_count - OVF_BUFFER;
	    if(debuglevel & DEBUG1) {
		printf("%d:setting performance overflow for %lld branches\n", 
				task->pid, ocount);
	    }

	    if(ztrace(ZTRACE_PERFOVF, task->pid, (unsigned long)&ocount,
		    		&bret) < 0) {
		perror("performance overflow");
	    }

	    taken=ADJ_BC(bret.ctr);
	    task->counter=taken;

	    if(taken > real_count) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
		fprintf(stderr, "Overflow overcounted! wanted %lld [%llx %x %x], got %lld[%llx %x %x]\n",
				real_count, 
				log->ctr.hw_branch,
				log->ctr.hw_intr,
				log->ctr.ex_pf,
				taken,
				bret.ctr.hw_branch,
				bret.ctr.hw_intr,
				bret.ctr.ex_pf);
		printf("Overflow overcounted! wanted %lld [%llx %x %x], got %lld[%llx %x %x]\n",
				real_count, 
				log->ctr.hw_branch,
				log->ctr.hw_intr,
				log->ctr.ex_noerror,
				taken,
				bret.ctr.hw_branch,
				bret.ctr.hw_intr,
				bret.ctr.ex_pf);
#else
		fprintf(stderr, "Overflow overcounted! wanted %lld [%llx], got %lld[%llx]\n",
				real_count, 
				log->ctr.hw_branch,
				taken,
				bret.ctr.hw_branch);
		printf("Overflow overcounted! wanted %lld [%llx], got %lld[%llx ]\n",
				real_count, 
				log->ctr.hw_branch,
				taken,
				bret.ctr.hw_branch);
#endif
		debug_pause();
	    }
	    if(taken == real_count) {
		printf("taken==real_count!\n");
	    }

	    b2=bret.ctr.hw_branch - b1 - ocount;

	    RECORD_MAX(b2, rstat.max_ovf_error);

	    if(debuglevel&DEBUG1) {
		printf("%d: overflow returned %lld (error %lld)\n",
			task->pid, taken, b2);
	    }
	}

#endif
	
#ifdef ZTRACE_BREAKPOINT_METHOD
	iterations=0;
	/* 
	 * We may not have any loops, but still need to move forward
	 * to the next IP address.  We may also be in a system call,
	 * in which case we need not move forward at all.
	 */
	ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	if((state.revirt_flags&REPLAY_PROCESS_IN_UMODE) 
			!= (log->flags & REPLAY_PROCESS_IN_UMODE))
		flag_doone=1;
	/*
	 * UMLinux specific: we only schedule for signals, so the last
	 * entry must have been a signal; zero count means we put two
	 * signals sequentially on the stack; but then we need to call
	 * do_signal again, which means we need to return to user-mode
	 * again... anyway, set the breakpoint.
	 */
	if(real_count == 0) {
		if(debuglevel & DEBUG0)
		 printf("real_count==0, setting flag_doone.\n");
		flag_doone=1;
	}

	while(task->counter!=real_count 
		|| state.regs.eip!=log->eip 
		|| flag_doone) {
		s64 taken;
		iterations++;
		flag_doone=0; /* FIXME */

		ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
		STATEREC(staterec_index).debugreg[1]=logcount[0];
		STATEREC(staterec_index).debugreg[2]=iterations;
		staterec_index++;

		taken= breakpoint(task,log->eip, log->flags, log->wchan, real_count);
		task->counter= taken; 

		if(debuglevel & DEBUG1)
		 printf("%d:Breakpoint returns %lld.  Total now %lld; want %lld.\n", 
				logcount[0], taken, task->counter, real_count);

		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if (taken > real_count) {
		    fprintf(stderr, "%d: Breakpoint returns %lld [%llx], wanted %lld[%llx]\n",
				logcount[0],
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
		    printf("%d: Breakpoint returns %lld [%llx], wanted %lld[%llx]\n",
				logcount[0],
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
		    printf("last_log:\n");
		    print_state(&STATEREC(staterec_lastlog));
		    printf("last_bkpt:\n");
		    print_state(&STATEREC(staterec_index-1));
		    printf("current:\n");
		    print_state(&state);
		    debug_pause();
		    errors++;
		    break;
		} else if (taken == real_count-1 && (debuglevel & DEBUG3)) {
		    printf("Close shave: %lld [%llx], log %lld[%llx]\n",
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
		}

		if(debuglevel & DEBUG1)
			printf("  eip=0x%x ecx=%d rep=%c\n",
				state.regs.eip, state.regs.ecx,
				is_repstring(state.code)?'y':'n');
		if(state.code[0]==0xcc) {
			printf("Warning: bkpt on int3\n");
		    	print_state(&state);
			//debug_pause();
		}
	}
	ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
	STATEREC(staterec_index).debugreg[1]=logcount[0];
	STATEREC(staterec_index).debugreg[2]=iterations+1;
	staterec_index++;

	ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
#endif /* ZTRACE_BREAKPOINT_METHOD */

	/* 
	 * Now we're to the eip in question.  If we stopped halfway thru a
	 * syscall, walk to the right place; If we stopped in the middle of a
	 * repstring, walk until we're at the right iteration.
	 */
	if(!(log->flags & REPLAY_PROCESS_IN_UMODE)) {
		struct ztrace_xuser_stop zt_xstop;
		if(log->nrxuser || log->edi) {
		 zt_xstop.edi=log->edi;
		 zt_xstop.nrxuser=log->nrxuser;
		 if(debuglevel & DEBUG1) {
		  printf("Interrupted in syscall.  calling ZTRACE_XSTOP.\n");
		  printf("    edi:0x%x nrxuser:%d\n", zt_xstop.edi, 
				zt_xstop.nrxuser);
		 }
		 if(ztrace(ZTRACE_XSTOP, task->pid, (unsigned long)&zt_xstop,0))
		 { 
			perror("XSTOP");
		 }
		} else if(debuglevel & DEBUG1) {
		 printf("Stopped in kernel, but before any xusers.\n");
		}
	} else if (is_repstring(state.code) && (state.regs.ecx != log->data)) {
		if(debuglevel & DEBUG0)
		 printf(" Handling repeat string instruction...\n");
		handle_repstring(task,log->data);
		ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	}

	errors += ((unsigned)state.regs.eip != log->eip);
	errors += (task->counter != real_count);
	errors += (state.regs.ecx != log->data);

	if((debuglevel & DEBUG3) || errors)
		printf("log done. pid=%d\n"
		 "          eip= %08x count= %08llx ecx= %08x\n" 
		 "          eip= %08x count= %08llx ecx= %08x (logged)\n",
		 task->pid, 
		 (unsigned) state.regs.eip, task->counter, state.regs.ecx,
		 log->eip, real_count, log->data);
	if(errors) {
		printf("%d warnings.\n", errors);
		printf("Writing staterec to disk...");
		writestaterec();
		printf("done.\n");
		debug_pause();
	}
	task->lastlog=*log;
	if(ztrace(ZTRACE_CLEARBRANCH, task->pid, 0, 0)) {
		perror("ZTRACE_CLEARBRANCH!\n");
		try(0);
	}
}

void print_sigset_t(__kernel_sigset_t * t) {
	int i, j, count=1;

	for(i=0; i<_NSIG_WORDS_K; i++) {
	    for(j=0; j<_NSIG_BPW_K; j++, count++) {
		if(t->sig[i] & (1 << j)) 
			fprintf(stderr, "%d ", i*_NSIG_BPW_K + j);
	    }
	}

	fputc('\n', stderr);
}

void deliver_signals(struct rlog_entry_signal * s) {
	if(debuglevel & DEBUG3) {
	 fprintf(stderr, "Delivering signal to pid %d\n", s->pid);
	 fprintf(stderr, " blocked: ");
	 print_sigset_t(&s->blocked);
	 fprintf(stderr, " pending_signal: ");
	 print_sigset_t(&s->pending_signal);
	 fprintf(stderr, " dfl: ");
	 print_sigset_t(&s->dfl);
	}
	ztrace(ZTRACE_SIGNALQUEUE, s->pid, 0, s);
}

void do_exit_entry(struct rlog_exit_entry * ee) {
	struct task *task;

	task=find_by_pid(ee->pid);

	if(!task->has_run) {
	    try(!ztrace(ZTRACE_ATTACH, task->pid,0,0));
	    if(debuglevel & DEBUG0)
	     printf("Attached to pid %d\n", task->pid);
	    task->has_run = 1;
	}

	if(debuglevel & DEBUG0) 
		printf("Exit Log -- pid %d\n", task->pid);

	ztrace(ZTRACE_CLEARBRANCH, task->pid, 0, 0);

	ztrace(ZTRACE_EXIT, task->pid,0,0);

	ztrace(ZTRACE_DETACH, task->pid,0,0);
	unmap_task(task);
	return ;
}

void do_replay(void) {
	struct rlog_header * rh;
	struct rlog_data * rd;
	while(1) {
	    rh=zr_next_schedrec();
	    rd=DATAOF(rh);

	    switch(rh->type) {
	    	case ENTRY_SCHED:
			if(debuglevel & DEBUG0)
			 printf("Schedule entry %d\n", logcount[0]++);
			do_schedule_entry((struct rlog_schedule_entry *)rd);
			break;
		case ENTRY_SIGNAL:
			if(debuglevel & DEBUG0)
			 printf("Signal entry %d\n", logcount[2]++);
			deliver_signals((struct rlog_entry_signal *)rd);
			if(stop_request) {
				struct ztrace_state_struct zt_state;
				/* UML specific! */
				if(ztrace(ZTRACE_GETSTATE, tasklist[0].pid,
						0, &zt_state)) {
					perror("getstate");
				}
				if(zt_state.revirt_flags & REPLAY_SYSCALL) {
				 printf("Stop request received, but still in syscall.  Carrying on.\n");
				} else {
				 printf("Stop request received, exiting after signal entry.\n");
				 exit(1);
				}
			}
			break;
		case ENTRY_EXIT:
			if(debuglevel & DEBUG0)
			 printf("Exit entry %d\n", logcount[3]++);
			do_exit_entry((struct rlog_exit_entry *)rd);
			break;
		case ENTRY_TOP:
		case ENTRY_EXECPARM:
			if(debuglevel & DEBUG0)
			 printf("Silly entry %d, skipping.\n", logcount[4]++);
			break;
		default:
			fprintf(stderr, "Unknown log: 0x%x\n", rh->type);
			debug_pause();
			break;
	    }
    	    if (!mapl)
		break;
	}
}

void attach_to_task(int pid) {
	try(!ztrace(ZTRACE_ATTACH,pid,0,0));
	fprintf(stderr, "Attached to process %d\n", pid);
}

int main(int argc,char **argv) {
	int i;
	if (argc<3 || argc > 4) {
		fprintf(stderr, "usage: %s <pid> log [debugmask]\n",argv[0]);
		exit(1);
	}
	if(argc == 4) 
		debuglevel=atoi(argv[3]);
	else
		debuglevel=DEBUG0;
	if(signal(SIGINT, sigint)==SIG_ERR) {
		perror("installing sigint handler");
	}
	if(signal(SIGUSR1, sigusr1)==SIG_ERR) {
		perror("installing sigint handler");
	}
	printf("zr pid: %d\n", getpid());

#if 0
	if(debuglevel & THROTTLE) { 
		if(ztrace(ZTRACE_SET_OPT, getpid(), REPLAY_OPT_PACKETSIZE,
					2000)<0) {
			perror("Setting network throttle");
		}
		if(ztrace(ZTRACE_SET_OPT, getpid(), REPLAY_OPT_THROTTLE_NETWORK,
					REPLAY_OPT_THROTTLE_NETWORK)<0) {
			perror("Setting network throttle");
		}
	} else {
		if(ztrace(ZTRACE_SET_OPT, getpid(), 
					REPLAY_OPT_THROTTLE_NETWORK, 0)<0) {
			perror("Setting network throttle");
		}
	}
	if(ztrace(ZTRACE_SET_OPT, getpid(), 
				REPLAY_OPT_LOGSYS_ON_REPLAY, 
				0
//				REPLAY_OPT_LOGSYS_ON_REPLAY
				)<0) {
		perror("Setting logsys on replay");
	}
#endif
	bzero(&rstat, sizeof(rstat));

	zr_open_log(argv[2]);
	mapl= 0;
	new_task(atoi(argv[1]));

	do_replay();

	printf("ztrace calls:\n");
	for(i=0; i<8; i++) {
		printf("[%2d] %6d [%2d] %6d [%2d] %6d [%2d] %6d\n",
			i, rstat.ztrace_call[i],
			i+8, rstat.ztrace_call[i+8],
			i+16, rstat.ztrace_call[i+16],
			i+24, rstat.ztrace_call[i+24]);
	}
	printf("repstring nr: %d cnt:%d\nrepstring sstep nr: %d cnt: %d\n", 
			rstat.nr_repstring, rstat.cnt_repstring,
			rstat.nr_repstring_sstep, rstat.cnt_repstring_sstep);
	printf("repstring distribution: %d %d %d %d %d %d %d %d\n",
			rstat.replen[0],
			rstat.replen[1],
			rstat.replen[2],
			rstat.replen[3],
			rstat.replen[4],
			rstat.replen[5],
			rstat.replen[6],
			rstat.replen[7]);
	printf("Highest overflow error: %d\n", rstat.max_ovf_error);
	printf("Max hw_branch: %lld\nMax hw_intr: %d\nMax real_count: %lld\n",
			rstat.max_bc,
			rstat.max_intr,
			rstat.max_real);
	return 0;
}
