/*
 * $Id: lilo.c,v 1.1.1.1 2003/05/27 21:45:34 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <netinet/in.h>
#include <sys/socket.h>

#include <sys/types.h>
#include <sys/time.h>

/* can't use <asm/page.h> for assembler code */
/* #include <asm/page.h> */
#define PAGE_SIZE	4096
#include <asm/unistd.h>

#include <elf.h>

#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <net/if.h>

#include "external_constants.h"
#include "lilo.h"
#include "lotol_fdesc.h"
#include "fi_if_lilo.h"


/* see arch/i386-linux/kernel/head.c */
#define PARAM (MEM_BASE + 0x90000)
/* ... */
#define CL_MAGIC_ADDR (*(unsigned short *) (PARAM+0x20))
#define CL_OFFSET (*(unsigned short *) (PARAM+0x22))
/* ... */
#define MOUNT_ROOT_RDONLY (*(unsigned short *) (PARAM+0x1F2))
#define RAMDISK_FLAGS (*(unsigned short *) (PARAM+0x1F8))
/* ... */
#define LOADER_TYPE (*(unsigned char *) (PARAM+0x210))
#define KERNEL_START (*(unsigned long *) (PARAM+0x214))
#define INITRD_START (*(unsigned long *) (PARAM+0x218))
#define INITRD_SIZE (*(unsigned long *) (PARAM+0x21c))
/* ... */
#define COMMAND_LINE ((char *) (PARAM+2048))
#define COMMAND_LINE_SIZE 256

#define CL_MAGIC 0xA33F

char *progname;
int opt_p = 0;
char *sys;

unsigned long mem_size;

typedef struct {
	pid_t pid;
	char *map;
	int size;
} video_struct;

video_struct *video;

/* convert macro into string */
#define STR2(x)	#x
#define STR(x)	STR2(x)

/* static */ void
do_map_code_begin(void)
{
	/*
	 * will receive
	 * - mem_base in %eax
	 * - mem_size in %ebx
	 * - current pointer in %ecx
	 */
	asm(
	"	pushl $" STR(PAGE_SIZE) "\n"		/* offset */
	"	pushl $" STR(FD_MEMORY) "\n"		/* fd */
	"	pushl $" STR(MAP_SHARED|MAP_FIXED) "\n"	/* flags */
	"	pushl $" STR(PROT_READ|PROT_WRITE|PROT_EXEC) "\n" /* prot */
	"	subl $" STR(PAGE_SIZE) ", %%ebx\n"	/* len */
	"	pushl %%ebx\n"
	"	pushl $" STR(MEM_BASE+PAGE_SIZE) "\n"	/* addr */
	"	movl %%esp, %%ebx\n"
	"	movl $" STR(__NR_mmap) ", %%eax\n"
	"	int $0x80\n"

	: /* no output */
	: /* no input */
	: "memory"
	);
}
/* static */ void
do_map_code_end(void)
{
}

/* static */ void
do_unmap_code_begin(void)
{
	/*
	 * will receive
	 * - mem_base in %eax
	 * - mem_size in %ebx
	 * - current pointer in %ecx
	 */
	asm(
	"	movl %%ecx, %%edx\n"		/* save current pointer */

	"	movl $" STR(MEM_BASE+PAGE_SIZE) ", %%ebx\n"	/* addr */
	"	movl %%edx, %%ecx\n"				/* len */
	"	addl $" STR(PAGE_SIZE) ", %%ecx\n"
	"	subl %%ebx, %%ecx\n"
	"	movl $" STR(__NR_munmap) ", %%eax\n"
	"	int $0x80\n"

	"	movl %%edx, %%ebx\n"				/* addr */
	"	addl $" STR(PAGE_SIZE) "*2, %%ebx\n"
	"	movl $0x80000000, %%ecx\n"			/* len */
	"	subl %%ebx, %%ecx\n"
	"	movl $" STR(__NR_munmap) ", %%eax\n"
	"	int $0x80\n"

	: /* no output */
	: /* no input */
	: "memory"
	);
}
/* static */ void
do_unmap_code_end(void)
{
}

/* Initialize memory */
int
ummem_init(void)
{
	int config_fd;
	char c;
	char map_path[256];
	int map_fd;
	unsigned long addr;
	int i;
	int ret;

	/* read memory size */
	config_fd = open("memory", O_RDONLY);
	if (config_fd < 0) {
		fprintf(stderr, "%s: Can't open %s: %s.\n", progname,
				"memory", strerror(errno));
		exit(1);
	}

	mem_size = 0;
	for (i = 0; i < 8; i++) {
		ret = read(config_fd, &c, 1);
		assert(ret == 1);	/* improve error handling FIXME VOSSI */
		mem_size *= 16;
		if ('0' <= c && c <= '9') {
			mem_size += c - '0';
		} else if ('A' <= c && c <= 'F') {
			mem_size += c - 'A' + 10;
		} else if ('a' <= c && c <= 'f') {
			mem_size += c - 'a' + 10;
		} else {
			assert(0);	/* improve error handling FIXME VOSSI */
		}
	}
	assert(0 < mem_size);	/* improve error handling FIXME VOSSI */

	ret = close(config_fd);
	assert(0 <= ret);

	/* create memory file */
	sprintf(map_path, "/tmp/page%08x", getpid());
	map_fd = open(map_path, O_RDWR | O_CREAT | O_EXCL, 0700);
	assert(0 <= map_fd);

	ret = unlink(map_path);
	assert(0 <= ret);

	ret = dup2(map_fd, FD_MEMORY);
	assert(ret == FD_MEMORY);
	ret = close(map_fd);
	assert(0 <= ret);

	for (addr = MEM_BASE; addr < MEM_BASE + mem_size; addr += PAGE_SIZE) {
		static char zero[0x10000];

		ret = write(FD_MEMORY, zero, PAGE_SIZE);
		assert(ret == PAGE_SIZE);
	}

	return 0;
}

/* Initialize console */
int
umlcon_init(void)
{
	sysfd_t fd;
	int console_socket;
	char conf[8+1+4+1];
	unsigned long ipaddr;
	unsigned short ipport;
	struct sockaddr_in addr;
	int ret;

	fd = open("console", O_RDONLY, 0);
	assert(0 <= (int)fd && FD_LOTOL_MIN > (int)fd);

	ret = read(fd, conf, sizeof(conf));
	assert(sizeof(conf) == ret);

	ret = close(fd);
	assert(0 <= ret);

	sscanf(conf, "%lx:%*x\n", &ipaddr);
	sscanf(conf, "%*x:%hx\n", &ipport);

	/* create socket */
	console_socket = socket(PF_INET, SOCK_STREAM, 0);
	assert(0 <= console_socket && FD_LOTOL_MIN > console_socket);

	/* connect to console process */
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(ipaddr);
	addr.sin_port = htons(ipport);
	ret = connect(console_socket,
			(struct sockaddr *) &addr, sizeof(addr));
	assert(0 <= ret);

	ret = dup2(console_socket, FD_CONSOLE);
	assert(FD_CONSOLE == ret);

	ret = close(console_socket);
	assert(0 <= ret);

	return 0;
}

/* Try to initialize floppy drives */
int
umfloppy_init(void)
{
	char drive, offset;
	char path[32];
	char name[] = "fd.";
	sysfd_t fd;
	int ret;

	for (drive = FD_FIRST_DRIVE; drive <= FD_LAST_DRIVE; drive++) {
		sprintf(name, "fd%c", drive);
		sprintf(path, "%s/media",  name);
		fprintf(stderr,"path=%s\n",path);fflush(stderr);

		offset = FD_OFFSET(drive);
		fprintf(stderr,"   offset=%d\n",(int)offset);fflush(stderr);

		/* media */
		errno=0;
		fd = open(path, O_RDWR, 0);
		fprintf(stderr,"   media=%d (%s)\n",fd,strerror(errno));fflush(stderr);

		if ((int)fd < 0 ) {
			errno=0;
			fd = open(path, O_RDONLY, 0);
			fprintf(stderr,"   media-ro=%d (%s)\n",fd,strerror(errno));fflush(stderr);
		}
		if (0<(int)fd){
			ret = dup2(fd, (FD_FD0_MEDIA + offset));
			fprintf(stderr,"   dup2-media=%d[%d] (%s)\n",ret,(FD_FD0_MEDIA + offset),strerror(errno));fflush(stderr);
			assert((FD_FD0_MEDIA + offset) == ret);

			ret = close(fd);
			assert(0 <= ret);
		}
	}
	return 0;
}
/* Try to initialize IDE drives */
int
umide_init(void)
{
	char drive, offset;
	char path[32];
	char name[] = "hd.";
	sysfd_t fd;
	int ret;

	for (drive = HD_FIRST_IDE_DRIVE; drive <= HD_LAST_IDE_DRIVE; drive++) {
		sprintf(name, "hd%c", drive);
		sprintf(path, "%s/model",  name);
		fprintf(stderr,"path=%s\n",path);fflush(stderr);

		offset = HD_OFFSET(drive) ;
		fprintf(stderr,"   offset=%d\n",(int)offset);fflush(stderr);

		/* model */
		errno=0;
		fd = open(path, O_RDONLY, 0);
		fprintf(stderr,"   model=%d (%s)\n",fd,strerror(errno));fflush(stderr);

		if (0 <= (int)fd) {
			assert(FD_LOTOL_MIN > (int)fd);

			errno=0;
			ret = dup2(fd, (FD_HDA_MODEL + offset)); /* XXX */
			fprintf(stderr,"   dup2-model=%d[%d] (%s)\n",ret,(FD_HDA_MODEL + offset),strerror(errno));fflush(stderr);
			assert((FD_HDA_MODEL + offset) == ret);

			ret = close(fd);
			assert(0 <= ret);

			/* config (hd) or pconfig (cdrom) */
			sprintf(path, "%s/config", name);
			fprintf(stderr,"   path=%s\n",path);fflush(stderr);
			errno=0;
			fd = open(path, O_RDONLY, 0);
			fprintf(stderr,"   config=%d (%s)\n",fd,strerror(errno));fflush(stderr);

			if (0 <= (int)fd) { /* harddisk */
				assert(FD_LOTOL_MIN > (int)fd);

				errno=0;
				ret = dup2(fd, (FD_HDA_CONFIG + offset));
				fprintf(stderr,"   dup2-config=%d[%d] (%s)\n",ret, (FD_HDA_CONFIG + offset),strerror(errno));fflush(stderr);
				assert((FD_HDA_CONFIG + offset) == ret);

				ret = close(fd);
				assert(0 <= ret);
				/* media */
				sprintf(path, "%s/media", name);

				errno=0;
				fd = open(path, O_RDWR, 0);
				fprintf(stderr,"   media-hd=%d (%s)\n",fd,strerror(errno));fflush(stderr);
				assert(0 <= ret);

			} else { /* cdrom */
				sprintf(path, "%s/pconfig", name);
				fprintf(stderr,"   path=%s\n",path);fflush(stderr);
				errno=0;
				fd = open(path, O_RDONLY, 0);
				fprintf(stderr,"   pconfig=%d (%s)\n",fd,strerror(errno));fflush(stderr);
				/*
				 * We need either config or pconfig!
				 * Therefore this open must not fail!
				 */
				assert(0 <= (int)fd && FD_LOTOL_MIN > (int)fd);

				errno=0;
				ret = dup2(fd, (FD_HDA_PCONFIG + offset));
				fprintf(stderr,"   dup2-pconfig=%d[%d] (%s)\n",ret,(FD_HDA_PCONFIG + offset),strerror(errno));fflush(stderr);
				assert((FD_HDA_PCONFIG + offset) == ret);

				ret = close(fd);
				assert(0 <= ret);
				/* media */
				sprintf(path, "%s/media", name);

				errno=0;
				fd = open(path, O_RDONLY, 0);
				fprintf(stderr,"   media=%d (%s)\n",fd,strerror(errno));fflush(stderr);
				assert(0 <= ret);

			}

			errno=0;
			ret = dup2(fd, (FD_HDA_MEDIA + offset));
			fprintf(stderr,"   dup2-media=%d[%d] (%s)\n",ret,(FD_HDA_MEDIA + offset),strerror(errno));fflush(stderr);
			assert((FD_HDA_MEDIA + offset) == ret);

			ret = close(fd);
			assert(0 <= ret);
		}
	}

	return 0;
}


/* Try to initialize ethernet */
int
umnet_init(void)
{
	sysfd_t fd;
	sysfd_t tfd;
	sysfd_t skfd;
	sysfd_t tffd;
	int eth = 0;
	int ret;
	int i;
	int flags;
	unsigned int myself;
	char name[] = "eth.";
	char line[80];
	char nibble1, nibble2, digit1, digit2;
	char *sret;
	unsigned short ipport;
	struct sockaddr_in addr;
	FILE *sfd, *sfd1;
	struct ifreq ifr;
	struct ifreq ifr1;

	/*
	for (eth = 0; eth < MAX_ETH_CARDS; eth++) {
		sprintf(name, "eth%d", eth);

		fd = open(name, O_RDONLY, 0);

		if (0 <= (int)fd) {
			assert(FD_LOTOL_MIN > (int)fd);

			ret = dup2(fd, (FD_ETH0_CONF + eth));
			assert((FD_ETH0_CONF + eth) == ret);

			ret = close(fd);
			assert(0 <= ret);
		} else
			continue;

		fd = FD_ETH0_CONF + eth;

		// open an I/O-stream, much easier to handle... 
		sfd = fdopen(fd, "r");
		assert(NULL != sfd);

		sret = fgets(line, sizeof(line) - 1, sfd);
		assert(sret == line);

		sscanf(line, "%x", &myself);	

		// Skip over to "our" line
		for (i = 0; i <= myself; i++) {
			sret = fgets(line, sizeof(line) - 1, sfd);
			assert(sret == line);
		}

		// reset descriptor
		ret = lseek(fd, 0, 0);
		assert(0 == ret);

		sscanf(line, "%*x:%*x:%*x:%*x:%*x:%*x %*x:%hx\n", &ipport);

		// create socket
		fd = socket(PF_INET, SOCK_DGRAM, 0);
		assert(0 <= (int)fd && FD_LOTOL_MIN > (int)fd);	
		// XXX Maybe we should allow a failure?

		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(ipport);
		addr.sin_addr.s_addr = INADDR_ANY;

		ret = dup2(fd, (FD_ETH0 + eth));
		assert((FD_ETH0 + eth) == ret);

		ret = close(fd);
		assert(0 <= ret);

		fd = FD_ETH0 + eth;

		fprintf(stderr, "%s: binding %s(%d) [0x%04x].\n", progname, name, fd, ipport);fflush(stderr);
		ret = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
		if (ret < 0) {
			fprintf(stderr, "%s: Can't bind %s(%d) [0x%04x]: %s.\n", progname,
				name, fd, ipport, strerror(errno));fflush(stderr);
		}
		assert(0 <= ret);

		flags = fcntl(fd, F_GETFL, 0);
		flags |= O_ASYNC;
		flags |= O_NONBLOCK;
		ret = fcntl(fd, F_SETFL, flags);
		assert(0 <= ret);

		ret = fcntl(fd, F_SETOWN, -1 * getpgrp());
		assert(0 <= ret);
	}
	*/	

	/*
	 * TAP INIT
	 */
	// I am not sure why this is needed, but perhaps the answer lies in 
	// UMLsrc/linux/umlinux/build-2.2.20/drivers/net/net_init.c ??? -M

	sprintf(name, "eth%d", eth);

	fd = open(name, O_RDWR, 0);
	assert(0 <= fd);

	if (0 <= (int)fd) {
		assert(FD_LOTOL_MIN > (int)fd);
		
		ret = dup2(fd, (FD_ETH0_CONF));
		assert((FD_ETH0_CONF) == ret);
		
		ret = close(fd);
		assert(0 <= ret);
	}
	
	fd = FD_ETH0_CONF;

	// open an I/O-stream, much easier to handle... 
	sfd = fdopen(fd, "r+");
	assert(NULL != sfd);

	sret = fgets(line, sizeof(line) - 1, sfd);
	assert(sret == line);
	
	sscanf(line, "%x", &myself);	

	// Skip over to "our" line
	for (i = 0; i <= myself; i++) {
		sret = fgets(line, sizeof(line) - 1, sfd);
		//fprintf(stderr, "Line: %s", line);
		assert(sret == line);
	}

	// reset descriptor
	ret = lseek(fd, 0, 0);
	assert(0 == ret);

	// File needs to be reopened?
	ret = fclose(sfd);
	assert(ret == 0);

	sscanf(line, "%*x:%*x:%*x:%*x:%*x:%*x %*x:%hx\n", &ipport);
	
	// create socket
	fd = socket(PF_INET, SOCK_DGRAM, 0);
	assert(0 <= (int)fd && FD_LOTOL_MIN > (int)fd);	
	// XXX Maybe we should allow a failure?
	
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(ipport);
	addr.sin_addr.s_addr = INADDR_ANY;
	
	ret = dup2(fd, (FD_ETH0));
	assert((FD_ETH0) == ret);
	
	ret = close(fd);
	assert(0 <= ret);
	
	fd = FD_ETH0;
	
	fprintf(stderr, "%s: binding %s(%d) [0x%04x].\n", progname, name, fd,
		ipport);
	fflush(stderr);
	ret = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
	if (ret < 0) {
		fprintf(stderr, "%s: Can't bind %s(%d) [0x%04x]: %s.\n", 
			progname, name, fd, ipport, strerror(errno));
		fflush(stderr);
	}
	assert(0 <= ret);
	
	flags = fcntl(fd, F_GETFL, 0);
	flags |= O_ASYNC;
	flags |= O_NONBLOCK;
	ret = fcntl(fd, F_SETFL, flags);
	assert(0 <= ret);
	
	ret = fcntl(fd, F_SETOWN, -1 * getpgrp());
	assert(0 <= ret);

	// The stuff below here is surely needed -M

	tfd = open("/dev/net/tun", O_RDWR);
	fprintf(stderr, "%s: tfd: %d\n", progname, tfd);
	if (tfd < 0)
		fprintf(stderr, "\terrno: %s\n", strerror(errno));
	assert(0 <= (int)tfd && FD_LOTOL_MIN > (int)tfd);
	
	errno = 0;
	ret = dup2(tfd, (FD_TAP));
	fprintf(stderr,"   dup2-TAP=%d[%d] (%s)\n", ret, FD_TAP, 
		strerror(errno));
	fflush(stderr);
	assert(FD_TAP == ret);
	
	ret = close(tfd);
	assert(0 <= ret);
	
	tfd = FD_TAP;
	
	memset(&ifr, 0, sizeof(ifr));
	
	ifr.ifr_flags = (IFF_TAP | IFF_NO_PI);
	strncpy(ifr.ifr_name, "", sizeof(ifr.ifr_name) - 1);

	if(ioctl(tfd, TUNSETIFF, (void *) &ifr) < 0) {
		perror("TUNSETIFF");
		exit(1);
	}
   
	if(ioctl(tfd, TUNSETNOCSUM, 1) < 0) {
		perror("TUNSETNOCSUM");
		exit(1);
	}
	/*
	if(ioctl(tfd, TUNSETPERSIST, 1) < 0){
		perror("TUNSETPERSIST");
		exit(1);
	}
	
	if(ioctl(tfd, TUNSETOWNER, 0) < 0){
		perror("TUNSETOWNER");
		exit(1);
	} 

	fprintf(stderr, "Set '%s' persistent and owned by uid %d\n",
	        ifr.ifr_name, 0);
	*/

	fprintf(stderr, "***** TAP: Set '%s' *****\n", ifr.ifr_name);

	// Create a channel to the NET kernel.
	if ((skfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	
	// Get the hw address of the tap interface.
	bzero(&ifr1, sizeof(ifr1));
	strcpy(ifr1.ifr_name, ifr.ifr_name);
	if (ioctl(skfd, SIOCGIFHWADDR, &ifr1) < 0) {
		perror ("ioctl(SIOCGIFHWADDR)");
		exit (1);
	}
	printf("HW address of tap interface is ");
	for (i = 0; i < 6; i++) {
		printf("%02x%s", (ifr1.ifr_hwaddr.sa_data [i]) & 0xff,
			(i == 5) ? "\n" : ":");
	}
	
	// Write it to the eth0 file.
	sprintf(name, "eth%d", eth);

	fd = open(name, O_RDWR, 0);
	assert(0 <= fd);

	if (0 <= (int)fd) {
		assert(FD_LOTOL_MIN > (int)fd);
		
		ret = dup2(fd, (FD_ETH0_CONF));
		assert((FD_ETH0_CONF) == ret);
		
		ret = close(fd);
		assert(0 <= ret);
	}

	fd = FD_ETH0_CONF;

	// open an I/O-stream, much easier to handle... 
	sfd1 = fdopen(fd, "r+");
	assert(NULL != sfd1);

	sret = fgets(line, sizeof(line) - 1, sfd1);
	assert(sret == line);
	
	sscanf(line, "%x", &myself);	

	// Skip over to begining of "our" line
	for (i = 0; i < myself ; i++) {
		sret = fgets(line, sizeof(line) - 1, sfd1);
		//fprintf(stderr, "Line: %s", line);
		assert(sret == line);
	}
	for (i = 0; i < 6; i++) {
		nibble1 = (ifr1.ifr_hwaddr.sa_data[i] & 0xf0) >> 4;
		nibble2 = ifr1.ifr_hwaddr.sa_data[i] & 0x0f;
		if ( nibble1 < 10 )
			digit1 = nibble1 + '0';
		else
			digit1 = nibble1 - 10 + 'a';
		if ( nibble2 < 10 )
			digit2 = nibble2 + '0';
		else
			digit2 = nibble2 - 10 + 'a';
		ret = fputc(digit1, sfd1);
		assert(ret == digit1);
		ret = fputc(digit2, sfd1);
		assert(ret == digit2);
		fputc( ((i == 5) ? ' ' : ':') , sfd1);
	}
	
	ret = fflush(sfd1);
	assert( ret == 0 );

	// Write interface name to tap file.
	tffd = creat("tap", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	assert(0 <= tffd);

	ret = write(tffd, ifr.ifr_name, strlen(ifr.ifr_name));
	assert(ret == strlen(ifr.ifr_name));

	ret = close(tffd);
	assert( ret == 0 );


       	flags = fcntl(tfd, F_GETFL, 0);
	flags |= O_ASYNC;
	flags |= O_NONBLOCK;
	ret = fcntl(tfd, F_SETFL, flags);
	assert(0 <= ret);
	
	ret = fcntl(tfd, F_SETOWN, -1 * getpgrp());
	assert(0 <= ret);

	return 0;
}

static video_struct *video_init(int argc, char **argv)
{
  int i;
  int ret, len;
  video_struct *v = NULL;

  for(i=0; i<argc; i++)
	  if(!strncmp(argv[i], "vga=", 4))
		  switch(atoi(argv[i]+4)) {
		  case 769: /* 640x480x8 */
			  if(!(v = (video_struct*)malloc(sizeof(video_struct)))) {
				  fprintf(stderr, "%s: video_struct malloc failure\n", __FILE__);
				  return NULL;
			  }
			  v->size = 640*480;
			  v->map = (char*)mmap(0, v->size, PROT_READ, MAP_SHARED, FD_MEMORY, 0xa0000);
			  assert(0 < (int)v->map);
			  break;
		  default: /* con */
			  break;
		  }

  if (! v) {
	  return (video_struct *) 0;
  }

  switch(v->pid = fork()) {
  case -1:
	  fprintf(stderr, "oops\n");
	  exit(-1);
  case 0:
	  while(1) {
		  len = 0;
		  do {
			  ret = write(FD_CONSOLE, v->map + len, v->size - len);
			  if (ret < 0) {
				  perror("write");
				  sleep(1);
				  ret = 0;
			  }
			  assert(0 <= ret);
			  len += ret;
		  } while((len % v->size) != 0);
		  sleep(1);
	  }
  }

  return v;
}

/* Initialize hardware */
int
hwinitialize(int argc, char **argv)
{
	int ret;

	ret = ummem_init();
	assert(0 == ret);

	ret = umlcon_init();
	assert(0 == ret);

	ret = umfloppy_init();
	assert(0 == ret);

	ret = umide_init();
	assert(0 == ret);

	ret = umnet_init();
	assert(0 == ret);

	video = video_init(argc, argv);

	return 0;	
}

/* Try to shutdown ethernet */
int
umnet_shutdown(void)
{
	/*
	int eth;

	for (eth = 0; eth < MAX_ETH_CARDS; eth++) {
		(void) close(FD_ETH0 + eth);
	}
	*/
	(void) close(FD_ETH0);
	(void) close(FD_TAP);

	return 0;
}

/* Try to shutdown IDE drives */
int
umide_shutdown(void)
{
	int drive;
	int offset;

	for (drive = HD_FIRST_IDE_DRIVE; drive <= HD_LAST_IDE_DRIVE; drive++) {
		offset = HD_OFFSET(drive);
		(void) close(FD_HDA_MEDIA + offset);
	}

	return 0;
}

/* Try to shutdown floppy drives */
int
umfloppy_shutdown(void)
{
	int drive;
	int offset;

	for (drive = FD_FIRST_DRIVE; drive <= FD_LAST_DRIVE; drive++) {
		offset = FD_OFFSET(drive);
		(void) close(FD_FD0_MEDIA + offset);
	}

	return 0;
}
/* Shutdown console */
int
umlcon_shutdown(void)
{
	int ret;

	ret = close(FD_CONSOLE);
	assert(0 <= ret);

	return 0;
}

int
ummem_shutdown(void)
{
	int ret;

	ret = close(FD_MEMORY);
	assert(0 <= ret);

	return 0;
}

int
video_shutdown(void)
{
	if (!video)
		kill(video->pid, SIGKILL);

	return 0;
}

/* Shutdown hardware */
int
hwshutdown(void)
{
	int ret;

	ret = video_shutdown();
	assert(0 == ret);

	ret = umnet_shutdown();
	assert(0 == ret);

	ret = umide_shutdown();
	assert(0 == ret);

	ret = umfloppy_shutdown();
	assert(0 == ret);

	ret = umlcon_shutdown();
	assert(0 == ret);

	ret = ummem_shutdown();
	assert(0 == ret);

	return 0;	
}

static void
(*do_load(int argc, char **argv))(void)
{
	int sys_fd;
	Elf32_Ehdr elf_hdr;
	Elf32_Phdr elf_phdr;
	int p;
	char *initrd;
	int initrd_fd;
	unsigned long initrd_start;
	unsigned long initrd_end;
	int i;
	void (*func)(void);
	int ret;
	int a;
	struct timeval tvStart, tvStop;

	/* ========== */
	/* map memory */
	/* ========== */

	printf("my fd memory is %d\n",FD_MEMORY);

	ret = (int) mmap((char *) MEM_BASE, mem_size,
			PROT_READ | PROT_WRITE | PROT_EXEC,
			MAP_SHARED | MAP_FIXED, FD_MEMORY, 0);
	assert(ret == MEM_BASE);



	/* ========================================= */
	/* write trampoline code to address MEM_BASE */
	/* ========================================= */

	memcpy((char *) MEM_BASE +    0, do_map_code_begin,
			(unsigned long) do_map_code_end
			- (unsigned long) do_map_code_begin);

	memcpy((char *) MEM_BASE + PAGE_SIZE / 2, do_unmap_code_begin,
			(unsigned long) do_unmap_code_end
			- (unsigned long) do_unmap_code_begin);

	ret = mprotect((char *) MEM_BASE, PAGE_SIZE, PROT_READ | PROT_EXEC);
	assert(0 <= ret);

	assert(*(int *) do_map_code_begin == *(int *) MEM_BASE);
	assert(*(int *) do_unmap_code_begin == *(int *) (MEM_BASE + PAGE_SIZE / 2));



	/* ===================== */
	/* load system to memory */
	/* ===================== */

	/* assuming ELF binary (FIXME!) */

	fprintf(stderr,"sys=%s\n",sys);fflush(stderr);
	sys_fd = open(sys, O_RDONLY);
	assert(0 <= sys_fd);

	/* read elf header */
	ret = read(sys_fd, (void *) &elf_hdr, sizeof(elf_hdr));
	assert(ret == sizeof(elf_hdr));

#if 0
	printf("e_type: %d\n", elf_hdr.e_type);
	printf("e_machine: %d\n", elf_hdr.e_machine);
	printf("e_version: %d\n", elf_hdr.e_version);
	printf("e_entry: %p\n", elf_hdr.e_entry);
	printf("e_phoff: %d\n", elf_hdr.e_phoff);
	printf("e_shoff: %d\n", elf_hdr.e_shoff);
	printf("e_flags: %d\n", elf_hdr.e_flags);
	printf("e_ehsize: %d\n", elf_hdr.e_ehsize);
	printf("e_phentsize: %d\n", elf_hdr.e_phentsize);
	printf("e_phnum: %d\n", elf_hdr.e_phnum);
	printf("e_shentsize: %d\n", elf_hdr.e_shentsize);
	printf("e_shnum: %d\n", elf_hdr.e_shnum);
	printf("e_shstrndx: %d\n", elf_hdr.e_shstrndx);
#endif

	/* read elf program header */
	for (p = 0; p < elf_hdr.e_phnum; p++) {
		ret = lseek(sys_fd, elf_hdr.e_phoff + p * elf_hdr.e_phentsize,
				SEEK_SET);
		assert(ret == elf_hdr.e_phoff + p * elf_hdr.e_phentsize);

		ret = read(sys_fd, (void *) &elf_phdr, sizeof(elf_phdr));
		assert(ret == sizeof(elf_phdr));

#if 0
		printf("p_type: %d\n", elf_phdr.p_type);
		printf("p_offset: 0x%x\n", elf_phdr.p_offset);
		printf("p_vaddr: %p\n", elf_phdr.p_vaddr);
		printf("p_paddr: %p\n", elf_phdr.p_paddr);
		printf("p_filesz: 0x%x\n", elf_phdr.p_filesz);
		printf("p_memsz: 0x%x\n", elf_phdr.p_memsz);
		printf("p_flags: 0x%x\n", elf_phdr.p_flags);
		printf("p_align: 0x%x\n", elf_phdr.p_align);
#endif

		ret = lseek(sys_fd, elf_phdr.p_offset, SEEK_SET);
		assert(ret == elf_phdr.p_offset);

		ret = read(sys_fd, (char *) elf_phdr.p_vaddr, elf_phdr.p_filesz);
		if (ret < 0) {
			fprintf(stderr, "%s: %s: %s.\n",
					progname, sys, strerror(errno));
			exit(1);
		}
		if (ret < elf_phdr.p_filesz) {
			fprintf(stderr, "%s: %s: %s.\n",
					progname, sys, "Short read");
			exit(1);
		}
		assert(ret == elf_phdr.p_filesz);

		if (elf_phdr.p_filesz != elf_phdr.p_memsz) {
			memset((char *) elf_phdr.p_vaddr + elf_phdr.p_filesz, 0,
					elf_phdr.p_memsz - elf_phdr.p_filesz);
		}
	}

	ret = close(sys_fd);
	assert(0 <= ret);

	/* ===================== */
	/* load initrd to memory */
	/* ===================== */

	for (i = 0; ; i++) {
		if (i == argc) {
			/* no initrd */
			initrd = NULL;
			break;
		}
		if (strncmp(argv[i], "initrd=", 7) == 0) {
			initrd = argv[i] + 7;
			break;
		}
	}
	if (initrd == NULL) {
		initrd_start = 0;
		initrd_end = 0;

	} else {
		initrd_fd = open(initrd, O_RDONLY);
		if (initrd_fd < 0) {
			fprintf(stderr, "%s: %s: %s.\n",
					progname, initrd,
					strerror(errno));
			exit(1);
		}

		initrd_start = mem_size / 2; /* FIXME */
		initrd_end = initrd_start;
		for (;;) {
			int len;

			len = read(initrd_fd, (char *) MEM_BASE + initrd_end,
					1024);
			if (len < 0) {
				fprintf(stderr, "%s: %s: %s.\n",
						progname, initrd,
						strerror(errno));
				exit(1);
			}
			if (len == 0) {
				break;
			}
			initrd_end += len;
		}

		ret = close(initrd_fd);
		assert(0 <= ret);
	}


	/* ============= */
	/* set parameter */
	/* ============= */

	for (i = 0; ; i++) {
		if (i == argc) {
			MOUNT_ROOT_RDONLY = 1;
			break;
		}
		if (strcmp(argv[i], "read-only") == 0) {
			MOUNT_ROOT_RDONLY = 1;
			break;
		} else if (strcmp(argv[i], "read-write") == 0) {
			MOUNT_ROOT_RDONLY = 0;
			break;
		}
	}
	RAMDISK_FLAGS = 0;

	LOADER_TYPE = 1;	/* must be != 0 */
	/* KERNEL_START = 0;	FIXME */
	INITRD_START = initrd_start;
	INITRD_SIZE = initrd_end - initrd_start;

	CL_MAGIC_ADDR = CL_MAGIC;
	CL_OFFSET = (char *) COMMAND_LINE - (char *) PARAM;
	strcpy(COMMAND_LINE, "BOOT_IMAGE=");
	strcat(COMMAND_LINE, sys);
	for (i = 0; i < argc; i++) {
		strcat(COMMAND_LINE, " ");
		strcat(COMMAND_LINE, argv[i]);
	}

	func = (void (*)(void)) elf_hdr.e_entry;




	/*
	gettimeofday(&tvStart,NULL);
	for(a=0;a<500000;a++) {
		ret = mprotect((char *) 0x70000000,0x04000000,PROT_NONE);
		assert(ret == 0);
		
		ret = mprotect((char *) 0x70000000,0x04000000,PROT_READ | PROT_WRITE | PROT_EXEC);
		assert (ret == 0);
	}
	gettimeofday(&tvStop,NULL);
	
	printf("testMap sec = %ld, usec = %ld\n",
	       tvStop.tv_sec - tvStart.tv_sec,
	       tvStop.tv_usec - tvStart.tv_usec);
	
	exit(0);

	*/

	return func;
}

static void
usage(int retval)
{
	fprintf(stderr, "Usage: %s [-p] <system>\n", progname);
	fprintf(stderr, "\t-p: protect kernel memory\n");
	exit(retval);
}

int
main(int argc, char **argv)
{
	void (*entry)(void);
	extern int tracer(int pid);
	FILE *fp;
	int c;
	int pid;
	int ret;
	int myfd;
	void *mymem;


	/*
	 * get program name
	 */
	progname = *argv;

	/*
	 * get options
	 */
	while ((c = getopt(argc, argv, "p")) != -1) {
		switch (c) {
		case 'p':
			opt_p = 1;
			break;
		default:
			usage(1);
		}
	}
	argc -= optind;
	argv += optind;

	/*
	 * get parameter
	 */
	if (0 < argc) {
		sys = *argv;
		argv++;
		argc--;
	} else {
		usage(1);
	}

	/*
	 * write pid to file
	 */
	fp = fopen("pid", "w");
	if (fp == (FILE *) 0) {
		fprintf(stderr, "pid: %s.\n", strerror(errno));
	} else {
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	}

	/* 
	 * Put us in our own process group, which will receive
	 * SIGIO and SIGURG for events on file descriptor fd.
	 *
	 * XXX Handling of these signals should be serialized by
	 * SMP code the kernel.
	 */ 
	ret = setpgrp();
	assert(0 <= ret);

	fi_init(0); /* initialize faultinjection */

	/* setup hardware */
	ret = hwinitialize(argc, argv);
	assert(0 <= ret);

	do {

		/* load image to high memory */
		entry = do_load(argc, argv);
		
		/*
		myfd = open("/home/kingst/mtest/tmp_ram",O_RDWR);
		assert(myfd == 5);
		printf("myfd = %d\n",myfd);

		mymem = mmap((char *) 0x80000000, 0x04000000,
			PROT_READ | PROT_WRITE | PROT_EXEC,
			MAP_SHARED | MAP_FIXED, myfd, 0);
		assert(mymem != MAP_FAILED);

		printf("mmap loaded to 0x%08x\n",mymem);
		*/
		
		ret = ptrace(PTRACE_TRACEME_UMLINUX, 0, 0, 0);

		
		(*entry)();
		
		
		//do_unload();
		ret = 1;

		/* load image to high memory */
		//entry = do_load(argc, argv);

		/* fork (child will run image, parent will remap system calls) */
		//pid = fork();
		//switch (pid) {
		//case -1:
		//fprintf(stderr, "fork: %s.\n", strerror(errno));
		//exit(1);
			/*NOTREACHED*/
		//case 0:
		/* child */
		//ret = ptrace(PTRACE_TRACEME, 0, 0, 0);
		//assert(0 <= ret);
		//pid = getpid();
		//assert(0 < pid);
		//ret = kill(pid, SIGTRAP);
		//assert(0 <= ret);

		//(*entry)();
		///*NOTREACHED*/
		//default:
		//fi_init(pid); /* re-initialize faultinjection */
		//ret = tracer(pid);
		///* fi_exit(); */
		//}

	} while (ret == 0);

	ret = hwshutdown();
	assert(0 <= ret);

	fi_exit() ; /* close faultinjection socket to frontend */

	/*
	 * remove pid file
	 */
	ret = unlink("pid");
	if (ret < 0) {
		fprintf(stderr, "pid: %s.", strerror(errno));
	}


	return (0);
}

/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
